//-----------------------------------------------------------------------------
// Datei: Revo3d.h
// Beschreibung: Smtliche Definitionen fr die Revo3d Engine.
//-----------------------------------------------------------------------------


//Info: Folgende Bibliotheken werden zustzlich bentigt (Linker-Einstellungen):
//winmm.lib dxguid.lib dxerr8.lib d3d8.lib d3dx8.lib dinput8.lib dsound.lib



#include <d3d8.h>
#include <d3dx8.h>
#include <dinput.h>
#include <dsound.h>
#include <dplay8.h>
#include <dxerr8.h>



//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//									DIRECT 3D
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************


const int TL_FVF = D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1;
const int VERTEX_FVF = D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1;

const int OBJ_2DTEXTURE = 0;
const int OBJ_2DTEXT = 1;
const int OBJ_3DOBJ = 2;
const int OBJ_3DPOLYS = 3;

const int NW_BUFSIZE = 128 * 1024;


class Revo3dEngine;
class RevoPhysics;


typedef struct TLVertex
{
	float x, y;			//Bildschirmposition
	float z;			//z-Buffer-Tiefe
	float rhw;
	DWORD Color;		//Farbe
	float tu, tv;		//Texturkoordinaten
} TLVertex;


typedef struct Vertex
{
	float x, y, z;
	float nx, ny, nz;
	float tu, tv;
} Vertex;


typedef struct Revo3dRes
{
	int ResX, ResY;		//Auflsung
	D3DFORMAT Format;	//Pixelformat
} Revo3dRes;


typedef struct Revo3dCaps
{
	int MaxAntiAlias;
	D3DXVECTOR2 MaxTextureSize;
	bool AnisotropicAviable;
	bool FogAviable;
	bool ShadowAviable;
	bool ShadowNoTransparencyAviable;
	bool SpecularAviable;
} Revo3dCaps;


typedef struct Revo3dPlane
{
	D3DXVECTOR3 Points[3], NormalVec;
	D3DXPLANE Plane;
} Revo3dPlane;


typedef struct RegObj
{
	void *pObj;
	int ObjType;
} RegObj;



typedef enum PolyType
{
	POINTLIST = 1,
	LINELIST = 2,
	LINESTRIP = 3,
	TRIANGLELIST = 4,
	TRIANGLESTRIP = 5,
	TRIANGLEFAN = 6
} PolyType;

typedef enum PolyVisibility
{
	NOCHANGE = 0,
	BOTHSIDES = 1,
	BACKSIDE = 2,
	FRONTSIDE = 3
} PolyVisibility;

typedef enum LightType
{
	POINTLIGHT = 1,
	SPOTLIGHT = 2,
	DIRECTIONALLIGHT = 3
} LightType;




class Revo2dTexture
{
private:
	D3DCOLOR colKey;
	bool UseTransMap;
	int MipLevels, eolOffset;
	D3DLOCKED_RECT LockRect;
	D3DSURFACE_DESC TexDesc;
	BYTE **pData;
	LPDIRECT3DTEXTURE8 pTexture;
	LPDIRECT3DDEVICE8 pDevice;
	Revo3dEngine *pEngine;

	void StoreData ();
	void ReStoreData ();

public:
	Revo2dTexture ();
	Revo2dTexture &operator= (Revo2dTexture &pOrig);
	~Revo2dTexture ();
	void Initialize (Revo3dEngine *Eng);
	void Cleanup ();
	bool LoadFromFile (char *File, D3DCOLOR ColorKey, int MipLevels, bool EnableEdit);
	bool LoadForEdit (int x, int y);
	bool EditStart ();
	void Edit (D3DXVECTOR2 *Pos, int a, int r, int g, int b);
	void EditEnd ();

	friend Revo3dEngine;
};


class Revo2dPicture
{
private:
	TLVertex Vert[4], NoRotVert[4];
	float RotAngle;
	D3DXVECTOR2 MP;
	int alpha, red, green, blue;

public:
	Revo2dPicture ();
	void SetPosition (int x1, int y1, int x2, int y2);
	void Move (int MoveX, int MoveY);
	void Rotate (float Angle);
	void SetPictureRange (float x1, float y1, float x2, float y2);
	void SetBackcolor (int r, int g, int b);
	void SetTransparency (int Transparency);

	friend Revo3dEngine;
};


class Revo2dText
{
private:
	HFONT FontHandle;
	LPD3DXFONT pFont;
	char Str[1024];
	RECT Rect;
	DWORD For;
	D3DCOLOR Col;
	LPDIRECT3DDEVICE8 pDevice;
	Revo3dEngine *pEngine;

public:
	Revo2dText ();
	Revo2dText &operator= (Revo2dText &pOrig);
	~Revo2dText ();
	void Initialize (Revo3dEngine *Eng);
	void Cleanup ();
	void SetFont (HFONT Font);
	void SetText (char *String) { strcpy (Str, String); }
	void SetPosition (int x1, int y1, int x2, int y2);
	void SetFormat (int FormatH, int FormatV);
	void SetColor (D3DCOLOR Color) { Col = Color; }

	friend Revo3dEngine;
};


class Revo3dObject
{
private:
	LPD3DXMESH pMesh;
	D3DMATERIAL8 *pMaterials;
	Revo2dTexture *pTextures;
	DWORD MaterialCnt;
	D3DXVECTOR3 Position;
	D3DXMATRIX matTrans, matRot;
	DWORD FVF;
	UINT VertexSize;
	DWORD VertexCnt, FaceCnt;
	BYTE *pVData, *pIData, *pAData;
	LPDIRECT3DDEVICE8 pDevice;
	Revo3dEngine *pEngine;

	void StoreData ();
	void ReStoreData ();
	D3DXVECTOR3 GetExtremePoint (int Index);
	bool GetTransparency (DWORD MaterialIndex);

public:
	Revo3dObject ();
	Revo3dObject &operator= (Revo3dObject &pOrig);
	~Revo3dObject ();
	void Initialize (Revo3dEngine *Eng);
	void Cleanup ();
	bool LoadFromFile (char *File, int MipLevels);
	void SetTransparency (int Transparency);
	void SetSpecPower (float SpecPower);
	
	void SetPosition (D3DXVECTOR3 *Pos);
	void Move (D3DXVECTOR3 *MoveDelta);
	D3DXVECTOR3 *GetPosition () { return (&Position); }
	D3DXVECTOR3 GetLowPoint () { return (GetExtremePoint (0)); }
	D3DXVECTOR3 GetHighPoint () { return (GetExtremePoint (1)); }
	D3DXVECTOR3 GetMidPoint () { return (GetExtremePoint (2)); }
	void SetRotationMatrix (D3DXMATRIX *Matrix);
	D3DXMATRIX *GetRotationMatrix () { return (&matRot); }	
	void SetRotation (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin);
	void Rotate (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin);

	friend Revo3dEngine;
	friend RevoPhysics;
};


class Revo3dPolygons
{
private:
	PolyType PolyFormat;
	Revo2dTexture *pTex;
	D3DMATERIAL8 Mat;
	PolyVisibility Visibility;
	bool Indexed, Dynamic;
	int VertexMaxCnt, IndexMaxCnt;
	BYTE *pVData, *pIData;
	D3DXVECTOR3 Position;
	D3DXMATRIX matTrans, matRot;
	LPDIRECT3DVERTEXBUFFER8 pVBuf;
	LPDIRECT3DINDEXBUFFER8 pIBuf;
	LPDIRECT3DDEVICE8 pDevice;
	Revo3dEngine *pEngine;

	void StoreData ();
	void ReStoreData ();
	//void PolySwap (USHORT VIndex1, USHORT VIndex2, BYTE *pVertices);

public:
	Revo3dPolygons ();
	Revo3dPolygons &operator= (Revo3dPolygons &pOrig);
	~Revo3dPolygons ();
	void Initialize (Revo3dEngine *Eng, bool Dynamic);
	void Cleanup ();
	bool ChangeBufferSize (int NewStaticVCnt, int NewStaticICnt);
	//void TransparencyEnsure (int Offset, int PolyCnt, bool Indexed, D3DXVECTOR3 *pCamPos);

	void SetVertexData (Vertex *vData, int Offset, int cnt);
	void GetVertexData (Vertex *vData, int Offset, int cnt);
	void SetIndexData (unsigned short *iData, int Offset, int cnt);
	void GetIndexData (unsigned short *iData, int Offset, int cnt);
	void SetPolyFormat (PolyType PolyFormat) { this->PolyFormat = PolyFormat; }
	void SetTexture (Revo2dTexture *Tex) { pTex = Tex; }
	void SetMaterial (D3DMATERIAL8 *Mat);
	void SetVisibility (PolyVisibility Visibility) { this->Visibility = Visibility; }
	void SetIndexed (bool Indexed) { this->Indexed = Indexed; }

	void SetPosition (D3DXVECTOR3 *Pos);
	void Move (D3DXVECTOR3 *MoveDelta);
	D3DXVECTOR3 *GetPosition () { return (&Position); }
	void SetRotationMatrix (D3DXMATRIX *Matrix);
	D3DXMATRIX *GetRotationMatrix () { return (&matRot); }
	void SetRotation (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin);
	void Rotate (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin);

	friend Revo3dEngine;	
};


class Revo3dEngine
{
private:
	//Variablen
	//allgemein
	WNDCLASSEX wc;
	HWND hWnd;
	LARGE_INTEGER TimerFrq, TimerBase;
	float TimerLast;
	bool ObjCleaned;
	int RegObjCnt;
	RegObj *RegisterObj;

	//Direct3D-Variablen
	LPDIRECT3DTEXTURE8 LastTexture;
	D3DMATERIAL8 *LastMaterial;
	D3DMATERIAL8 DefaultMaterial;

	int AdapterUse;										//Nummer des Display-Adapters, der verwendet wird
	Revo3dRes Res;										//Auflsung
	int AntiAlias;										//Anti-Aliasing
	bool Windowed;										//Windowed Mode
	int MaxLightIndex;									//Hchster verwendeter Light-Index
	D3DDEVTYPE DevType;									//HW oder SW Rendering
	DWORD TLType;										//HW oder SW T&L
	D3DCAPS8 RdCaps;									//Speichert die Eigenschaften der Grafikkarte
	D3DPRESENT_PARAMETERS PParam;						//Die aktuellen Present-Parameters
	D3DXVECTOR2 LT, RB;									//Links oben und rechts unten des Viewports
	D3DXVECTOR3 CamPos;									//Speichert die Kameraposition

	bool ShadowAble, ShadowNoBlendAble;					//Ob Schatten bzw. ungeblendete Schatten verwendet werden knnen
	bool UseShadow;										//Ob Schatten verwendet wird.
	D3DXMATRIX matShadow;

	bool UseMirror;
	D3DXMATRIX matViewSave;

	Revo3dRes *lpRes;
	LPDIRECT3D8 pD3D;									//Pointer auf Direct3d
	LPDIRECT3DDEVICE8 pD3DDevice;						//Pointer auf das Render-Device


	//Funktionen
	bool IsUsualRes (int ResX, int ResY);
	bool IsRenderState (int State);
	void SetTransform (D3DXMATRIX *matTrans, D3DXMATRIX *matRot, D3DXMATRIX *matMultiply = NULL);
	D3DCULL InvertCullMode (D3DCULL Mode);
	void RegisterObject (void *pObj, int ObjType);
	void UnregisterObject (void *pObj);
	void CleanupObjects ();
	void RestoreObjects ();

public:
	//Funktionen

	Revo3dEngine ();
	~Revo3dEngine ();
	bool Initialize (HWND hWnd, char *WindowName);
	void Cleanup ();
	int GetPossibleRes (bool AllowBelow640, bool AllowUnusualRes, bool Allow32bpp, Revo3dRes **outRes);
	Revo3dCaps GetPossibleCaps (D3DFORMAT Format, bool Windowed);
	D3DMATERIAL8 GetDefaultMaterial () { return (DefaultMaterial); }
	HWND GethWnd () { return (hWnd); }

	bool SetRes (Revo3dRes *Res, int AntiAlias, bool Windowed);
	int GetFormatBPP (D3DFORMAT Format);
	void RenderStart (D3DXVECTOR2 *xy1, D3DXVECTOR2 *xy2);
	void RenderEnd ();
	void RenderShow ();
	void RenderClear (bool ClearScreen, D3DCOLOR ClearColor, bool WillRender3d, bool WillRenderBlendedShadow);
	void Render2dPicture (Revo2dPicture *Pic, Revo2dTexture *Tex);
	void Render2dText (Revo2dText *Text);
	void Render2dLine (int x1, int y1, int x2, int y2, D3DCOLOR Color);
	void Render2dBox (int x1, int y1, int x2, int y2, bool Filled, D3DCOLOR Color);
	void Render3dPolygons (Revo3dPolygons *Polygons, int Offset, int PolyCnt);
	void Render3dObject (Revo3dObject *Obj);
	bool SetShadowState (Revo3dPlane *PlaneTo, D3DXVECTOR3 *LightPos, int Intensity, LightType Type = POINTLIGHT, bool ClipEdges = true);
	void SetReflectionState (Revo3dPlane *PlaneTo, bool ClipEdges = true);
	bool SetLight (int Index, D3DLIGHT8 *Light);
	void SetLightState (int Index, bool State);
	void SetGlobalLight (D3DCOLOR Color);
	void SetSpecularState (bool State);
	void SetZState (bool ReadEnable, bool WriteEnable);
	void SetAnisotropicState (bool State);
	void SetFog (int Density);
	void Revo3dEngine::SetLookDistance (float MaxView, float MaxAngle = D3DX_PI / 4);
	void SetCamera (D3DXVECTOR3 *CameraPos, D3DXVECTOR3 *CameraLookAt, float RollAngle);

	void HandleMessages ();
	void StartTimer ();
	float GetRelTimeDiff ();
	float GetAbsTimeDiff ();

	friend Revo2dTexture;
	friend Revo2dPicture;
	friend Revo2dText;
	friend Revo3dPolygons;
	friend Revo3dObject;
	friend LPDIRECT3DTEXTURE8 TextureMake (Revo3dEngine *Eng, char *File, D3DCOLOR ColorKey, int MipLevels, bool UseTransMap = false);
};



//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//									DIRECT INPUT
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************


class RevoInput
{
private:
	LPDIRECTINPUT8 pDI;									//Pointer auf DirectInput
	LPDIRECTINPUTDEVICE8 pKeyb;							//Pointer auf das Tastatur-Device
	LPDIRECTINPUTDEVICE8 pMouse;						//Pointer auf das Maus-Device
	DIMOUSESTATE MouseData;
	BYTE KeybData[256];
	BYTE KeybDown[256];

	char GetChar (int KeyCode, bool ShiftDown);

public:
	RevoInput ();
	~RevoInput ();
	bool Initialize (HWND hWnd, HINSTANCE hInstance, bool NoWinKeys, bool NoWinMouse);
	void Cleanup ();
	BYTE *ReadKeyboard (char *AppendTo = NULL, int BufferSize = 0);
	DIMOUSESTATE *ReadMouse ();
};



//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//									DIRECT SOUND
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************


const int VolDiff = DSBVOLUME_MAX - DSBVOLUME_MIN;


#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }


class RevoSoundEngine;


class CWaveFile
{
private:
    HRESULT ReadMMIO();

public:
    WAVEFORMATEX* m_pwfx;        // Pointer to WAVEFORMATEX structure
    HMMIO         m_hmmio;       // MM I/O handle for the WAVE
    MMCKINFO      m_ck;          // Multimedia RIFF chunk
    MMCKINFO      m_ckRiff;      // Use in opening a WAVE file
    DWORD         m_dwSize;      // The size of the wave file
    MMIOINFO      m_mmioinfoOut;
    DWORD         m_dwFlags;
    BYTE*         m_pbData;
    BYTE*         m_pbDataCur;
    ULONG         m_ulDataSize;

    CWaveFile();
    ~CWaveFile();
    HRESULT Open( LPTSTR strFileName, WAVEFORMATEX* pwfx );
    HRESULT Close();
    HRESULT Read( BYTE* pBuffer, DWORD dwSizeToRead, DWORD* pdwSizeRead );
    DWORD   GetSize();
	WAVEFORMATEX* GetFormat() { return m_pwfx; }
    HRESULT ResetFile();
};


class Revo2dSound
{
private:
	char File[1024];
	LPDIRECTSOUNDBUFFER pSoundBuf;
	DWORD BufSize;
	LPDIRECTSOUND8 pDSound;

public:
	Revo2dSound ();
	Revo2dSound &operator= (Revo2dSound &pOrig);
	~Revo2dSound ();
	void Initialize (RevoSoundEngine *Snd);
	void Cleanup ();
	bool LoadFromFile (char *File);
	void SetPlayPosition (float Pos);
	void SetVolume (float Vol);

	friend RevoSoundEngine;
};


class RevoMP3
{
private:
	char File[1024];
	char sMP3Ind[32];
	long MP3Len;

public:
	RevoMP3 ();
	RevoMP3 &operator= (RevoMP3 &pOrig);
	~RevoMP3 ();
	void Cleanup ();
	bool LoadFromFile (char *File);
	void SetPlayPosition (float Pos);

	friend RevoSoundEngine;
};


class Revo3dSound
{
private:
	char File[1024];
	LPDIRECTSOUNDBUFFER pSoundBuf;
	LPDIRECTSOUND3DBUFFER pSound3DBuf;
	DWORD BufSize;
	D3DXVECTOR3 Position;
	LPDIRECTSOUND8 pDSound;

public:
	Revo3dSound ();
	Revo3dSound &operator= (Revo3dSound &pOrig);
	~Revo3dSound ();
	void Initialize (RevoSoundEngine *Snd);
	void Cleanup ();
	bool LoadFromFile (char *File);
	void SetDistances (float MinDist, float MaxDist);
	void SetPlayPosition (float Pos);
	void Set3dPosition (D3DXVECTOR3 *Pos);
	void Move3dPosition (D3DXVECTOR3 *MoveDelta);
	void SetSpeed (D3DXVECTOR3 *Speed);
	void SetVolume (float Vol);

	friend RevoSoundEngine;
};


class RevoSoundEngine
{
private:
	LPDIRECTSOUNDBUFFER pBufPrimary;
	LPDIRECTSOUND3DLISTENER p3DListener;
	LPDIRECTSOUND8 pDSound;
	D3DXVECTOR3 Position;

public:
	RevoSoundEngine ();
	~RevoSoundEngine ();
	bool Initialize (HWND hWnd);
	void Cleanup ();
	void SetListener3dPosition (D3DXVECTOR3 *Pos, D3DXVECTOR3 *LookAt);
	void MoveListener3dPosition (D3DXVECTOR3 *MoveDelta, D3DXVECTOR3 *LookAt);
	void SetListenerSpeed (D3DXVECTOR3 *Speed);
	void Play2dSound (Revo2dSound *Sound, bool Looping);
	void Stop2dSound (Revo2dSound *Sound);
	void PlayMP3 (RevoMP3 *MP3, bool Looping);
	void StopMP3 (RevoMP3 *MP3);
	void Play3dSound (Revo3dSound *Sound, bool Looping);
	void Stop3dSound (Revo3dSound *Sound);
	void Update3dSounds ();

	friend Revo2dSound;
	friend Revo3dSound;
};



//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//									DIRECT PLAY
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************


const BYTE MSGTYPE_PLAYERLIST = 0;
const BYTE MSGTYPE_DATA = 1;


typedef struct ServerDesc
{
	char *IP;
	char *SessionName;
} ServerDesc;


class RevoClient
{
private:
	IDirectPlay8Client *pClient;
	IDirectPlay8Address *pAddrClient, *pAddrServer;
	DPN_APPLICATION_DESC AppDesc;
	GUID Ident;
	static void (* fp) (BYTE *, int);
	static DPNID *pClientIdents;
	static int NrClients;

public:
	RevoClient ();
	~RevoClient ();
	bool Initialize (unsigned int Identifier, unsigned int Port, void (* Callback) (BYTE *, int));
	void Cleanup ();
	int GetServers (float TimeOut, ServerDesc **outServers);
	bool Connect (char *IP, float TimeOut);
	void Disconnect ();
	int GetPlayers (unsigned long **outIdentifiers);
	bool SendData (unsigned long DestIdentifier, BYTE *pData, unsigned int DataSize, float TimeOut);
	
	friend HRESULT WINAPI ClientMsgHandler (void *pvUserContext, DWORD dwMessageId, void *pMsgBuffer);
};

HRESULT WINAPI ClientMsgHandler (void *pvUserContext, DWORD dwMessageId, void *pMsgBuffer);


class RevoServer
{
private:
	static IDirectPlay8Server *pServer;
	IDirectPlay8Address *pAddrServer;
	DPN_APPLICATION_DESC AppDesc;
	GUID Ident;
	static void (* fp) (BYTE *, int);
	static bool AllowClientConn;
	static DPNID *pClientIdents;
	static int NrClients;

	static void AddPlayer (DPNID Ident);
	static void DeletePlayer (DPNID Ident);
	static void SendPlayerList ();

public:
	RevoServer ();
	~RevoServer ();
	bool Initialize (unsigned int Identifier, unsigned int Port, void (* Callback) (BYTE *, int));
	void Cleanup ();
	bool Host (char *SessionName, int MaxClients);
	void Unhost ();
	void SetConnectionState (bool AllowClientConnection);
	int GetPlayers (unsigned long **outIdentifiers);
	static bool SendData (unsigned long DestIdentifier, BYTE *pData, unsigned int DataSize, float TimeOut);
	int GetQueueSize (unsigned long DestIdentifier);

	friend HRESULT WINAPI ServerMsgHandler (void *pvUserContext, DWORD dwMessageId, void *pMsgBuffer);
};

HRESULT WINAPI ServerMsgHandler (void *pvUserContext, DWORD dwMessageId, void *pMsgBuffer);



//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//									FUNKTIONEN
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************

//RevoFunctions
//2d-Vektoren
D3DXVECTOR2 Vector2dMake (float x, float y);
D3DXVECTOR2 NormalVector2dMake (D3DXVECTOR2* v1, D3DXVECTOR2 *v2);
D3DXVECTOR2 Vector2dRotate (D3DXVECTOR2 *Vector, float Angle, D3DXVECTOR2 *Origin);
//3d-Vektoren
D3DXVECTOR3 Vector3dMake (float x, float y, float z);
D3DXVECTOR3 NormalVector3dMake (D3DXVECTOR3* v1, D3DXVECTOR3* v2, D3DXVECTOR3* v3);
D3DXVECTOR3 Vector3dRotate (D3DXVECTOR3 *Vector, D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin);
//Ebenen
Revo3dPlane PlaneMake (D3DXVECTOR3 *p1, D3DXVECTOR3 *p2, D3DXVECTOR3 *p3);
//Vertexe
Vertex VertexMake (float x, float y, float z, float nx, float ny, float nz, float tu, float tv);
TLVertex TLVertexMake (float x, float y, D3DCOLOR Color, float tu, float tv);
void BillboardVertexMake (D3DXVECTOR3 *MidPos, float Width, float Height, D3DXVECTOR3 *CamPos, bool AllowXRotation, Vertex *outV);
//Farben
D3DCOLOR ColorMake (int r, int g, int b);
D3DCOLOR ColorAlphaMake (int a, int r, int g, int b);
//Sonstiges
LPDIRECT3DTEXTURE8 TextureMake (Revo3dEngine *Eng, char *File, D3DCOLOR ColorKey, int MipLevels, bool EnableEdit);
D3DMATERIAL8 MaterialMake (int rReflect, int gReflect, int bReflect, int rOwn, int gOwn, int bOwn, float SpecPower, int Transparency);
D3DLIGHT8 LightMake (D3DLIGHTTYPE Type, int rEmit, int gEmit, int bEmit, int rSpecular, int gSpecular, int bSpecular, D3DVECTOR *Position, D3DVECTOR *Direction, float Range, float Attenuation, float InnerCore, float OuterCore);
D3DLIGHT8 PointLightMake (int rEmit, int gEmit, int bEmit, int Specular, D3DVECTOR *Position, float Range);
D3DLIGHT8 SpotLightMake (int rEmit, int gEmit, int bEmit, int Specular, D3DVECTOR *Position, D3DVECTOR *Direction, float Range, float InnerCore, float OuterCore);
D3DLIGHT8 DirectionalLightMake (int rEmit, int gEmit, int bEmit, int Specular, D3DVECTOR *Direction);
HFONT FontMake (char *FontName, int Size, bool Bold, bool Italic, bool Underline);
Revo3dRes ResolutionMake (int x, int y, int bpp);

//Eigene Funktionen ("private", werden nur intern verwendet)
LRESULT CALLBACK WindowProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
D3DXMATRIX RotationMatrixMake (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order);
D3DXVECTOR3 MirrorPoint (Revo3dPlane *pPlane, D3DXVECTOR3 *pPoint);
void GetFolder (char *Folder, char *File);
GUID GuidMake (unsigned int Identifier);
WCHAR *UnicodeStringMake (char *String);
char *UnicodeStringTranslate (WCHAR *UnicodeString);