//-----------------------------------------------------------------------------
// Datei: Revo3dEngine.cpp
// Beschreibung: Enthlt die gesamte Revo3dEngine.
//-----------------------------------------------------------------------------


#include <windows.h>
#include "revo3d.h"




//******************** GLOBALE FUNKTIONEN ********************

//********** INITSIALISIEREN **********



Revo3dEngine::Revo3dEngine ()
{
	pD3D = NULL;
	pD3DDevice = NULL;
	lpRes = NULL;
	wc.hInstance = NULL;
	hWnd = 0;
	RegisterObj = NULL;
}

Revo3dEngine::~Revo3dEngine ()
{
	Cleanup ();
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert die Revo3d Engine.
// Parameter: hWnd (Definiert das Fenster, das verwendet werden soll. Wird 0 bergeben,
// erstellt die Revo3dEngine selbst ein Fenster.), WindowName (Der Name des Fensters, das
// erstellt wird. Wird nur bercksichtigt falls hWnd 0 ist.)
//-----------------------------------------------------------------------------
bool Revo3dEngine::Initialize (HWND hWnd, char *WindowName)
{
	int i;
	int cntAdapter;
	bool HWTL;

	//Variablen initsialisieren
	Cleanup ();
	AdapterUse = -1;
	HWTL = false;
	ShadowAble = false;
	ShadowNoBlendAble = false;
	UseShadow = false;
	UseMirror = false;
	LastTexture = NULL;
	LastMaterial = NULL;
	DefaultMaterial = MaterialMake (255, 255, 255, 0, 0, 0, 0, 0);
	ObjCleaned = false;
	RegObjCnt = 0;
	MaxLightIndex = -1;

	if (!hWnd)
	{
		//Eine Klasse und ein Fenster erstellen
		wc.cbSize = sizeof (WNDCLASSEX);
		wc.style = CS_CLASSDC;
		wc.lpfnWndProc = WindowProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = GetModuleHandle (NULL);
		wc.hIcon = NULL;
		wc.hCursor = NULL;
		wc.hbrBackground = NULL;
		wc.lpszMenuName = NULL;
		wc.lpszClassName = "RenderClass";
		wc.hIconSm = NULL;
		if (!RegisterClassEx (&wc))
			return (false);

		this->hWnd = CreateWindow ("RenderClass", WindowName, WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, NULL, CW_USEDEFAULT, NULL, GetDesktopWindow (), NULL, wc.hInstance, NULL);
		if (!this->hWnd)
			return (false);
	}
	else
		this->hWnd = hWnd;
	ShowWindow (this->hWnd, SW_SHOWDEFAULT);
	UpdateWindow (this->hWnd);


	//********************DIRECT 3D********************
	if (!(pD3D = Direct3DCreate8 (D3D_SDK_VERSION)))			//erstellt das D3D-Objekt
		return (false);

	cntAdapter = pD3D->GetAdapterCount ();
	for (i = 0; i < cntAdapter; i++)
	{
		//Ermitteln der bestmglichen Rendereinstellungen (Hardware Rendering, Hardware T&L).
		if (FAILED (pD3D->GetDeviceCaps (i, D3DDEVTYPE_HAL, &RdCaps)))
			continue;
		//Es muss entweder kein Vertical-Sync oder Refresh-Rate flipping supported werden.
		if (!(RdCaps.PresentationIntervals & D3DPRESENT_INTERVAL_IMMEDIATE) && !(RdCaps.PresentationIntervals & D3DPRESENT_INTERVAL_ONE))
			continue;
		if (HWTL && !(RdCaps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT))
			continue;
		//Adapter (Grafikkarte) kann prinzipiell verwendet werden, zustzliche Informationen auslesen
		if (RdCaps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT)
			HWTL = true;
		AdapterUse = i;
	}
	if (AdapterUse == -1)
		return (false);

	//HW T&L setzen, falls mglich
	if (HWTL)
		TLType = D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		TLType = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert die Revo3dEngine.
//-----------------------------------------------------------------------------
void Revo3dEngine::Cleanup ()
{
	if (RegisterObj)
	{
		CleanupObjects ();
		free (RegisterObj);
		RegisterObj = NULL;
	}
	if (lpRes)
	{
		free (lpRes);
		lpRes = NULL;
	}
	if (pD3DDevice)
	{
		pD3DDevice->SetTexture (0, NULL);
		pD3DDevice->Release ();
		pD3DDevice = NULL;
	}
	if (pD3D)
	{
		pD3D->Release ();
		pD3D = NULL;
	}
	if (hWnd)
	{
		DestroyWindow (hWnd);
		hWnd = 0;
	}
	if (wc.hInstance)
	{
		UnregisterClass ("RenderClass", wc.hInstance);
		wc.hInstance = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: GetPossibleRes
// Beschreibung: Liefert alle Auflsungen zurck, die verwendet werden knnen.
// Parameter: AllowBelow640 (ob Auflsungen < 640x480 verwendet werden drfen),
// AllowUnusualRes (ob unbliche Auflsungen verwendet werden
// drfen), Allow32bpp (ob 32 Bit Farbtiefe erlaubt ist), **outRes
// (die Adresse eines Revo3dRes Pointers auf den Anfang des Feldes)
// Rckgabewert: Anzahl der Auflsungen
//-----------------------------------------------------------------------------
int Revo3dEngine::GetPossibleRes (bool AllowBelow640, bool AllowUnusualRes, bool Allow32bpp, Revo3dRes **outRes)
{
	if (!pD3D || !outRes)
		return (0);

	UINT i, j;
	UINT cntMode, cntAct = 0;
	bool ExAlr;
	D3DDISPLAYMODE RdMode;

	if (lpRes)
	{
		free (lpRes);
		lpRes = NULL;
	}
	cntMode = pD3D->GetAdapterModeCount (AdapterUse);
	for (i = 0; i < cntMode; i++)
	{
		//Alle mglichen Auflsungen durchlaufen
		pD3D->EnumAdapterModes (AdapterUse, i, &RdMode);
		if (!AllowBelow640 && (RdMode.Width < 640 || RdMode.Height < 480))
			continue;
		if (!AllowUnusualRes && !IsUsualRes (RdMode.Width, RdMode.Height))
			continue;
		if (RdMode.Format != D3DFMT_R5G6B5 && RdMode.Format != D3DFMT_X8R8G8B8)
			continue;

		//Prfen, ob dieses Auflsung schon vorkommt, nur mit einer anderen Refresh-Rate
		ExAlr = false;
		for (j = 0; j < cntAct; j++)
		{
			if (lpRes[j].ResX == (int) RdMode.Width && lpRes[j].ResY == (int) RdMode.Height && lpRes[j].Format == RdMode.Format)
			{
				ExAlr = true;
				break;
			}
		}
		//Wenn diese Auflsung noch nicht existiert, wird sie hinzugefgt
		if (!ExAlr)
		{
			if (!lpRes)
				lpRes = (Revo3dRes *) malloc (sizeof (Revo3dRes));
			else
				lpRes = (Revo3dRes *) realloc (lpRes, sizeof (Revo3dRes) * (cntAct + 1));
			lpRes[cntAct].ResX = RdMode.Width;
			lpRes[cntAct].ResY = RdMode.Height;
			lpRes[cntAct].Format = RdMode.Format;
			cntAct++;
		}
	}
	*outRes = lpRes;
	return (cntAct);
}

//-----------------------------------------------------------------------------
// Funktion: GetPossibleCaps
// Beschreibung: Gibt zurck, welche Eigenschaften die Hardware beherrscht und fr
// die Engine relevant sind.
// Parameter: bpp (die Farbtiefe, die spter eingestellt werden soll), Windowed (ob
// die Anwendung spter im Fenster laufen wird oder nicht)
//-----------------------------------------------------------------------------
Revo3dCaps Revo3dEngine::GetPossibleCaps (D3DFORMAT Format, bool Windowed)
{
	int i;
	D3DDISPLAYMODE RdMode;
	D3DFORMAT BackBufFormat;
	Revo3dCaps retCaps;

	//Backbuffer-Modus feststellen
	if (Windowed)
	{
		pD3D->GetAdapterDisplayMode (AdapterUse, &RdMode);
		BackBufFormat = RdMode.Format;
	}
	else
		BackBufFormat = Format;
	//Caps-Struktur ausfllen
	ZeroMemory (&retCaps, sizeof (Revo3dCaps));
	if (!pD3D)
		return (retCaps);
	//Anisotropischer Filter
	if (RdCaps.MaxAnisotropy > 1)
		retCaps.AnisotropicAviable = true;
	//Nebel
	if (RdCaps.RasterCaps & D3DPRASTERCAPS_ZFOG)
		retCaps.FogAviable = true;
	//Anti-Aliasing
	for (i = 2; i <= 8; i *= 2)
	{
		if (FAILED (pD3D->CheckDeviceMultiSampleType (AdapterUse, D3DDEVTYPE_HAL, BackBufFormat, Windowed, (D3DMULTISAMPLE_TYPE) i)))
			break;
	}
	retCaps.MaxAntiAlias = i / 2;
	if (retCaps.MaxAntiAlias == 1)
		retCaps.MaxAntiAlias = 0;
	//Texturengre
	retCaps.MaxTextureSize.x = (float) RdCaps.MaxTextureWidth;
	retCaps.MaxTextureSize.y = (float) RdCaps.MaxTextureHeight;
	//Schatten
	if (RdCaps.RasterCaps & D3DPRASTERCAPS_ZBIAS)
	{
		if (FAILED (pD3D->CheckDepthStencilMatch (AdapterUse, D3DDEVTYPE_HAL, BackBufFormat, BackBufFormat, D3DFMT_D15S1)))
		{
			if (FAILED (pD3D->CheckDepthStencilMatch (AdapterUse, D3DDEVTYPE_HAL, BackBufFormat, BackBufFormat, D3DFMT_D24S8)))
			{
				if (SUCCEEDED (pD3D->CheckDepthStencilMatch (AdapterUse, D3DDEVTYPE_HAL, BackBufFormat, BackBufFormat, D3DFMT_D24X4S4)))
					retCaps.ShadowAviable = true;
			}
			else
				retCaps.ShadowAviable = true;
		}
		else
			retCaps.ShadowAviable = true;
		retCaps.ShadowAviable = retCaps.ShadowAviable && (RdCaps.StencilCaps & D3DSTENCILCAPS_DECR);
		retCaps.ShadowNoTransparencyAviable = true;
	}
	//Glanzlichter
	if (RdCaps.ShadeCaps & D3DPSHADECAPS_SPECULARGOURAUDRGB)
		retCaps.SpecularAviable = true;

	return (retCaps);
}



//-----------------------------------------------------------------------------
// Funktion: SetRes
// Beschreibung: Schaltet in eine bestimmte Auflsung um.
// Parameter: Res (die Auflsung, die eingestellt werden soll), AntiAlias (das zu
// verwendende Anti-Aliasing), Windowed (Ob die Anwendung im Fenster laufen soll
// oder im Fullscreen-Modus. Im Fenster-Modus wird der bpp-Wert ignoriert.)
// Bemerkung: Setzt als Default-Eigenschaft die Sichtweite auf 1000 Meter, die
// Kameraposition auf (0, 0, 0), den Zielpunkt der Kamera auf (0, 0, 1), die
// Kamera-Drehung auf 0 und den Texturenfilter auf linear.
//-----------------------------------------------------------------------------
bool Revo3dEngine::SetRes (Revo3dRes *Res, int AntiAlias, bool Windowed)
{
	if (!pD3D || !Res)
		return (false);	

	int i;
	D3DPRESENT_PARAMETERS PresentParam;
	D3DDISPLAYMODE RdMode;
	D3DFORMAT BackBufFormat, StencilFormat;
	D3DXMATRIX RdMatrix[2];
	D3DLIGHT8 *pRdLight;
	int *pRdLightEnable;
	DWORD RdState[255], TexState;
	D3DVIEWPORT8 RdVPort;
	bool Failed = false, PrevSA, PrevSBA, PrevWindowed;
	Revo3dRes PrevRes;
	int PrevAntiAlias;


	//Prensent-Parameter initsialisieren
	ZeroMemory (&PresentParam, sizeof (D3DPRESENT_PARAMETERS));
	PresentParam.BackBufferWidth = Res->ResX;
	PresentParam.BackBufferHeight = Res->ResY;
	if (Windowed)
	{
		PresentParam.Windowed = true;
		pD3D->GetAdapterDisplayMode (AdapterUse, &RdMode);
		BackBufFormat = RdMode.Format;
	}
	else
	{
		BackBufFormat = Res->Format;
		if (RdCaps.PresentationIntervals & D3DPRESENT_INTERVAL_IMMEDIATE)
			PresentParam.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
		else if (RdCaps.PresentationIntervals & D3DPRESENT_INTERVAL_ONE)
			PresentParam.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	}
	PresentParam.BackBufferFormat = BackBufFormat;
	PresentParam.BackBufferCount = 1;
	PresentParam.EnableAutoDepthStencil = true;
	//Testen, ob Stencil Buffer mglich ist. Wird zum Rendern von Schatten bentigt.
	if (FAILED (pD3D->CheckDepthStencilMatch (AdapterUse, D3DDEVTYPE_HAL, BackBufFormat, BackBufFormat, D3DFMT_D15S1)))
	{
		if (FAILED (pD3D->CheckDepthStencilMatch (AdapterUse, D3DDEVTYPE_HAL, BackBufFormat, BackBufFormat, D3DFMT_D24S8)))
		{
			if (FAILED (pD3D->CheckDepthStencilMatch (AdapterUse, D3DDEVTYPE_HAL, BackBufFormat, BackBufFormat, D3DFMT_D24X4S4)))
				//Stencil Buffer ist nicht supported
				StencilFormat = D3DFMT_D16;
			else
				StencilFormat = D3DFMT_D24X4S4;
		}
		else
			StencilFormat = D3DFMT_D24S8;
	}
	else
		StencilFormat = D3DFMT_D15S1;
	PrevSA = ShadowAble;								//Sichern fr den Fall, das das Device wiederhergestellt werden muss
	PrevSBA = ShadowNoBlendAble;
	if (RdCaps.RasterCaps & D3DPRASTERCAPS_ZBIAS)
	{
		ShadowNoBlendAble = true;
		if (StencilFormat != D3DFMT_D16)
			ShadowAble = true;
		else
			ShadowAble = false;
	}
	PresentParam.AutoDepthStencilFormat = StencilFormat;
	PresentParam.SwapEffect = D3DSWAPEFFECT_DISCARD;
	PresentParam.MultiSampleType = (D3DMULTISAMPLE_TYPE) AntiAlias;
	PresentParam.hDeviceWindow = hWnd;

	PrevRes = this->Res;								//Sichern fr den Fall, das das Device wiederhergestellt werden muss
	PrevAntiAlias = this->AntiAlias;
	PrevWindowed = this->Windowed;
	this->Res = *Res;
	this->AntiAlias = AntiAlias;
	this->Windowed = Windowed;

	ShowWindow (hWnd, SW_SHOWDEFAULT);
	UpdateWindow (hWnd);
	if (!pD3DDevice)
	{
		//Neu erstellen
		if (FAILED (pD3D->CreateDevice (AdapterUse, D3DDEVTYPE_HAL, hWnd, TLType, &PresentParam, &pD3DDevice)))
			return (false);
		PParam = PresentParam;
		SetLookDistance (1000);
		SetCamera (&Vector3dMake (0, 0, 0), &Vector3dMake (0, 0, 1), 0);
		SetAnisotropicState (false);
		//Viewport setzen
		LT.x = 0;
		LT.y = 0;
		RB.x = (float) this->Res.ResX;
		RB.y = (float) this->Res.ResY;

		//Allgemeine Render-Eigenschaften setzen
		pD3DDevice->SetRenderState (D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
		pD3DDevice->SetRenderState (D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
		pD3DDevice->SetRenderState (D3DRS_FOGCOLOR, ColorMake (150, 150, 150));
		pD3DDevice->SetRenderState (D3DRS_FOGTABLEMODE, D3DFOG_EXP2);
		pD3DDevice->SetRenderState (D3DRS_STENCILFUNC, D3DCMP_LESS);
		pD3DDevice->SetRenderState (D3DRS_STENCILPASS, D3DSTENCILOP_DECR);
	}
	else
	{
		//Reseten.
		//Aktuelle Rendereinstellungen sichern
		pD3DDevice->GetTransform (D3DTS_VIEW, &RdMatrix[0]);
		pD3DDevice->GetTransform (D3DTS_PROJECTION, &RdMatrix[1]);
		if (MaxLightIndex != -1)
		{
			pRdLight = new D3DLIGHT8[MaxLightIndex + 1];
			pRdLightEnable = new int[MaxLightIndex + 1];
			for (i = 0; i <= MaxLightIndex; i++)
			{
				pD3DDevice->GetLight (i, &pRdLight[i]);
				pD3DDevice->GetLightEnable (i, &pRdLightEnable[i]);
			}
		}
		for (i = 0; i < 255; i++)
		{
			if (IsRenderState (i))
				pD3DDevice->GetRenderState ((D3DRENDERSTATETYPE) i, &RdState[i]);
		}
		pD3DDevice->GetTextureStageState (0, D3DTSS_MAGFILTER, &TexState);
		pD3DDevice->GetViewport (&RdVPort);

		//Alle Objekte deallokieren, Device reseten und Objekte wiederherstellen
		pD3DDevice->SetTexture (0, NULL);
		LastTexture = NULL;
		CleanupObjects ();

		if (FAILED (pD3DDevice->Reset (&PresentParam)))
		{
			//Reset ist fehlgeschlagen. Versuchen, das vorige Device wiederherzustellen (das offenbar funktionierte)
			if (FAILED (pD3DDevice->Reset (&PParam)))
			{
				if (MaxLightIndex != -1)
				{
					delete []pRdLight;
					delete []pRdLightEnable;
				}
				return (false);
			}
			//Gesicherte Daten vom alten Device wiederherstellen
			ShadowAble = PrevSA;
			ShadowNoBlendAble = PrevSBA;
			this->Res = PrevRes;
			this->AntiAlias = PrevAntiAlias;
			this->Windowed = PrevWindowed;
			Failed = true;
		}
		else
			PParam = PresentParam;
		RestoreObjects ();

		//Vorige Rendereinstellungen wiederherstellen
		pD3DDevice->SetTransform (D3DTS_VIEW, &RdMatrix[0]);
		pD3DDevice->SetTransform (D3DTS_PROJECTION, &RdMatrix[1]);
		if (MaxLightIndex != -1)
		{
			for (i = 0; i <= MaxLightIndex; i++)
			{
				pD3DDevice->SetLight (i, &pRdLight[i]);
				pD3DDevice->LightEnable (i, pRdLightEnable[i]);
			}
			delete []pRdLight;
			delete []pRdLightEnable;
		}
		for (i = 0; i < 255; i++)
		{
			if (IsRenderState (i))
				pD3DDevice->SetRenderState ((D3DRENDERSTATETYPE) i, RdState[i]);
		}
		pD3DDevice->SetTextureStageState (0, D3DTSS_MAGFILTER, TexState);
		pD3DDevice->SetTextureStageState (0, D3DTSS_MINFILTER, TexState);
		pD3DDevice->SetViewport (&RdVPort);
	}

	//Allgemeine Texture-Eigenschaften setzen
	pD3DDevice->SetTextureStageState (0, D3DTSS_MAXANISOTROPY, RdCaps.MaxAnisotropy);
	pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE);
	pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);

	return (!Failed);
}

//-----------------------------------------------------------------------------
// Funktion: GetFormatBPP
// Beschreibung: Gibt die Farbtiefe eines bestimmten Pixelformats zurck.
// Parameter: Format (das Pixelformat)
// Rckgabewert: Die bpp oder 0 falls das Format unbekannt ist.
//-----------------------------------------------------------------------------
int Revo3dEngine::GetFormatBPP (D3DFORMAT Format)
{
	if (Format == D3DFMT_R3G3B2 || Format == D3DFMT_A8)
		return (8);
	else if (Format == D3DFMT_R5G6B5 || Format == D3DFMT_X1R5G5B5 || Format == D3DFMT_A1R5G5B5 || Format == D3DFMT_A4R4G4B4 || Format == D3DFMT_A8R3G3B2 || Format == D3DFMT_X4R4G4B4)
		return (16);
	else if (Format == D3DFMT_R8G8B8)
		return (24);
	else if (Format == D3DFMT_A8R8G8B8 || Format == D3DFMT_X8R8G8B8)
		return (32);
	else
		return (0);
}



//********** RENDERN **********

//-----------------------------------------------------------------------------
// Funktion: RenderClear
// Beschreibung: Lscht den Bildschirm und trifft zustzliche Vorbereitungen.
// Parameter: ClearScreen (ob der Bildschirm gelscht werden soll),
// ClearColor (Farbe, mit der der Bildschirm gelscht wird),
// WillRender3d (ob irgendwelche 3d-Objekte in diesem Frame gerendert werden),
// WillRenderBlendedShadow (ob irgendwelche Schatten, die eine Intensitt von
// weniger als 255 besitzen, in diesem Frame gerendert werden)
//-----------------------------------------------------------------------------
void Revo3dEngine::RenderClear (bool ClearScreen, D3DCOLOR ClearColor, bool WillRender3d, bool WillRenderBlendedShadow)
{
	if (!pD3DDevice)
		return;

	DWORD Flags = 0;

	if (ClearScreen)
		Flags = D3DCLEAR_TARGET;
	if (WillRender3d)
		Flags = Flags | D3DCLEAR_ZBUFFER;
	if (ShadowAble && WillRenderBlendedShadow)
		Flags = Flags | D3DCLEAR_STENCIL;
	pD3DDevice->Clear (0, NULL, Flags, ClearColor, 1, 1);
}



//-----------------------------------------------------------------------------
// Funktion: RenderStart
// Beschreibung: Muss zu Beginn des Renderns eines Fensters aufgerufen werden.
// Parameter: xy1 (Die linke, obere Ecke des Zielfensters. Ist dieser Parameter
// NULL, wird die linke, obere Ecke des Bildschirms verwendet.), xy2 (Die rechte,
// untere Ecke des Zielfensters. Ist dieser Parameter NULL, wird die rechte, untere
// Ecke des Bildschirms verwendet.)
// Bemerkung: Wenn der Fokus verloren wird, wird diese Funktion so lange ausgefhrt,
// bis der Fokus wieder erhalten wurde. Der Timer wird nicht angehalten.
//-----------------------------------------------------------------------------
void Revo3dEngine::RenderStart (D3DXVECTOR2 *xy1, D3DXVECTOR2 *xy2)
{
	if (!pD3DDevice)
		return;

	HRESULT hr;
	D3DVIEWPORT8 vPort;
	D3DXVECTOR2 cmpVec[2];

	do
	{
		//Messages verarbeiten lassen
		HandleMessages ();
		//Device reseten, falls ntig
		hr = pD3DDevice->TestCooperativeLevel ();
		if (hr == D3DERR_DEVICENOTRESET)
		{
			if (!SetRes (&Res, AntiAlias, Windowed))
				hr = E_FAIL;
			else
				hr = S_OK;
		}
	}
	while (FAILED (hr));

	if (FAILED (pD3DDevice->BeginScene ()))
		return;

	if (xy1)
		cmpVec[0] = *xy1;
	else
		ZeroMemory (cmpVec, sizeof (D3DXVECTOR2));
	if (xy2)
		cmpVec[1] = *xy2;
	else
	{
		cmpVec[1].x = (float) Res.ResX;
		cmpVec[1].y = (float) Res.ResY;
	}
	if (cmpVec[0].x != LT.x || cmpVec[0].y != LT.y || cmpVec[1].x != RB.x || cmpVec[1].y != RB.y)
	{
		//Viewport neu setzen
		vPort.X = (DWORD) cmpVec[0].x;
		vPort.Y = (DWORD) cmpVec[0].y;
		vPort.Width = (DWORD) (cmpVec[1].x - cmpVec[0].x);
		vPort.Height = (DWORD) (cmpVec[1].y - cmpVec[0].y);
		vPort.MinZ = 0;
		vPort.MaxZ = 1;
		pD3DDevice->SetViewport (&vPort);
		LT = cmpVec[0];
		RB = cmpVec[1];
	}
}



//-----------------------------------------------------------------------------
// Funktion: RenderEnd
// Beschreibung: Muss beim Ende des Renderns eines Fensters aufgerufen werden.
//-----------------------------------------------------------------------------
void Revo3dEngine::RenderEnd ()
{
	if (!pD3DDevice)
		return;
	pD3DDevice->EndScene ();
}



//-----------------------------------------------------------------------------
// Funktion: RenderShow
// Beschreibung: Muss beim Ende des Renderns eines Frames aufgerufen werden.
//-----------------------------------------------------------------------------
void Revo3dEngine::RenderShow ()
{
	if (!pD3DDevice)
		return;
	pD3DDevice->Present (NULL, NULL, hWnd, NULL);
}



//-----------------------------------------------------------------------------
// Funktion: Render2dPicture
// Beschreibung: Rendert ein 2d-Bitmap.
// Parameter: Pic (das Bild, das gerendert werden soll), Tex (die Textur, die
// zum Rendern des Bildes verwendet werden soll).
//-----------------------------------------------------------------------------
void Revo3dEngine::Render2dPicture (Revo2dPicture *Pic, Revo2dTexture *Tex)
{
	if (!pD3DDevice || !Pic || !Tex || !Tex->pTexture)
		return;

	bool UseTrans, UseKey;

	if (255 - Pic->alpha)
		UseTrans = true;
	else
		UseTrans = false;
	if (Tex->colKey)
		UseKey = true;
	else
		UseKey = false;
	pD3DDevice->SetVertexShader (TL_FVF);
	if (UseTrans || UseKey)
	{
		//Eine Blending-Operation ist erforderlich, daher Alpha Blending aktivieren
		pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, true);
		if (UseTrans && UseKey)
			pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
		else if (UseTrans)
			pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
		else
			pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);
	}
	if (LastTexture != Tex->pTexture)
	{
		pD3DDevice->SetTexture (0, Tex->pTexture);
		LastTexture = Tex->pTexture;
	}
	pD3DDevice->DrawPrimitiveUP (D3DPT_TRIANGLESTRIP, 2, Pic->Vert, sizeof (TLVertex));
	pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, false);
}



//-----------------------------------------------------------------------------
// Funktion: Render2dText
// Beschreibung: Rendert einen 2d-Text.
// Parameter: Text (der Text, der gerendert werden soll)
//-----------------------------------------------------------------------------
void Revo3dEngine::Render2dText (Revo2dText *Text)
{
	if (!Text || !Text->pFont)
		return;

	Text->pFont->DrawTextA (Text->Str, -1, &Text->Rect, Text->For, Text->Col);
}



//-----------------------------------------------------------------------------
// Funktion: Render2dLine
// Beschreibung: Rendert eine 2d-Linie.
// Parameter: x1, y1 (1. Punkt), x2, y2 (2. Punkt), Color (die Farbe)
//-----------------------------------------------------------------------------
void Revo3dEngine::Render2dLine (int x1, int y1, int x2, int y2, D3DCOLOR Color)
{
	if (!pD3DDevice)
		return;

	TLVertex v[2];

	v[0] = TLVertexMake ((float) x1, (float) y1, Color, 0, 0);
	v[1] = TLVertexMake ((float) x2, (float) y2, Color, 0, 0);
	pD3DDevice->SetVertexShader (TL_FVF);
	if (LastTexture)
	{
		pD3DDevice->SetTexture (0, NULL);
		LastTexture = NULL;
	}
	pD3DDevice->DrawPrimitiveUP (D3DPT_LINELIST, 1, v, sizeof (TLVertex));
}



//-----------------------------------------------------------------------------
// Funktion: Render2dBox
// Beschreibung: Rendert ein 2d-Rechteck.
// Parameter: x1, y1 (die linke, obere Ecke), x2, y2 (die rechte, untere Ecke),
// Filled (ob das Rechteck ausgefllt werden soll), Color (die Farbe)
//-----------------------------------------------------------------------------
void Revo3dEngine::Render2dBox (int x1, int y1, int x2, int y2, bool Filled, D3DCOLOR Color)
{
	if (!pD3DDevice)
		return;	

	TLVertex v[5];

	pD3DDevice->SetVertexShader (TL_FVF);
	if (LastTexture)
	{
		pD3DDevice->SetTexture (0, NULL);
		LastTexture = NULL;
	}
	v[0] = TLVertexMake ((float) x1, (float) y1, Color, 0, 0);
	v[1] = TLVertexMake ((float) x2, (float) y1, Color, 0, 0);
	if (Filled)
	{
		v[2] = TLVertexMake ((float) x1, (float) y2, Color, 0, 0);
		v[3] = TLVertexMake ((float) x2, (float) y2, Color, 0, 0);
		pD3DDevice->DrawPrimitiveUP (D3DPT_TRIANGLESTRIP, 2, v, sizeof (TLVertex));
	}
	else
	{
		v[2] = TLVertexMake ((float) x2, (float) y2, Color, 0, 0);
		v[3] = TLVertexMake ((float) x1, (float) y2, Color, 0, 0);
		v[4] = v[0];
		pD3DDevice->DrawPrimitiveUP (D3DPT_LINESTRIP, 4, v, sizeof (TLVertex));
	}
}



//-----------------------------------------------------------------------------
// Funktion: Render3dPolygons
// Beschreibung: Rendert benutzerdefinierte 3d-Polygone.
// Parameter: Polygons (die zu rendernden Polygone), Offset (der Index des 1. zu
// rendernden Vertex bzw. Index, beginnend bei 0), PolyCnt (die Anzahl der zu rendernden Polygone)
//-----------------------------------------------------------------------------
void Revo3dEngine::Render3dPolygons (Revo3dPolygons *Polygons, int Offset, int PolyCnt)
{
	if (!pD3DDevice || !Polygons)
		return;

	LPDIRECT3DTEXTURE8 UseTexture;

	//Schatten bercksichtigen
	pD3DDevice->SetVertexShader (VERTEX_FVF);
	if (!UseShadow)
	{
		SetTransform (&Polygons->matTrans, &Polygons->matRot);
		if (!Polygons->pTex)
			UseTexture = NULL;
		else
		{
			UseTexture = Polygons->pTex->pTexture;
			if (Polygons->pTex->MipLevels != 1)
				pD3DDevice->SetTextureStageState (0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
			else
				pD3DDevice->SetTextureStageState (0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
		}
		if (LastTexture != UseTexture)
		{
			pD3DDevice->SetTexture (0, UseTexture);
			LastTexture = UseTexture;
		}
		pD3DDevice->SetMaterial (&Polygons->Mat);
		LastMaterial = &Polygons->Mat;
		if (Polygons->Mat.Diffuse.a != 1 || (Polygons->pTex && Polygons->pTex->colKey) || (Polygons->pTex && Polygons->pTex->UseTransMap))
		{
			//Eine Blending-Operation ist erforderlich, daher Alpha Blending aktivieren
			pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, true);
			if (Polygons->Mat.Diffuse.a != 1 && Polygons->pTex && (Polygons->pTex->colKey || Polygons->pTex->UseTransMap))
				pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
			else if (Polygons->Mat.Diffuse.a != 1)
				pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
			else
				pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);
		}
	}
	else
		SetTransform (&Polygons->matTrans, &Polygons->matRot, &matShadow);
	if (Polygons->Visibility)
	{
		//Spiegelung bercksichtigen
		if (!UseMirror)				
			pD3DDevice->SetRenderState (D3DRS_CULLMODE, (D3DCULL) Polygons->Visibility);
		else
			pD3DDevice->SetRenderState (D3DRS_CULLMODE, InvertCullMode ((D3DCULL) Polygons->Visibility));
	}
	//Polygone rendern
	if (Polygons->Dynamic)
	{
		if (Polygons->Indexed && Polygons->pIData)
			pD3DDevice->DrawIndexedPrimitiveUP ((D3DPRIMITIVETYPE) Polygons->PolyFormat, 0, Polygons->VertexMaxCnt, PolyCnt, Polygons->pIData, D3DFMT_INDEX16, Polygons->pVData, sizeof (Vertex));
		else
			pD3DDevice->DrawPrimitiveUP ((D3DPRIMITIVETYPE) Polygons->PolyFormat, PolyCnt, Polygons->pVData, sizeof (Vertex));
	}
	else
	{
		pD3DDevice->SetStreamSource (0, Polygons->pVBuf, sizeof (Vertex));
		if (Polygons->Indexed && Polygons->pIBuf)
		{
			pD3DDevice->SetIndices (Polygons->pIBuf, 0);
			pD3DDevice->DrawIndexedPrimitive ((D3DPRIMITIVETYPE) Polygons->PolyFormat, 0, Polygons->VertexMaxCnt, Offset, PolyCnt);
		}
		else
			pD3DDevice->DrawPrimitive ((D3DPRIMITIVETYPE) Polygons->PolyFormat, Offset, PolyCnt);
	}
	if (!UseShadow)
		pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, false);
}



//-----------------------------------------------------------------------------
// Funktion: Render3dObject
// Beschreibung: Rendert ein 3d-Objekt.
// Parameter: Obj (das Objekt, das gerendert werden soll)
//-----------------------------------------------------------------------------
void Revo3dEngine::Render3dObject (Revo3dObject *Obj)
{
	if (!pD3DDevice || !Obj || !Obj->pMesh || !Obj->pTextures || !Obj->pMaterials)
		return;

	DWORD i;

	pD3DDevice->SetVertexShader (VERTEX_FVF);
	//Schatten bercksichtigen
	if (!UseShadow)
	{
		SetTransform (&Obj->matTrans, &Obj->matRot);
		pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
	}
	else
		SetTransform (&Obj->matTrans, &Obj->matRot, &matShadow);
	//Spiegelung bercksichtigen
	if (!UseMirror)
		pD3DDevice->SetRenderState (D3DRS_CULLMODE, D3DCULL_CCW);
	else
		pD3DDevice->SetRenderState (D3DRS_CULLMODE, D3DCULL_CW);
	//Objekt rendern
	pD3DDevice->SetTextureStageState (0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
	for (i = 0; i < Obj->MaterialCnt; i++)
	{
		if (!UseShadow)
		{
			pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, Obj->GetTransparency (i));
			if (LastTexture != Obj->pTextures[i].pTexture)
			{
				pD3DDevice->SetTexture (0, Obj->pTextures[i].pTexture);
				LastTexture = Obj->pTextures[i].pTexture;
			}
			if (LastMaterial != &Obj->pMaterials[i])
			{
				pD3DDevice->SetMaterial (&Obj->pMaterials[i]);
				LastMaterial = &Obj->pMaterials[i];
			}
		}
		Obj->pMesh->DrawSubset (i);
	}
	if (!UseShadow)
		pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, false);
}



//-----------------------------------------------------------------------------
// Funktion: SetShadowState
// Beschreibung: Legt fest, ob Schatten gerendert wird oder nicht.
// Parameter: PlaneTo (eine Revo3dPlane, die die Flche beschreibt, die den Schatten "auffangt"),
// LightPos (Position der Lichtquelle, von der der Schatten berechnet wird), Intensity (die Intensitt des Schattens),
// Type (der Typ der Lichtquelle), ClipEdges (ob der Schatten an den Ecken der Ebene aufhren soll)
// Rckgabewert: Der Rckgabewert sagt aus, ob Schatten jetzt aktiviert sind oder nicht.
// Bemerkung: Es sollte abgefragt werden, ob es mglich ist, transparente Schatten zu rendern.
// Dies kann mit der Funktion GetPossibleCaps getan werden.
// Schatten sollten als letztes in einem Frame gerendert werden.
// Der Rckgabewert sollte abgefragt werden, da bei false der Schatten nicht aktiviert wurde.
//-----------------------------------------------------------------------------
bool Revo3dEngine::SetShadowState (Revo3dPlane *PlaneTo, D3DXVECTOR3 *LightPos, int Intensity, LightType Type, bool ClipEdges)
{
	if (!pD3DDevice)
		return (false);
	if (!PlaneTo || !LightPos || !Intensity || !ShadowNoBlendAble || (Intensity != 255 && !ShadowAble))
	{
		//Schatten abstellen
		UseShadow = false;
		pD3DDevice->SetRenderState (D3DRS_STENCILENABLE, false);
		pD3DDevice->SetRenderState (D3DRS_CLIPPLANEENABLE, false);
		pD3DDevice->SetRenderState (D3DRS_ZBIAS, 0);
		pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, false);
		return (false);
	}

	D3DMATERIAL8 ShadowMat;
	D3DXVECTOR3 *pVecs;
	D3DXVECTOR4 Light;
	D3DXPLANE ClipPlane;

	//Schatten einstellen
	UseShadow = true;
	ShadowMat = MaterialMake (0, 0, 0, 0, 0, 0, 0, 255 - Intensity);
	//Textur, Material
	pD3DDevice->SetTexture (0, NULL);
	LastTexture = NULL;
	pD3DDevice->SetMaterial (&ShadowMat);
	LastMaterial = NULL;
	if (Intensity != 255)
	{
		//Blending
		pD3DDevice->SetTextureStageState (0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
		pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, true);
		pD3DDevice->SetRenderState (D3DRS_STENCILENABLE, true);
	}
	else
	{
		pD3DDevice->SetRenderState (D3DRS_ALPHABLENDENABLE, false);
		pD3DDevice->SetRenderState (D3DRS_STENCILENABLE, false);
	}
	//Lichtquelle "erstellen"
	memcpy (&Light, LightPos, sizeof (D3DXVECTOR3));
	if (Type == DIRECTIONALLIGHT)
		Light.w = 0;
	else
		Light.w = 1;	
	if (ClipEdges)
	{
		pD3DDevice->SetRenderState (D3DRS_CLIPPLANEENABLE, D3DCLIPPLANE0 | D3DCLIPPLANE1 | D3DCLIPPLANE2);
		pVecs = PlaneTo->Points;
		pD3DDevice->SetClipPlane (0, *D3DXPlaneFromPoints (&ClipPlane, &pVecs[1], &pVecs[0], &CamPos));
		pD3DDevice->SetClipPlane (1, *D3DXPlaneFromPoints (&ClipPlane, &pVecs[2], &pVecs[1], &CamPos));
		pD3DDevice->SetClipPlane (2, *D3DXPlaneFromPoints (&ClipPlane, &pVecs[0], &pVecs[2], &CamPos));
	}
	else
		pD3DDevice->SetRenderState (D3DRS_CLIPPLANEENABLE, false);
	pD3DDevice->SetRenderState (D3DRS_ZBIAS, 5);
	D3DXMatrixShadow (&matShadow, &Light, &PlaneTo->Plane);
	return (true);
}



//-----------------------------------------------------------------------------
// Funktion: SetReflectionState
// Beschreibung: Legt fest, ob Spiegelungen (Reflexionen) gerendert werden oder nicht.
// Parameter: PlaneTo (eine Revo3dPlane, die die Flche beschreibt, an der die Spiegelung auftritt),
// ClipEdges (ob die Spiegelung an den Ecken der Ebene aufhren soll)
// Bemerkung: Spiegelungen sollten gerendert werden, bevor die spiegelnde Flche (z.B ein Boden)
// selbst gerendert wird.
//-----------------------------------------------------------------------------
void Revo3dEngine::SetReflectionState (Revo3dPlane *PlaneTo, bool ClipEdges)
{
	if (!pD3DDevice)
		return;
	if (!PlaneTo)
	{
		//Spiegeln abstellen
		pD3DDevice->SetTransform (D3DTS_VIEW, &matViewSave);
		pD3DDevice->SetRenderState (D3DRS_CLIPPLANEENABLE, false);
		UseMirror = false;
		return;
	}

	D3DXMATRIX matReflect;
	D3DXVECTOR3 *pVecs, MirrorCamPos;
	D3DXPLANE ClipPlane;

	if (!UseMirror)
	{
		UseMirror = true;
		//Aktuelle View-Matrix in der Engine sichern
		pD3DDevice->GetTransform (D3DTS_VIEW, &matViewSave);
	}
	D3DXMatrixReflect (&matReflect, &PlaneTo->Plane);
	D3DXMatrixMultiply (&matReflect, &matReflect, &matViewSave);
	pD3DDevice->SetTransform (D3DTS_VIEW, &matReflect);
	if (ClipEdges)
	{
		pD3DDevice->SetRenderState (D3DRS_CLIPPLANEENABLE, D3DCLIPPLANE0 | D3DCLIPPLANE1 | D3DCLIPPLANE2);
		MirrorCamPos = MirrorPoint (PlaneTo, &CamPos);
		pVecs = PlaneTo->Points;
		pD3DDevice->SetClipPlane (0, *D3DXPlaneFromPoints (&ClipPlane, &pVecs[1], &MirrorCamPos, &pVecs[0]));
		pD3DDevice->SetClipPlane (1, *D3DXPlaneFromPoints (&ClipPlane, &pVecs[2], &MirrorCamPos, &pVecs[1]));
		pD3DDevice->SetClipPlane (2, *D3DXPlaneFromPoints (&ClipPlane, &pVecs[0], &MirrorCamPos, &pVecs[2]));
	}
}



//-----------------------------------------------------------------------------
// Funktion: SetLookDistance
// Beschreibung: Legt fest, wie weit die maximale Sicht begrenzt ist.
// Parameter: MaxView (die maximale Sichtweite), MaxAngle (der Winkel der Kamera)
//-----------------------------------------------------------------------------
void Revo3dEngine::SetLookDistance (float MaxView, float MaxAngle)
{
	if (!pD3DDevice)
		return;	

	D3DXMATRIX matProj;

	D3DXMatrixPerspectiveFovLH (&matProj, MaxAngle, 1, 1, MaxView);
	pD3DDevice->SetTransform (D3DTS_PROJECTION, &matProj);
}



//-----------------------------------------------------------------------------
// Funktion: SetCamera
// Beschreibung: Setzt die Kameraposition und -ausrichtung.
// Parameter: CameraPos (die Position der Kamera), CameraLookAt (wohin die
// Kamera schaut), RollAngle (die Drehung der Kamera, in Radiant)
//-----------------------------------------------------------------------------
void Revo3dEngine::SetCamera (D3DXVECTOR3 *CameraPos, D3DXVECTOR3 *CameraLookAt, float RollAngle)
{
	if (!pD3DDevice || !CameraPos || !CameraLookAt)
		return;

	D3DXMATRIX matView, matCalc;

	CamPos = *CameraPos;
	D3DXMatrixLookAtLH (&matView, CameraPos, CameraLookAt, &Vector3dMake (0, 1, 0));
	if (RollAngle)
	{
		D3DXMatrixRotationZ (&matCalc, RollAngle);
		D3DXMatrixMultiply (&matView, &matView, &matCalc);
	}
	pD3DDevice->SetTransform (D3DTS_VIEW, &matView);
}



//-----------------------------------------------------------------------------
// Funktion: SetLight
// Beschreibung: Fgt eine neue Lichtquelle hinzu oder ersetzt eine vorhandene
// Lichtquelle und schaltet dieses Lichtquelle ein.
// Parameter: Index (der Index der Lichtquelle, die gendert werden soll),
// Light (die Beschreibung der Lichtquelle)
//-----------------------------------------------------------------------------
bool Revo3dEngine::SetLight (int Index, D3DLIGHT8 *Light)
{
	if (!pD3DDevice || !Light)
		return (false);
	if (FAILED (pD3DDevice->SetLight (Index, Light)))
		return (false);
	pD3DDevice->LightEnable (Index, true);
	if (Index > MaxLightIndex)
		MaxLightIndex = Index;
	return (true);
}



//-----------------------------------------------------------------------------
// Funktion: SetLightState
// Beschreibung: Schaltet eine Lichtquelle aus oder ein.
// Parameter: Index (der Index der Lichtquelle, die gendert werden soll),
// State (der Status, den die Lichtquelle annehmen soll)
//-----------------------------------------------------------------------------
void Revo3dEngine::SetLightState (int Index, bool State)
{
	if (!pD3DDevice)
		return;
	pD3DDevice->LightEnable (Index, State);
}



//-----------------------------------------------------------------------------
// Funktion: SetSpecularState
// Beschreibung: Bestimmt, ob Specular-Effekte verwendet werden knnen.
// Standardmig auf falsch gestellt.
// Parameter: State (der Status)
//-----------------------------------------------------------------------------
void Revo3dEngine::SetSpecularState (bool State)
{
	if (!pD3DDevice)
		return;
	pD3DDevice->SetRenderState (D3DRS_SPECULARENABLE, State);
}



//-----------------------------------------------------------------------------
// Funktion: SetZState
// Beschreibung: Bestimmt, wie der Z-Buffer verwendet wird.
// Parameter: ReadEnable (ob vom Z-Buffer gelesen werden kann), WriteEnable (ob auf den Z-Buffer
// geschrieben werden kann)
// Bemerkung: Wird der Z-Buffer lesend deaktiviert, werden alle Polygone bzw. Objekte in den Vordergrund gerendert.
// Wenn der Z-Buffer lesend nicht verwendet werden kann, kann er auch nicht schreibend verwendet werden.
//-----------------------------------------------------------------------------
void Revo3dEngine::SetZState (bool ReadEnable, bool WriteEnable)
{
	if (!pD3DDevice)
		return;
	pD3DDevice->SetRenderState (D3DRS_ZENABLE, ReadEnable);
	if (ReadEnable)
		pD3DDevice->SetRenderState (D3DRS_ZWRITEENABLE, WriteEnable);
}



//-----------------------------------------------------------------------------
// Funktion: SetAnisotropicState
// Beschreibung: Bestimmt, ob der anisoptische Texturenfilter verwendet wird.
// Standardmig auf falsch gestellt.
// Parameter: State (der Status)
//-----------------------------------------------------------------------------
void Revo3dEngine::SetAnisotropicState (bool State)
{
	if (!pD3DDevice)
		return;
	if (State)
	{
		pD3DDevice->SetTextureStageState (0, D3DTSS_MAGFILTER, D3DTEXF_ANISOTROPIC);
		pD3DDevice->SetTextureStageState (0, D3DTSS_MINFILTER, D3DTEXF_ANISOTROPIC);
	}
	else
	{
		pD3DDevice->SetTextureStageState (0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
		pD3DDevice->SetTextureStageState (0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	}
}



//-----------------------------------------------------------------------------
// Funktion: SetGlobalLight
// Beschreibung: Setzt ein Licht, das immer mindestens vorhanden ist.
// Parameter: Color (die Farbe des Lichtes)
//-----------------------------------------------------------------------------
void Revo3dEngine::SetGlobalLight (D3DCOLOR Color)
{
	if (!pD3DDevice)
		return;
	pD3DDevice->SetRenderState (D3DRS_AMBIENT, Color);
}



//-----------------------------------------------------------------------------
// Funktion: SetFog
// Beschreibung: Bestimmt die Intensitt des Nebels.
// Parameter: Density (die Intensitt, von 0 bis 255)
//-----------------------------------------------------------------------------
void Revo3dEngine::SetFog (int Density)
{
	float d;

	if (!pD3DDevice)
		return;
	if (Density)
		pD3DDevice->SetRenderState (D3DRS_FOGENABLE, true);
	else
	{
		pD3DDevice->SetRenderState (D3DRS_FOGENABLE, false);
		return;
	}
	d = (float) Density / 2550;
	pD3DDevice->SetRenderState (D3DRS_FOGDENSITY, *(DWORD *) &d);
}



//********** ALLGEMEIN **********

//-----------------------------------------------------------------------------
// Funktion: HandleMessages
// Beschreibung: Verarbeitet die Messages von Windows fr das Fenster.
//-----------------------------------------------------------------------------
void Revo3dEngine::HandleMessages ()
{
	MSG msg;

	while (PeekMessage (&msg, hWnd, 0, 0, PM_REMOVE))
	{
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}
}



//-----------------------------------------------------------------------------
// Funktion: StartTimer
// Beschreibung: Startet den Timer.
//-----------------------------------------------------------------------------
void Revo3dEngine::StartTimer ()
{
	QueryPerformanceFrequency (&TimerFrq);
	QueryPerformanceCounter (&TimerBase);
	TimerLast = 0;
}



//-----------------------------------------------------------------------------
// Funktion: GetRelTimeDiff
// Beschreibung: Gibt die Zeit in Sekunden zurck, die seit dem letzten Aufruf
// dieser Funktion bzw. dem Aufruf der StartTimer-Funktion vergangen ist.
//-----------------------------------------------------------------------------
float Revo3dEngine::GetRelTimeDiff ()
{
	LARGE_INTEGER TimerCnt;
	float TimerTotal, TimerDiff;

	QueryPerformanceCounter (&TimerCnt);
	TimerTotal = (float) (TimerCnt.QuadPart - TimerBase.QuadPart) / TimerFrq.QuadPart;
	TimerDiff = TimerTotal - TimerLast;
	TimerLast = TimerTotal;
	return (TimerDiff);
}



//-----------------------------------------------------------------------------
// Funktion: GetAbsTimeDiff
// Beschreibung: Gibt die Zeit in Sekunden zurck, die seit dem Aufruf der
// StartTimer-Funktion vergangen ist.
//-----------------------------------------------------------------------------
float Revo3dEngine::GetAbsTimeDiff ()
{
	LARGE_INTEGER TimerCnt;

	QueryPerformanceCounter (&TimerCnt);
	return ((float) (TimerCnt.QuadPart - TimerBase.QuadPart) / TimerFrq.QuadPart);
}



//******************** PRIVATE FUNKTIONEN ********************

bool Revo3dEngine::IsUsualRes (int ResX, int ResY)
{
	if (ResX == 320 && ResY == 240)
		return (true);
	else if (ResX == 640 && ResY == 480)
		return (true);
	else if (ResX == 800 && ResY == 600)
		return (true);
	else if (ResX == 1024 && ResY == 768)
		return (true);
	else if (ResX == 1280 && ResY == 1024)
		return (true);
	else if (ResX == 1600 && ResY == 1200)
		return (true);
	else if (ResX == 1920 && ResY == 1440)
		return (true);
	else if (ResX == 2048 && ResY == 1536)
		return (true);
	else
		return (false);
}



bool Revo3dEngine::IsRenderState (int State)
{
	if (State >=7 && State <= 10)
		return (true);
	if (State >=14 && State <= 16)
		return (true);
	if (State == 19 || State == 20 || State == 40 || State == 47 || State == 48)
		return (true);
	if (State >= 22 && State <= 30)
		return (true);
	if (State >= 34 && State <= 38)
		return (true);
	if (State == 40)
		return (true);
	if (State >= 52 && State <= 60)
		return (true);
	if (State >= 128 && State <= 137)
		return (true);
	if (State >= 139 && State <= 143)
		return (true);
	if (State >= 145 && State <= 148)
		return (true);
	if (State >= 151 && State <= 168)
		return (true);
	if (State == 170 || State == 171)
		return (true);
	return (false);
}



void Revo3dEngine::SetTransform (D3DXMATRIX *matTrans, D3DXMATRIX *matRot, D3DXMATRIX *matMultiply)
{
	if (!pD3DDevice || !matTrans || !matRot)
		return;

	D3DXMATRIX matWorld;

	D3DXMatrixMultiply (&matWorld, matRot, matTrans);
	if (matMultiply)
		D3DXMatrixMultiply (&matWorld, &matWorld, matMultiply);
	pD3DDevice->SetTransform (D3DTS_WORLD, &matWorld);
}



D3DCULL Revo3dEngine::InvertCullMode (D3DCULL Mode)
{
	if (Mode == D3DCULL_CW)
		return (D3DCULL_CCW);
	else if (Mode == D3DCULL_CCW)
		return (D3DCULL_CW);
	else
		return (D3DCULL_NONE);
}



void Revo3dEngine::RegisterObject (void *pObj, int ObjType)
{
	if (RegisterObj)
		RegisterObj = (RegObj *) realloc (RegisterObj, sizeof (RegObj) * (RegObjCnt + 1));
	else
		RegisterObj = (RegObj *) malloc (sizeof (RegObj) * (RegObjCnt + 1));
	RegisterObj[RegObjCnt].pObj = pObj;
	RegisterObj[RegObjCnt].ObjType = ObjType;
	RegObjCnt++;
}

void Revo3dEngine::UnregisterObject (void *pObj)
{
	if (!RegisterObj)
		return;

	int i, j;

	for (i = 0; i < RegObjCnt; i++)
	{
		if (RegisterObj[i].pObj == pObj)
		{
			if (pD3DDevice && (RegisterObj[i].ObjType == OBJ_2DTEXTURE || RegisterObj[i].ObjType == OBJ_3DOBJ))
			{
				pD3DDevice->SetTexture (0, NULL);
				LastTexture = NULL;
			}
			for (j = i; j < RegObjCnt - 1; j++)
				RegisterObj[j] = RegisterObj[j + 1];
			RegisterObj = (RegObj *) realloc (RegisterObj, sizeof (RegObj) * (RegObjCnt - 1));
			RegObjCnt--;
			return;
		}
	}
}

void Revo3dEngine::CleanupObjects ()
{
	if (!RegisterObj || ObjCleaned)
		return;

	int i;
	Revo2dTexture *pTex;
	Revo2dText *pText;
	Revo3dObject *pObj;
	Revo3dPolygons *pPolys;

	//Alle erstellten Objekte deallokieren
	for (i = 0; i < RegObjCnt; i++)
	{
		if (RegisterObj[i].ObjType == OBJ_2DTEXTURE)
		{
			pTex = (Revo2dTexture *) RegisterObj[i].pObj;
			pTex->StoreData ();
			pTex->Cleanup ();
		}
		else if (RegisterObj[i].ObjType == OBJ_2DTEXT)
		{
			pText = (Revo2dText *) RegisterObj[i].pObj;
			pText->Cleanup ();
		}
		else if (RegisterObj[i].ObjType == OBJ_3DOBJ)
		{
			pObj = (Revo3dObject *) RegisterObj[i].pObj;
			pObj->StoreData ();
			//Nicht die Cleanup-Funktion verwenden, da diese auch die Texturen und Materialien lschen wrde!
			if (pObj->pMesh)
			{
				pObj->pMesh->Release ();
				pObj->pMesh = NULL;
			}
		}
		else if (RegisterObj[i].ObjType == OBJ_3DPOLYS)
		{
			pPolys = (Revo3dPolygons *) RegisterObj[i].pObj;
			pPolys->StoreData ();
			if (!pPolys->Dynamic)
				//Nur nicht-dynamische Polygone "cleanen".
				pPolys->Cleanup ();
		}
	}
	ObjCleaned = true;
}

void Revo3dEngine::RestoreObjects ()
{
	if (!RegisterObj || !ObjCleaned)
		return;
	
	int i;
	Revo2dTexture *pTex;
	Revo2dText *pText;
	Revo3dObject *pObj;
	Revo3dPolygons *pPolys;

	//Alle deallokieren Objekte wiederherstellen
	for (i = 0; i < RegObjCnt; i++)
	{
		if (RegisterObj[i].ObjType == OBJ_2DTEXTURE)
		{
			pTex = (Revo2dTexture *) RegisterObj[i].pObj;
			pTex->ReStoreData ();
		}
		else if (RegisterObj[i].ObjType == OBJ_2DTEXT)
		{
			pText = (Revo2dText *) RegisterObj[i].pObj;
			pText->SetFont (pText->FontHandle);
		}
		else if (RegisterObj[i].ObjType == OBJ_3DOBJ)
		{
			pObj = (Revo3dObject *) RegisterObj[i].pObj;
			pObj->ReStoreData ();
		}
		else if (RegisterObj[i].ObjType == OBJ_3DPOLYS)
		{
			pPolys = (Revo3dPolygons *) RegisterObj[i].pObj;
			pPolys->ReStoreData ();
		}
	}
	ObjCleaned = false;
}



LRESULT CALLBACK WindowProc (HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	return (DefWindowProc (hWnd, msg, wParam, lParam));
}




















Revo2dTexture::Revo2dTexture ()
{
	pTexture = NULL;
	LockRect.pBits = NULL;
	pData = NULL;
	pEngine = NULL;
}

Revo2dTexture &Revo2dTexture::operator= (Revo2dTexture &pOrig)
{
	if (&pOrig == this || !pOrig.pEngine || !pOrig.pTexture || pOrig.LockRect.pBits)
		return (*this);

	DWORD i;
	D3DSURFACE_DESC MipMapDesc;
	D3DLOCKED_RECT SrcRect, DstRect;

	Cleanup ();
	pEngine = pOrig.pEngine;
	pDevice = pEngine->pD3DDevice;
	//Eigenschaften der originalen Textur auslesen
	pOrig.pTexture->GetLevelDesc (0, &TexDesc);
	//Textur "lokal" neu erstellen
	if (FAILED (D3DXCreateTexture (pDevice, TexDesc.Width, TexDesc.Height, pOrig.MipLevels, 0, TexDesc.Format, D3DPOOL_MANAGED, &pTexture)))
		return (*this);
	//Textur kopieren
	for (i = 0; i < pTexture->GetLevelCount (); i++)
	{
		pTexture->GetLevelDesc (i, &MipMapDesc);
		//Surface kopieren
		pOrig.pTexture->LockRect (i, &SrcRect, NULL, D3DLOCK_READONLY);
		pTexture->LockRect (i, &DstRect, NULL, 0);
		memcpy (DstRect.pBits, SrcRect.pBits, MipMapDesc.Size);
		pTexture->UnlockRect (i);
		pOrig.pTexture->UnlockRect (i);
	}
	colKey = pOrig.colKey;
	MipLevels = pOrig.MipLevels;
	UseTransMap = pOrig.UseTransMap;

	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_2DTEXTURE);
	return (*this);
}

Revo2dTexture::~Revo2dTexture ()
{
	ReStoreData ();
	Cleanup ();
	if (pEngine)
		pEngine->UnregisterObject (this);
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert die 2d-Textur.
// Parameter: Eng (die Hauptengine)
//-----------------------------------------------------------------------------
void Revo2dTexture::Initialize (Revo3dEngine *Eng)
{
	if (!Eng)
		return;

	Cleanup ();
	pEngine = Eng;
	pDevice = pEngine->pD3DDevice;
	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_2DTEXTURE);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert die Revo2dTexture.
//-----------------------------------------------------------------------------
void Revo2dTexture::Cleanup ()
{
	EditEnd ();
	if (pTexture)
	{
		pTexture->Release ();
		pTexture = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: LoadFromFile
// Beschreibung: Ldt ein Bitmap von einer Datei in die Textur.
// Parameter: File (Pfad der Bitmap-Datei), ColorKey (transparente Farbe),
// MipLevels (Wie viele MipMapping-Ebenen erstellt werden sollen. 0, um eine
// komplette MipMap zu erstellen.), EnableEdit (Ob die Textur nachtrglich gendert
// werden kann. Dies ist nur mglich wenn MipLevels == 1.)
//-----------------------------------------------------------------------------
bool Revo2dTexture::LoadFromFile (char *File, D3DCOLOR ColorKey, int MipLevels, bool EnableEdit)
{
	if (!File)
		return (false);

	Cleanup ();
	pTexture = TextureMake (pEngine, File, ColorKey, MipLevels, EnableEdit);
	if (!pTexture)
		return (false);

	pTexture->GetLevelDesc (0, &TexDesc);
	colKey = ColorKey;
	this->MipLevels = MipLevels;
	UseTransMap = false;

	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: LoadForEdit
// Beschreibung: Erstellt eine leere Textur zum Editieren.
// Parameter: x (Breite der Textur), y (Hhe der Textur)
// Bemerkung: x und y sollten beide eine 2er-Potenz grer 0 sein.
//-----------------------------------------------------------------------------
bool Revo2dTexture::LoadForEdit (int x, int y)
{
	Cleanup ();
	if (FAILED (D3DXCreateTexture (pDevice, x, y, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &pTexture)))
		return (false);
	pTexture->GetLevelDesc (0, &TexDesc);
	colKey = 0;
	MipLevels = 1;
	UseTransMap = false;
	return (true);
}



//-----------------------------------------------------------------------------
// Funktion: EditStart
// Beschreibung: Muss zu Beginn einer nderungs-Operation der Textur aufgerufen werden.
//-----------------------------------------------------------------------------
bool Revo2dTexture::EditStart ()
{	
	if (!pTexture || MipLevels != 1 || LockRect.pBits)
		return (false);

	//Pixel-Format berprfen (D3DFMT_A8R8G8B8 bentigt)
	if (TexDesc.Format != D3DFMT_A8R8G8B8)
		return (false);
	//Gesamte Textur locken
	LockRect.pBits = NULL;
	if (FAILED (pTexture->LockRect (0, &LockRect, NULL, 0)))
		return (false);
	//EndOfLine-Offset ausrechnen
	eolOffset = LockRect.Pitch - TexDesc.Width * 4;
	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: Edit
// Beschreibung: ndert ein einzelnes Texel auf der Textur.
// Parameter: Pos (Die x- und y-Position des zu ndernden Texels. Bei z.B. einer
// 128x128 Textur reicht x und y jeweils von 0 bis 127.), a (der Transparenz-Anteil),
// r, g, b (die Farbanteile). Bei -1 wird die jeweilige Komponente nicht gendert.
//-----------------------------------------------------------------------------
void Revo2dTexture::Edit (D3DXVECTOR2 *Pos, int a, int r, int g, int b)
{
	if (!pTexture || !LockRect.pBits || !Pos)
		return;
	if (Pos->x < 0 || Pos->y < 0 || Pos->x >= TexDesc.Width || Pos->y >= TexDesc.Height)
		return;

	BYTE *pTexData, Data;

	pTexData = (BYTE *) LockRect.pBits + (int) Pos->y * LockRect.Pitch + (int) Pos->x * 4;
	//1. Byte: b
	Data = b;
	if (b != -1)
		memcpy (pTexData, &Data, 1);
	pTexData++;
	//2. Byte: g
	Data = g;
	if (g != -1)
		memcpy (pTexData, &Data, 1);
	pTexData++;
	//3. Byte: r
	Data = r;
	if (r != -1)
		memcpy (pTexData, &Data, 1);
	pTexData++;
	//4. Byte: a
	Data = a;
	if (a != -1)
	{
		UseTransMap = true;
		memcpy (pTexData, &Data, 1);
	}
}

//-----------------------------------------------------------------------------
// Funktion: EditEnd
// Beschreibung: Muss beim Ende einer nderungs-Operation der Textur aufgerufen werden.
//-----------------------------------------------------------------------------
void Revo2dTexture::EditEnd ()
{
	if (!pTexture || !LockRect.pBits)
		return;

	pTexture->UnlockRect (0);
	LockRect.pBits = NULL;
}



void Revo2dTexture::StoreData ()
{
	if (!pTexture || pData)
		return;

	DWORD i, MipMapCnt = pTexture->GetLevelCount ();
	D3DSURFACE_DESC MipMapDesc;
	D3DLOCKED_RECT LRect;

	EditEnd ();
	pData = new BYTE *[MipMapCnt];
	for (i = 0; i < MipMapCnt; i++)
	{
		//Textur im Speicher sichern
		pTexture->GetLevelDesc (i, &MipMapDesc);
		pData[i] = new BYTE[MipMapDesc.Size];
		pTexture->LockRect (i, &LRect, NULL, D3DLOCK_READONLY);
		memcpy (pData[i], LRect.pBits, MipMapDesc.Size);
		pTexture->UnlockRect (i);
	}
}

void Revo2dTexture::ReStoreData ()
{
	if (!pData)
		return;

	DWORD i, MipMapCnt;
	D3DSURFACE_DESC MipMapDesc;
	D3DLOCKED_RECT LRect;	

	if (FAILED (D3DXCreateTexture (pDevice, TexDesc.Width, TexDesc.Height, MipLevels, 0, TexDesc.Format, D3DPOOL_MANAGED, &pTexture)))
		return;
	MipMapCnt = pTexture->GetLevelCount ();
	//Textur im Speicher wiederherstellen
	for (i = 0; i < MipMapCnt; i++)
	{
		pTexture->GetLevelDesc (i, &MipMapDesc);
		pTexture->LockRect (i, &LRect, NULL, 0);
		memcpy (LRect.pBits, pData[i], MipMapDesc.Size);
		pTexture->UnlockRect (i);
		delete []pData[i];
	}
	delete []pData;
	pData = NULL;
}



















Revo2dPicture::Revo2dPicture ()
{
	D3DCOLOR col;

	alpha = 255;
	red = 255;
	green = 255;
	blue = 255;
	col = ColorMake (255, 255, 255);
	Vert[0] = TLVertexMake (0, 0, col, 0, 0);
	Vert[1] = TLVertexMake (0, 0, col, 1, 0);
	Vert[2] = TLVertexMake (0, 0, col, 0, 1);
	Vert[3] = TLVertexMake (0, 0, col, 1, 1);
	MP = Vector2dMake (0, 0);
	RotAngle = 0;
}



//-----------------------------------------------------------------------------
// Funktion: SetPosition
// Beschreibung: Setzt die Position des Bildes am Bildschirm.
// Parameter: x1, y1 (die linke, obere Ecke), x2, y2 (die rechte, untere Ecke)
// Bemerkung: Der Aufruf dieser Funktion macht vorhergehende Rotationen "rckgngig".
//-----------------------------------------------------------------------------
void Revo2dPicture::SetPosition (int x1, int y1, int x2, int y2)
{
	Vert[0].x = (float) x1;
	Vert[0].y = (float) y1;
	Vert[1].x = (float) x2;
	Vert[1].y = (float) y1;
	Vert[2].x = (float) x1;
	Vert[2].y = (float) y2;
	Vert[3].x = (float) x2;
	Vert[3].y = (float) y2;
	memcpy (NoRotVert, Vert, 4 * sizeof (TLVertex));
	MP = Vector2dMake ((x1 + x2) / 2.0f, (y1 + y2) / 2.0f);
}

//-----------------------------------------------------------------------------
// Funktion: Move
// Beschreibung: Verschiebt die Position des Bildes.
// Parameter: MoveX, MoveY (die Verschiebung, relativ zum aktuellen Standpunkt)
//-----------------------------------------------------------------------------
void Revo2dPicture::Move (int MoveX, int MoveY)
{
	int i;

	for (i = 0; i < 4; i++)
	{
		Vert[i].x += MoveX;
		Vert[i].y += MoveY;
	}
}



//-----------------------------------------------------------------------------
// Funktion: Rotate
// Beschreibung: Rotiert das Bild um seinen Mittelpunkt, im Uhrzeigersinn.
// Parameter: Angle (der Winkel)
//-----------------------------------------------------------------------------
void Revo2dPicture::Rotate (float Angle)
{
	int i;
	float cosPhi, sinPhi;

	RotAngle += Angle;
	cosPhi = (float) cos (RotAngle);
	sinPhi = (float) sin (RotAngle);
	for (i = 0; i < 4; i++)
	{
		Vert[i].x = (NoRotVert[i].x - MP.x) * cosPhi - (NoRotVert[i].y - MP.y) * sinPhi + MP.x;
		Vert[i].y = (NoRotVert[i].x - MP.x) * sinPhi + (NoRotVert[i].y - MP.y) * cosPhi + MP.y;
	}
}



//-----------------------------------------------------------------------------
// Funktion: SetPictureRange
// Beschreibung: Legt fest, welcher Ausschnitt des Bildes dargestellt wird.
// Parameter: x1, y1 (die linke, obere Ecke), x2, y2 (die rechte, untere Ecke)
// Bemerkung: Alle Parameter sollten zwischen 0 und 1 sein.
//-----------------------------------------------------------------------------
void Revo2dPicture::SetPictureRange (float x1, float y1, float x2, float y2)
{
	Vert[0].tu = x1;
	Vert[0].tv = y1;
	Vert[1].tu = x2;
	Vert[1].tv = y1;
	Vert[2].tu = x1;
	Vert[2].tv = y2;
	Vert[3].tu = x2;
	Vert[3].tv = y2;
}



//-----------------------------------------------------------------------------
// Funktion: SetBackcolor
// Beschreibung: Setzt die Hintergrundfarbe (Beleuchtung) fr das Bild.
// Parameter: r, g, b (jeweils von 0 bis 255)
//-----------------------------------------------------------------------------
void Revo2dPicture::SetBackcolor (int r, int g, int b)
{
	int i;
	D3DCOLOR col;

	red = r;
	green = g;
	blue = b;
	col = ColorAlphaMake (alpha, red, green, blue);
	for (i = 0; i < 4; i++)
		Vert[i].Color = col;
}



//-----------------------------------------------------------------------------
// Funktion: SetTransparency
// Beschreibung: Setzt die Transparenz fr das Bild.
// Parameter: Trans (Transparenz, von 0 bis 255)
//-----------------------------------------------------------------------------
void Revo2dPicture::SetTransparency (int Transparency)
{
	int i;
	D3DCOLOR col;

	alpha = 255 - Transparency;
	col = ColorAlphaMake (alpha, red, green, blue);
	for (i = 0; i < 4; i++)
		Vert[i].Color = col;
}




















Revo2dText::Revo2dText ()
{
	pFont = NULL;
	pEngine = NULL;
}

Revo2dText &Revo2dText::operator= (Revo2dText &pOrig)
{
	if (&pOrig == this || !pOrig.pEngine || !pOrig.pFont)
		return (*this);

	Cleanup ();
	pEngine = pOrig.pEngine;
	pDevice = pEngine->pD3DDevice;
	Col = pOrig.Col;
	For = pOrig.For;
	Rect = pOrig.Rect;
	strcpy (Str, pOrig.Str);
	SetFont (pOrig.FontHandle);

	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_2DTEXT);
	return (*this);
}

Revo2dText::~Revo2dText ()
{
	Cleanup ();
	if (pEngine)
		pEngine->UnregisterObject (this);
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert den 2d-Text.
// Parameter: Eng (die Hauptengine)
// Bemerkung: Setzt als Default-Eigenschaften die Textfarbe auf schwarz und
// das Format auf links oben.
//-----------------------------------------------------------------------------
void Revo2dText::Initialize (Revo3dEngine *Eng)
{
	if (!Eng)
		return;

	Cleanup ();
	pEngine = Eng;
	pDevice = pEngine->pD3DDevice;
	For = DT_LEFT | DT_TOP;
	Col = ColorMake (0, 0, 0);
	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_2DTEXT);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert den Revo2dText.
//-----------------------------------------------------------------------------
void Revo2dText::Cleanup ()
{
	if (pFont)
	{
		pFont->Release ();
		pFont = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: SetFont
// Beschreibung: Setzt eine Schriftart fr den Text.
// Parameter: Font (die Schriftart, erstellbar mit FontMake)
//-----------------------------------------------------------------------------
void Revo2dText::SetFont (HFONT Font)
{
	Cleanup ();
	D3DXCreateFont (pDevice, Font, &pFont);
	FontHandle = Font;
}



//-----------------------------------------------------------------------------
// Funktion: SetPosition
// Beschreibung: Setzt die Position fr den Text.
// Parameter: x1, y1 (die linke, obere Ecke), x2, y2 (die rechte, untere Ecke)
//-----------------------------------------------------------------------------
void Revo2dText::SetPosition (int x1, int y1, int x2, int y2)
{
	Rect.left = x1;
	Rect.top = y1;
	Rect.right = x2;
	Rect.bottom = y2;
}



//-----------------------------------------------------------------------------
// Funktion: SetFormat
// Beschreibung: Setzt die Formatierung fr den Text.
// Parameter: FormatH (links = 0, zentriert = 1, rechts = 2),
// FormatV (oben = 0, zentriert = 1, unten = 2)
//-----------------------------------------------------------------------------
void Revo2dText::SetFormat (int FormatH, int FormatV)
{
	For = 0;
	if (FormatH == 0)
		For = DT_LEFT;
	else if (FormatH == 1)
		For = For | DT_CENTER;
	else
		For = For | DT_RIGHT;
	if (FormatV == 0)
		For = For | DT_TOP;
	else if (FormatV == 1)
		For = For | DT_VCENTER;
	else
		For = For | DT_BOTTOM;
}



















Revo3dObject::Revo3dObject ()
{
	pMesh = NULL;
	pTextures = NULL;
	pMaterials = NULL;
	pVData = pIData = pAData = NULL;
	pEngine = NULL;
}

Revo3dObject &Revo3dObject::operator= (Revo3dObject &pOrig)
{
	if (&pOrig == this || !pOrig.pEngine || !pOrig.pMesh || !pOrig.pTextures || !pOrig.pMaterials)
		return (*this);

	DWORD i;

	Cleanup ();
	pEngine = pOrig.pEngine;
	pDevice = pEngine->pD3DDevice;
	//Mesh kopieren
	if (FAILED (pOrig.pMesh->CloneMeshFVF (D3DXMESH_MANAGED, pOrig.pMesh->GetFVF (), pDevice, &pMesh)))
		return (*this);
	//Texturen kopieren, mit dem Zuweisungsoperator der Revo2dTexture
	pTextures = new Revo2dTexture[pOrig.MaterialCnt];
	for (i = 0; i < pOrig.MaterialCnt; i++)
		pTextures[i] = pOrig.pTextures[i];
	//Materialien kopieren
	pMaterials = new D3DMATERIAL8[pOrig.MaterialCnt];
	memcpy (pMaterials, pOrig.pMaterials, sizeof (D3DMATERIAL8) * pOrig.MaterialCnt);
	//Sonstige Variablen kopieren
	FVF = pOrig.FVF;
	MaterialCnt = pOrig.MaterialCnt;
	Position = pOrig.Position;
	matTrans = pOrig.matTrans;
	matRot = pOrig.matRot;
	VertexCnt = pOrig.VertexCnt;
	FaceCnt = pOrig.FaceCnt;
	VertexSize = pOrig.VertexSize;

	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_3DOBJ);
	return (*this);
}

Revo3dObject::~Revo3dObject ()
{
	Cleanup ();
	if (pVData)
		delete []pVData;
	if (pIData)
		delete []pIData;
	if (pAData)
		delete []pAData;
	if (pEngine)
		pEngine->UnregisterObject (this);
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert das 3d-Object.
// Parameter: Eng (die Hauptengine)
// Bemerkung: Setzt als Default-Eigenschaft die Position auf (0, 0, 0) und keine Rotation.
//-----------------------------------------------------------------------------
void Revo3dObject::Initialize (Revo3dEngine *Eng)
{
	if (!Eng)
		return;

	Cleanup ();
	pEngine = Eng;
	pDevice = pEngine->pD3DDevice;
	Position = Vector3dMake (0, 0, 0);
	D3DXMatrixIdentity (&matTrans);
	D3DXMatrixIdentity (&matRot);
	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_3DOBJ);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert das Revo3dObject.
//-----------------------------------------------------------------------------
void Revo3dObject::Cleanup ()
{
	if (pTextures)
	{
		delete []pTextures;
		pTextures = NULL;
	}
	if (pMaterials)
	{
		delete []pMaterials;
		pMaterials = NULL;
	}
	if (pMesh)
	{
		pMesh->Release ();
		pMesh = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: LoadFromFile
// Beschreibung: Ldt ein 3d-Objekt von einer x-Datei.
// Parameter: File (Pfad der Datei), MipLevels (Wie viele MipMapping-Ebenen fr jede Textur
// auf dem 3d-Objekt erstellt werden sollen. 0, um eine komplette MipMap zu erstellen.)
// Bemerkung: Setzt als Default-Eigenschaften die Transparenz auf 0 und die
// SpecPower ebenfalls auf 0.
//-----------------------------------------------------------------------------
bool Revo3dObject::LoadFromFile (char *File, int MipLevels)
{
	if (!File)
		return (false);

	DWORD i, *pAdj;
	LPD3DXBUFFER pMatBuffer;
	D3DXMATERIAL *pRdMat;
	char Folder[1024], OPstr[1024];

	Cleanup ();
	if (FAILED (D3DXLoadMeshFromX (File, D3DXMESH_MANAGED, pDevice, NULL, &pMatBuffer, &MaterialCnt, &pMesh)))
		return (false);

	if (!MaterialCnt)
		return (false);
	pRdMat = (D3DXMATERIAL *) pMatBuffer->GetBufferPointer ();
	pMaterials = new D3DMATERIAL8[MaterialCnt];
	pTextures = new Revo2dTexture[MaterialCnt];
	GetFolder (Folder, File);

	for (i = 0; i < MaterialCnt; i++)
	{
		//Material erstellen
		pMaterials[i] = pRdMat[i].MatD3D;
		pMaterials[i].Ambient = pMaterials[i].Diffuse;
		pMaterials[i].Specular.a = 1;
		pMaterials[i].Specular.r = 1;
		pMaterials[i].Specular.g = 1;
		pMaterials[i].Specular.b = 1;
		//Falls ntig, Textur erstellen
		pTextures[i].Initialize (pEngine);
		if (pRdMat[i].pTextureFilename)
		{
			strcpy (OPstr, Folder);
			strcat (OPstr, pRdMat[i].pTextureFilename);
			pTextures[i].LoadFromFile (OPstr, 0, MipLevels, false);
		}
	}

	pMatBuffer->Release ();
	//Objekt optimieren
	pAdj = new DWORD[3 * pMesh->GetNumFaces ()];
	pMesh->GenerateAdjacency (0, pAdj);
	pMesh->OptimizeInplace (D3DXMESHOPT_ATTRSORT | D3DXMESHOPT_COMPACT | D3DXMESHOPT_VERTEXCACHE, pAdj, NULL, NULL, NULL);
	delete []pAdj;
	//Eigenschaften lokal in der Klasse sichern
	FVF = pMesh->GetFVF ();
	VertexSize = D3DXGetFVFVertexSize (FVF);
	VertexCnt = pMesh->GetNumVertices ();
	FaceCnt = pMesh->GetNumFaces ();
	//Normalen neu berechnen
	D3DXComputeNormals (pMesh);

	SetTransparency (0);
	SetSpecPower (0);

	return (true);
}



//-----------------------------------------------------------------------------
// Funktion: SetPosition
// Beschreibung: Setzt die Position des Objektes im 3d-Raum.
// Parameter: Pos (die zu setzende Position)
//-----------------------------------------------------------------------------
void Revo3dObject::SetPosition (D3DXVECTOR3 *Pos)
{
	if (!Pos)
		return;
	Position = *Pos;
	D3DXMatrixTranslation (&matTrans, Position.x, Position.y, Position.z);
}

//-----------------------------------------------------------------------------
// Funktion: Move
// Beschreibung: Verschiebt die Position des Objektes im 3d-Raum.
// Parameter: MoveDelta (die zu verschiebende Strecke)
//-----------------------------------------------------------------------------
void Revo3dObject::Move (D3DXVECTOR3 *MoveDelta)
{
	if (!MoveDelta)
		return;
	D3DXVec3Add (&Position, &Position, MoveDelta);
	D3DXMatrixTranslation (&matTrans, Position.x, Position.y, Position.z);
}



//-----------------------------------------------------------------------------
// Funktion: SetRotationMatrix
// Beschreibung: Setzt die Rotation des Objektes neu anhand einer Rotations-Matrix.
// Parameter: Matrix (die Rotations-Matrix)
//-----------------------------------------------------------------------------
void Revo3dObject::SetRotationMatrix (D3DXMATRIX *Matrix)
{
	if (!Matrix)
		return;
	matRot = *Matrix;
}

//-----------------------------------------------------------------------------
// Funktion: SetRotation
// Beschreibung: Setzt die Rotation des Objektes neu, ausgehend vom originalen, nicht
// rotierten Objekt. Das heit, dass beim Angle-Parameter bisherige Rotationen dieses
// Objektes nicht bercksichtig werden mssen.
// Parameter: Angle (Winkel fr die Rotation um die x, y und z-Achse), Order (Die
// Reihenfolge, in der die Rotationen durchgefhrt werden sollen, beginnend bei 0.
// Wenn z.B. zuerst in y, dann in x und anschlieend in z rotiert werden soll, muss
// der Vektor so aussehen: x = 1, y = 0, z = 2. Wenn NULL bergeben wird, wird
// zuerst um x, dann um y und anschlieend um z rotiert.), Origin (der Mittelpunkt
// der Rotation)
//-----------------------------------------------------------------------------
void Revo3dObject::SetRotation (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin)
{
	if (!Angle || !Origin)
		return;

	D3DXMATRIX matTmp1, matTmp2;

	D3DXMatrixTranslation (&matTmp1, Position.x - Origin->x, Position.y - Origin->y, Position.z - Origin->z);
	matRot = RotationMatrixMake (Angle, Order);
	D3DXMatrixMultiply (&matTmp1, &matTmp1, &matRot);
	D3DXMatrixTranslation (&matTmp2, Origin->x - Position.x, Origin->y - Position.y, Origin->z - Position.z);
	D3DXMatrixMultiply (&matRot, &matTmp1, &matTmp2);
}

//-----------------------------------------------------------------------------
// Funktion: Rotate
// Beschreibung: Rotiert das Objekt, ausgehend vom aktuellen, mglicherweise bereits
// frher rotierten Objekt. Das heit, dass beim Angle-Parameter bisherige Rotationen dieses
// Objektes bercksichtig werden mssen.
// Parameter: Angle (Winkel fr die Rotation um die x, y und z-Achse), Order (Die
// Reihenfolge, in der die Rotationen durchgefhrt werden sollen, beginnend bei 0.
// Wenn z.B. zuerst in y, dann in x und anschlieend in z rotiert werden soll, muss
// der Vektor so aussehen: x = 1, y = 0, z = 2. Wenn NULL bergeben wird, wird
// zuerst um x, dann um y und anschlieend um z rotiert.), Origin (der Mittelpunkt
// der Rotation)
//-----------------------------------------------------------------------------
void Revo3dObject::Rotate (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin)
{
	if (!Angle || !Origin)
		return;

	D3DXMATRIX matTmp1, matTmp2;

	D3DXMatrixTranslation (&matTmp1, Position.x - Origin->x, Position.y - Origin->y, Position.z - Origin->z);
	matTmp2 = RotationMatrixMake (Angle, Order);
	D3DXMatrixMultiply (&matTmp1, &matTmp1, &matTmp2);
	D3DXMatrixTranslation (&matTmp2, Origin->x - Position.x, Origin->y - Position.y, Origin->z - Position.z);
	D3DXMatrixMultiply (&matTmp1, &matTmp1, &matTmp2);
	D3DXMatrixMultiply (&matRot, &matRot, &matTmp1);
}



//-----------------------------------------------------------------------------
// Funktion: SetTransparency
// Beschreibung: Legt die Transparenz fr das Objekt fest.
//-----------------------------------------------------------------------------
void Revo3dObject::SetTransparency (int Transparency)
{
	if (!pMaterials)
		return;

	DWORD i;

	for (i = 0; i < MaterialCnt; i++)
		pMaterials[i].Diffuse.a = 1 - Transparency / (float) 255.0;
}



//-----------------------------------------------------------------------------
// Funktion: SetSpecPower
// Beschreibung: Legt die Intensitt der Glanzlichteffekte fr das Objekt fest.
//-----------------------------------------------------------------------------
void Revo3dObject::SetSpecPower (float SpecPower)
{
	if (!pMaterials)
		return;

	DWORD i;

	for (i = 0; i < MaterialCnt; i++)
		pMaterials[i].Power = SpecPower;
}



void Revo3dObject::StoreData ()
{
	if (!pMesh || pVData || pIData || pAData)
		return;

	BYTE *pLockData;

	//Vertex-Daten sichern
	pVData = new BYTE[VertexCnt * VertexSize];
	pMesh->LockVertexBuffer (D3DLOCK_READONLY, &pLockData);
	memcpy (pVData, pLockData, VertexCnt * VertexSize);
	pMesh->UnlockVertexBuffer ();
	//Index-Daten sichern
	pIData = new BYTE[FaceCnt * 3 * sizeof (USHORT)];
	pMesh->LockIndexBuffer (D3DLOCK_READONLY, &pLockData);
	memcpy (pIData, pLockData, FaceCnt * 3 * sizeof (USHORT));
	pMesh->UnlockIndexBuffer ();
	//Attribute-Daten sichern
	pAData = new BYTE[FaceCnt * sizeof (DWORD)];
	pMesh->LockAttributeBuffer (D3DLOCK_READONLY, (DWORD **) &pLockData);
	memcpy (pAData, pLockData, FaceCnt * sizeof (DWORD));
	pMesh->UnlockAttributeBuffer ();
}

void Revo3dObject::ReStoreData ()
{
	if (!pVData || !pIData || !pAData)
		return;

	BYTE *pLockData;

	//Mesh neu erstellen
	if (FAILED (D3DXCreateMeshFVF (FaceCnt, VertexCnt, D3DXMESH_MANAGED, FVF, pDevice, &pMesh)))
		return;
	//Vertex-Daten zurcksichern
	pMesh->LockVertexBuffer (0, &pLockData);
	memcpy (pLockData, pVData, VertexCnt * VertexSize);
	pMesh->UnlockVertexBuffer ();
	delete []pVData;
	pVData = NULL;
	//Index-Daten zurcksichern
	pMesh->LockIndexBuffer (0, &pLockData);
	memcpy (pLockData, pIData, FaceCnt * 3 * sizeof (USHORT));
	pMesh->UnlockIndexBuffer ();
	delete []pIData;
	pIData = NULL;
	//Attribute-Daten zurcksichern
	pMesh->LockAttributeBuffer (0, (DWORD **) &pLockData);
	memcpy (pLockData, pAData, FaceCnt * sizeof (DWORD));
	pMesh->UnlockAttributeBuffer ();
	delete []pAData;
	pAData = NULL;
}



D3DXVECTOR3 Revo3dObject::GetExtremePoint (int Index)
{
	if (!pMesh)
		return (Vector3dMake (0, 0, 0));

	DWORD i;
	float lx, ly, lz, hx, hy, hz;
	BYTE *pVert;
	D3DXVECTOR3 RdVec, retVec;

	pMesh->LockVertexBuffer (D3DLOCK_READONLY, &pVert);
	for (i = 0; i < VertexCnt; i++)
	{
		memcpy (&RdVec, &pVert[i * VertexSize], sizeof (D3DXVECTOR3));
		if (i == 0 || RdVec.x < lx)
			lx = RdVec.x;
		if (i == 0 || RdVec.x > hx)
			hx = RdVec.x;
		if (i == 0 || RdVec.y < ly)
			ly = RdVec.y;
		if (i == 0 || RdVec.y > hy)
			hy = RdVec.y;
		if (i == 0 || RdVec.z < lz)
			lz = RdVec.z;
		if (i == 0 || RdVec.z > hz)
			hz = RdVec.z;
	}
	pMesh->UnlockVertexBuffer ();
	if (Index == 0)
		//Minima
		retVec = Vector3dMake (lx, ly, lz);
	else if (Index == 1)
		//Maxima
		retVec = Vector3dMake (hx, hy, hz);
	else
		//Mittelpunkt
		retVec = Vector3dMake ((lx + hx) / 2.0f, (ly + hy) / 2.0f, (lz + hz) / 2.0f);
	D3DXVec3Add (&retVec, &retVec, &Position);
	return (retVec);	
}



bool Revo3dObject::GetTransparency (DWORD MaterialIndex)
{
	if (!pMaterials)
		return (false);
	if (pMaterials[MaterialIndex].Diffuse.a != 1)
		return (true);
	else
		return (false);
}






















Revo3dPolygons::Revo3dPolygons ()
{
	pVBuf = NULL;
	pIBuf = NULL;
	pVData = NULL;
	pIData = NULL;
	pEngine = NULL;
}

Revo3dPolygons &Revo3dPolygons::operator= (Revo3dPolygons &pOrig)
{
	if (&pOrig == this || !pOrig.pEngine || !pOrig.pVBuf)
		return (*this);

	BYTE *pDest, *pSrc;

	Cleanup ();
	pEngine = pOrig.pEngine;
	pDevice = pEngine->pD3DDevice;
	Dynamic = pOrig.Dynamic;
	if (!ChangeBufferSize (pOrig.VertexMaxCnt, pOrig.IndexMaxCnt))
		return (*this);
	if (Dynamic)
	{
		//Daten-Buffer kopieren
		memcpy (pVData, pOrig.pVData, pOrig.VertexMaxCnt * sizeof (Vertex));
		if (pOrig.pIData)
			memcpy (pIData, pOrig.pIData, pOrig.IndexMaxCnt * sizeof (USHORT));
	}
	else
	{
		//Vertex-Buffer kopieren
		pVBuf->Lock (0, 0, &pDest, 0);
		pOrig.pVBuf->Lock (0, 0, &pSrc, D3DLOCK_READONLY);
		memcpy (pDest, pSrc, pOrig.VertexMaxCnt * sizeof (Vertex));
		pOrig.pVBuf->Unlock ();
		pVBuf->Unlock ();
		//Index-Buffer kopieren, falls erforderlich
		if (pIBuf)
		{
			pIBuf->Lock (0, 0, &pDest, 0);
			pOrig.pIBuf->Lock (0, 0, &pSrc, D3DLOCK_READONLY);
			memcpy (pDest, pSrc, pOrig.IndexMaxCnt * sizeof (USHORT));
			pOrig.pIBuf->Unlock ();
			pIBuf->Unlock ();
		}
	}
	//Sonstige Variablen kopieren
	VertexMaxCnt = pOrig.VertexMaxCnt;
	IndexMaxCnt = pOrig.IndexMaxCnt;
	Position = pOrig.Position;
	matTrans = pOrig.matTrans;
	matRot = pOrig.matRot;
	PolyFormat = pOrig.PolyFormat;
	pTex = pOrig.pTex;
	Mat = pOrig.Mat;
	Visibility = pOrig.Visibility;
	Indexed = pOrig.Indexed;

	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_3DPOLYS);
	return (*this);
}

Revo3dPolygons::~Revo3dPolygons ()
{
	Cleanup ();
	if (pVData)
		free (pVData);
	if (pIData)
		free (pIData);
	if (pEngine)
		pEngine->UnregisterObject (this);
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert die 3d-Polygons.
// Parameter: Eng (die Hauptengine), Dynamic (ob die Revo3dPolygons fr dynamische
// Verwendung optimiert werden sollen)
// Bemerkung: Setzt als Default-Eigenschaft die Position auf (0, 0, 0), das PolyFormat auf
// TRIANGLELIST, keine Textur, das Material auf das Default-Material der Revo3dEngine,
// die Sichtbarkeit auf die Vorderseite und den Indiziert-Status auf falsch.
//-----------------------------------------------------------------------------
void Revo3dPolygons::Initialize (Revo3dEngine *Eng, bool Dynamic)
{
	if (!Eng)
		return;

	Cleanup ();
	pEngine = Eng;
	pDevice = pEngine->pD3DDevice;
	VertexMaxCnt = 0;
	IndexMaxCnt = 0;
	Position = Vector3dMake (0, 0, 0);
	D3DXMatrixIdentity (&matTrans);
	D3DXMatrixIdentity (&matRot);
	PolyFormat = TRIANGLELIST;
	pTex = NULL;
	Mat = pEngine->DefaultMaterial;
	Visibility = FRONTSIDE;
	Indexed = false;
	this->Dynamic = Dynamic;
	pEngine->UnregisterObject (this);
	pEngine->RegisterObject (this, OBJ_3DPOLYS);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert die 3d-Polygons.
//-----------------------------------------------------------------------------
void Revo3dPolygons::Cleanup ()
{
	if (pVBuf)
	{
		pVBuf->Release ();
		pVBuf = NULL;
	}
	if (pIBuf)
	{
		pIBuf->Release ();
		pIBuf = NULL;
	}
	if (Dynamic)
	{
		//Wenn nicht dynamisch, werden diese Buffer fr ReStoreData gebraucht.
		if (pVData)
		{
			free (pVData);
			pVData = NULL;
		}
		if (pIData)
		{
			free (pIData);
			pIData = NULL;
		}
	}
}



//-----------------------------------------------------------------------------
// Funktion: SetVertexData
// Beschreibung: Setzt die Vertex-Daten fr die zu rendernden Polygone.
// Falls ntig, wird die Gre des Buffers angepasst.
// Parameter: vData (Pointer auf die Source-Vertexe), Offset (Startindex im Ziel-Buffer,
// in Vertexe, beginnend bei 0), cnt (Anzahl der zu kopierenden Vertexe)
//-----------------------------------------------------------------------------
void Revo3dPolygons::SetVertexData (Vertex *vData, int Offset, int cnt)
{
	if (!vData || !cnt)
		return;

	BYTE *pVertices;

	if ((Dynamic && !pVData) || (!Dynamic && !pVBuf) || Offset + cnt > VertexMaxCnt)
	{
		//Grennderung des Buffers ist notwendig
		if (!ChangeBufferSize (Offset + cnt, 0))
			return;
	}
	//Neue Vertexe (dazu)kopieren
	if (Dynamic)
		memcpy (&pVData[Offset * sizeof (Vertex)], vData, cnt * sizeof (Vertex));
	else
	{
		pVBuf->Lock (Offset * sizeof (Vertex), cnt * sizeof (Vertex), &pVertices, 0);
		memcpy (pVertices, vData, cnt * sizeof (Vertex));
		pVBuf->Unlock ();
	}
}

//-----------------------------------------------------------------------------
// Funktion: GetVertexData
// Beschreibung: Gibt die Vertex-Daten fr die zu rendernden Polygone zurck.
// Parameter: vData (Pointer auf die zu schreibenden Vertexe), Offset (Startindex
// im Buffer, in Vertexe, beginnend bei 0), cnt (Anzahl der zu lesenden Vertexe)
//-----------------------------------------------------------------------------
void Revo3dPolygons::GetVertexData (Vertex *vData, int Offset, int cnt)
{
	if ((Dynamic && !pVData) || (!Dynamic && !pVBuf) || Offset + cnt > VertexMaxCnt || !vData || !cnt)
		return;

	BYTE *pVertices;

	//Vertexe lesen
	if (Dynamic)
		memcpy (vData, &pVData[Offset * sizeof (Vertex)], cnt * sizeof (Vertex));
	else
	{
		pVBuf->Lock (Offset * sizeof (Vertex), cnt * sizeof (Vertex), &pVertices, D3DLOCK_READONLY);
		memcpy (vData, pVertices, cnt * sizeof (Vertex));
		pVBuf->Unlock ();
	}
}

//-----------------------------------------------------------------------------
// Funktion: SetIndexData
// Beschreibung: Setzt die Index-Daten fr die zu rendernden Polygone.
// Falls ntig, wird die Gre des Buffers angepasst.
// Parameter: iData (Pointer auf die Source-Indizes), Offset (Startindex im Ziel-Buffer,
// in Indizes, beginnend bei 0), cnt (Anzahl der zu kopierenden Indizes)
//-----------------------------------------------------------------------------
void Revo3dPolygons::SetIndexData (unsigned short *iData, int Offset, int cnt)
{
	if (!iData || !cnt)
		return;

	BYTE *pIndices;

	if ((Dynamic && !pIData) || (!Dynamic && !pIBuf) || Offset + cnt > IndexMaxCnt)
	{
		//Grennderung des Index-Buffers ist notwendig
		if (!ChangeBufferSize (0, Offset + cnt))
			return;
	}
	//Neue Indizes (dazu)kopieren
	if (Dynamic)
		memcpy (&pIData[Offset * sizeof (USHORT)], iData, cnt * sizeof (USHORT));
	else
	{
		pIBuf->Lock (Offset * sizeof (USHORT), cnt * sizeof (USHORT), &pIndices, 0);
		memcpy (pIndices, iData, cnt * sizeof (USHORT));
		pIBuf->Unlock ();
	}
}

//-----------------------------------------------------------------------------
// Funktion: GetIndexData
// Beschreibung: Gibt die Index-Daten fr die zu rendernden Polygone zurck.
// Parameter: iData (Pointer auf die zu schreibenden Indizes), Offset (Startindex
// im Buffer, in Indizes, beginnend bei 0), cnt (Anzahl der zu lesenden Indizes)
//-----------------------------------------------------------------------------
void Revo3dPolygons::GetIndexData (unsigned short *iData, int Offset, int cnt)
{
	if ((Dynamic && !pIData) || (!Dynamic && !pIBuf) || Offset + cnt > IndexMaxCnt || !iData || !cnt)
		return;

	BYTE *pIndices;

	//Indizes lesen
	if (Dynamic)
		memcpy (iData, &pIData[Offset * sizeof (USHORT)], cnt * sizeof (USHORT));
	else
	{
		pIBuf->Lock (Offset * sizeof (USHORT), cnt * sizeof (USHORT), &pIndices, D3DLOCK_READONLY);
		memcpy (iData, pIndices, cnt * sizeof (USHORT));
		pIBuf->Unlock ();
	}
}



//-----------------------------------------------------------------------------
// Funktion: SetMaterial
// Beschreibung: Setzt das Material der Polygone.
// Parameter: Mat (Pointer auf das zu setzende Material)
//-----------------------------------------------------------------------------
void Revo3dPolygons::SetMaterial (D3DMATERIAL8 *Mat)
{
	if (!Mat)
		return;
	this->Mat = *Mat;
}



//-----------------------------------------------------------------------------
// Funktion: SetPosition
// Beschreibung: Setzt die Position der Polygone im 3d-Raum.
// Parameter: Pos (die zu setzende Position)
//-----------------------------------------------------------------------------
void Revo3dPolygons::SetPosition (D3DXVECTOR3 *Pos)
{
	if (!Pos)
		return;
	Position = *Pos;
	D3DXMatrixTranslation (&matTrans, Position.x, Position.y, Position.z);
}

//-----------------------------------------------------------------------------
// Funktion: Move
// Beschreibung: Verschiebt die Position der Polygone im 3d-Raum.
// Parameter: MoveDelta (die zu verschiebende Strecke)
//-----------------------------------------------------------------------------
void Revo3dPolygons::Move (D3DXVECTOR3 *MoveDelta)
{
	if (!MoveDelta)
		return;
	D3DXVec3Add (&Position, &Position, MoveDelta);
	D3DXMatrixTranslation (&matTrans, Position.x, Position.y, Position.z);
}



//-----------------------------------------------------------------------------
// Funktion: SetRotationMatrix
// Beschreibung: Setzt die Rotation der Polygone neu anhand einer Rotations-Matrix.
// Parameter: Matrix (die Rotations-Matrix)
//-----------------------------------------------------------------------------
void Revo3dPolygons::SetRotationMatrix (D3DXMATRIX *Matrix)
{
	if (!Matrix)
		return;
	matRot = *Matrix;
}

//-----------------------------------------------------------------------------
// Funktion: SetRotation
// Beschreibung: Setzt die Rotation der Polygone neu, ausgehend von den originalen, nicht
// rotierten Polygonen. Das heit, dass beim Angle-Parameter bisherige Rotationen dieser
// Polygone nicht bercksichtig werden mssen.
// Parameter: Angle (Winkel fr die Rotation um die x, y und z-Achse), Order (Die
// Reihenfolge, in der die Rotationen durchgefhrt werden sollen, beginnend bei 0.
// Wenn z.B. zuerst in y, dann in x und anschlieend in z rotiert werden soll, muss
// der Vektor so aussehen: x = 1, y = 0, z = 2. Wenn NULL bergeben wird, wird
// zuerst um x, dann um y und anschlieend um z rotiert.), Origin (der Mittelpunkt
// der Rotation)
//-----------------------------------------------------------------------------
void Revo3dPolygons::SetRotation (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin)
{
	if (!Angle || !Origin)
		return;

	D3DXMATRIX matTmp1, matTmp2;

	D3DXMatrixTranslation (&matTmp1, Position.x - Origin->x, Position.y - Origin->y, Position.z - Origin->z);
	matRot = RotationMatrixMake (Angle, Order);
	D3DXMatrixMultiply (&matTmp1, &matTmp1, &matRot);
	D3DXMatrixTranslation (&matTmp2, Origin->x - Position.x, Origin->y - Position.y, Origin->z - Position.z);
	D3DXMatrixMultiply (&matRot, &matTmp1, &matTmp2);
}

//-----------------------------------------------------------------------------
// Funktion: Rotate
// Beschreibung: Rotiert die Polygone, ausgehend von den aktuellen, mglicherweise bereits
// frher rotierten Polygonen. Das heit, dass beim Angle-Parameter bisherige Rotationen dieser
// Polygone bercksichtig werden mssen.
// Parameter: Angle (Winkel fr die Rotation um die x, y und z-Achse), Order (Die
// Reihenfolge, in der die Rotationen durchgefhrt werden sollen, beginnend bei 0.
// Wenn z.B. zuerst in y, dann in x und anschlieend in z rotiert werden soll, muss
// der Vektor so aussehen: x = 1, y = 0, z = 2. Wenn NULL bergeben wird, wird
// zuerst um x, dann um y und anschlieend um z rotiert.), Origin (der Mittelpunkt
// der Rotation)
//-----------------------------------------------------------------------------
void Revo3dPolygons::Rotate (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin)
{
	if (!Angle || !Origin)
		return;

	D3DXMATRIX matTmp1, matTmp2;

	D3DXMatrixTranslation (&matTmp1, Position.x - Origin->x, Position.y - Origin->y, Position.z - Origin->z);
	matTmp2 = RotationMatrixMake (Angle, Order);
	D3DXMatrixMultiply (&matTmp1, &matTmp1, &matTmp2);
	D3DXMatrixTranslation (&matTmp2, Origin->x - Position.x, Origin->y - Position.y, Origin->z - Position.z);
	D3DXMatrixMultiply (&matTmp1, &matTmp1, &matTmp2);
	D3DXMatrixMultiply (&matRot, &matRot, &matTmp1);
}



//-----------------------------------------------------------------------------
// Funktion: ChangeBufferSize
// Beschreibung: ndert die Gre des Vertex- oder Index-Buffers. Beim Aufrufen von
// SetVertexData oder SetIndexData wird der Buffer automatisch angepasst.
// Parameter: NewStaticVCnt (die neue Gre des Vertex-Buffers, 0 um sie nicht zu
// ndern), NewStaticICnt (die neue Gre des Index-Buffers, 0 um sie nicht zu ndern)
//-----------------------------------------------------------------------------
bool Revo3dPolygons::ChangeBufferSize (int NewStaticVCnt, int NewStaticICnt)
{
	if (!pDevice)
		return (false);

	BYTE *pVertices, *pIndices;
	void *pSave;
	int SaveVertexCnt, SaveIndexCnt;

	//Vertex-Buffer ndern
	if (Dynamic)
	{
		if (NewStaticVCnt != VertexMaxCnt && NewStaticVCnt > 0)
		{
			VertexMaxCnt = NewStaticVCnt;
			if (pVData)
				pVData = (BYTE *) realloc (pVData, VertexMaxCnt * sizeof (Vertex));
			else
				pVData = (BYTE *) malloc (VertexMaxCnt * sizeof (Vertex));
		}
	}
	else
	{
		pSave = NULL;
		if ((NewStaticVCnt != VertexMaxCnt || !pVBuf) && NewStaticVCnt > 0)
		{
			if (pVBuf)
			{
				//Daten vom existierenden Buffer sichern
				if (NewStaticVCnt > VertexMaxCnt)
					SaveVertexCnt = VertexMaxCnt;
				else
					SaveVertexCnt = NewStaticVCnt;
				pSave = new Vertex[SaveVertexCnt];
				pVBuf->Lock (0, SaveVertexCnt * sizeof (Vertex), &pVertices, D3DLOCK_READONLY);
				memcpy (pSave, pVertices, SaveVertexCnt * sizeof (Vertex));
				pVBuf->Unlock ();
				pVBuf->Release ();
				pVBuf = NULL;
			}
			VertexMaxCnt = NewStaticVCnt;
			//Neu erstellen
			if (FAILED (pDevice->CreateVertexBuffer (VertexMaxCnt * sizeof (Vertex), 0, VERTEX_FVF, D3DPOOL_MANAGED, &pVBuf)))
			{
				if (pSave)
					delete []pSave;
				return (false);
			}
			//Vorherige Daten wiederherstellen
			if (pSave)
			{
				pVBuf->Lock (0, SaveVertexCnt * sizeof (Vertex), &pVertices, 0);
				memcpy (pVertices, pSave, SaveVertexCnt * sizeof (Vertex));
				pVBuf->Unlock ();
				delete []pSave;
			}
		}
	}

	//Index-Buffer ndern
	if (Dynamic)
	{
		if (NewStaticICnt != IndexMaxCnt && NewStaticICnt > 0)
		{
			IndexMaxCnt = NewStaticICnt;
			if (pIData)
				pIData = (BYTE *) realloc (pIData, IndexMaxCnt * sizeof (USHORT));
			else
				pIData = (BYTE *) malloc (IndexMaxCnt * sizeof (USHORT));
		}
	}
	else
	{	
		pSave = NULL;
		if ((NewStaticICnt != IndexMaxCnt || !pIBuf) && NewStaticICnt > 0)
		{
			if (pIBuf)
			{
				//Daten vom existierenden Buffer sichern
				if (NewStaticICnt > IndexMaxCnt)
					SaveIndexCnt = IndexMaxCnt;
				else
					SaveIndexCnt = NewStaticICnt;
				pSave = new USHORT[SaveIndexCnt];
				pIBuf->Lock (0, SaveIndexCnt * sizeof (USHORT), &pIndices, D3DLOCK_READONLY);
				memcpy (pSave, pIndices, SaveIndexCnt * sizeof (USHORT));
				pIBuf->Unlock ();
				pIBuf->Release ();
				pIBuf = NULL;
			}
			IndexMaxCnt = NewStaticICnt;
			//Neu erstellen
			if (FAILED (pDevice->CreateIndexBuffer (IndexMaxCnt * sizeof (USHORT), 0, D3DFMT_INDEX16, D3DPOOL_MANAGED, &pIBuf)))
			{
				if (pSave)
					delete []pSave;
				return (false);
			}
			//Vorherige Daten wiederherstellen
			if (pSave)
			{
				pIBuf->Lock (0, SaveIndexCnt * sizeof (USHORT), &pIndices, 0);
				memcpy (pIndices, pSave, SaveIndexCnt * sizeof (USHORT));
				pIBuf->Unlock ();
				delete []pSave;
			}
		}
	}
	return (true);
}



void Revo3dPolygons::StoreData ()
{
	if (Dynamic || !pVBuf || pVData || pIData)
		return;

	BYTE *pLockData;

	//Vertex-Daten sichern
	pVData = (BYTE *) malloc (VertexMaxCnt * sizeof (Vertex));
	pVBuf->Lock (0, 0, &pLockData, D3DLOCK_READONLY);
	memcpy (pVData, pLockData, VertexMaxCnt * sizeof (Vertex));
	pVBuf->Unlock ();
	//Index-Daten sichern, falls erforderlich
	if (pIBuf)
	{
		pIData = (BYTE *) malloc (IndexMaxCnt * sizeof (USHORT));
		pIBuf->Lock (0, 0, &pLockData, D3DLOCK_READONLY);
		memcpy (pIData, pLockData, IndexMaxCnt * sizeof (USHORT));
		pIBuf->Unlock ();
	}
}

void Revo3dPolygons::ReStoreData ()
{
	if (Dynamic || !pVData)
		return;

	BYTE *pLockData;

	//Buffer neu erstellen
	if (!ChangeBufferSize (VertexMaxCnt, IndexMaxCnt))
		return;
	//Vertex-Daten zurcksichern
	pVBuf->Lock (0, 0, &pLockData, 0);
	memcpy (pLockData, pVData, VertexMaxCnt * sizeof (Vertex));
	pVBuf->Unlock ();
	free (pVData);
	pVData = NULL;
	//Index-Daten zurcksichern, falls erforderlich
	if (pIData)
	{
		pIBuf->Lock (0, 0, &pLockData, 0);
		memcpy (pLockData, pIData, IndexMaxCnt * sizeof (USHORT));
		pIBuf->Unlock ();
		free (pIData);
		pIData = NULL;
	}
}