//-----------------------------------------------------------------------------
// Datei: RevoFunctions.cpp
// Beschreibung: Enthlt grundstzliche Funktionen.
//-----------------------------------------------------------------------------


#include <windows.h>
#include "revo3d.h"




D3DXMATRIX RotationMatrixMake (D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order)
{
	int i, j;
	D3DXMATRIX RMat, Rotation;

	D3DXMatrixIdentity (&Rotation);
	for (i = 0; i < 3; i++)
	{
		if (Order)
		{
			if (i == Order->x)
				j = 0;
			else if (i == Order->y)
				j = 1;
			else if (i == Order->z)
				j = 2;
		}
		else
			j = i;
		//j bezeichnet jetzt die Achse, um die rotiert werden soll.
		if (j == 0 && Angle->x)
		{
			D3DXMatrixRotationX (&RMat, Angle->x);
			D3DXMatrixMultiply (&Rotation, &Rotation, &RMat);
		}
		else if (j == 1 && Angle->y)
		{
			D3DXMatrixRotationY (&RMat, Angle->y);
			D3DXMatrixMultiply (&Rotation, &Rotation, &RMat);
		}
		else if (j == 2 && Angle->z)
		{
			D3DXMatrixRotationZ (&RMat, Angle->z);
			D3DXMatrixMultiply (&Rotation, &Rotation, &RMat);
		}
	}
	return (Rotation);
}



D3DXVECTOR3 MirrorPoint (Revo3dPlane *pPlane, D3DXVECTOR3 *pPoint)
{
	D3DXVECTOR3 OpVec, SpVec;

	//Spiegelt einen gegebenen Punkt auf die andere Seite der Ebene.
	D3DXVec3Add (&OpVec, pPoint, &pPlane->NormalVec);
	D3DXPlaneIntersectLine (&SpVec, &pPlane->Plane, pPoint, &OpVec);
	D3DXVec3Subtract (&OpVec, &SpVec, pPoint);
	D3DXVec3Add (&OpVec, &SpVec, &OpVec);
	return (OpVec);
}



void GetFolder (char *Folder, char *File)
{
	int i, l;

	l = strlen (File);
	for (i = l - 1; i >= 0; i--)
	{
		if (File[i] == '\\')
		{
			strcpy (Folder, File);
			Folder[i + 1] = '\0';
			return;
		}
	}
	Folder[0] = '\0';
}



GUID GuidMake (unsigned int Identifier)
{
	int i;
	GUID retGuid;

	retGuid.Data1 = (DWORD) Identifier;
	retGuid.Data2 = (WORD) Identifier;
	retGuid.Data3 = (WORD) Identifier;
	for (i = 0; i < 8; i++)
		retGuid.Data4[i] = (BYTE) Identifier;
	return (retGuid);
}



WCHAR *UnicodeStringMake (char *String)
{
	WCHAR *wStr;

	wStr = new WCHAR[strlen (String) + 1];
	MultiByteToWideChar (CP_ACP, 0, String, -1, wStr, strlen (String) + 1);
	return (wStr);
}

char *UnicodeStringTranslate (WCHAR *UnicodeString)
{
	char *Str;

	Str = new char[wcslen (UnicodeString) + 1];
	WideCharToMultiByte (CP_ACP, 0, UnicodeString, -1, Str, wcslen (UnicodeString) + 1, NULL, NULL);
	return (Str);
}



//-----------------------------------------------------------------------------
// Funktion: Vector2dMake
// Beschreibung: Erzeugt einen 2d-Vektor mit den angegebenen Koordinaten.
// Parameter: x, y
//-----------------------------------------------------------------------------
D3DXVECTOR2 Vector2dMake (float x, float y)
{
	D3DXVECTOR2 retVec;

	retVec.x = x;
	retVec.y = y;
	return (retVec);
}



//-----------------------------------------------------------------------------
// Funktion: NormalVector2dMake
// Beschreibung: Erzeugt einen Normalvektor zu den angegebenen zwei Punkten, wobei
// der Vektor von Punkt 1 zu Punkt 2 angenommen wird.
// Parameter: v1, v2 (die jeweiligen Punkte)
//-----------------------------------------------------------------------------
D3DXVECTOR2 NormalVector2dMake (D3DXVECTOR2* v1, D3DXVECTOR2 *v2)
{
	D3DXVECTOR2 ErgVec, retVec;

	D3DXVec2Subtract (&ErgVec, v2, v1);
	retVec.x = ErgVec.y;
	retVec.y = -ErgVec.x;
	D3DXVec2Normalize (&retVec, &retVec);
	return (retVec);
}



//-----------------------------------------------------------------------------
// Funktion: Rotate
// Beschreibung: Rotiert einen 2d-Vektor.
// Parameter: Vector (der Vektor, der rotiert werden soll), Angle (Winkel fr die
// Rotation), Origin (der Mittelpunkt der Rotation)
//-----------------------------------------------------------------------------
D3DXVECTOR2 Vector2dRotate (D3DXVECTOR2 *Vector, float Angle, D3DXVECTOR2 *Origin)
{
	if (!Vector || !Origin)
		return (Vector2dMake (0, 0));

	float cosPhi, sinPhi;
	D3DXVECTOR2 retVec;

	cosPhi = (float) cos (Angle);
	sinPhi = (float) sin (Angle);
	retVec.x = (Vector->x - Origin->x) * cosPhi - (Vector->y - Origin->y) * sinPhi + Origin->x;
	retVec.y = (Vector->x - Origin->x) * sinPhi + (Vector->y - Origin->y) * cosPhi + Origin->y;
	return (retVec);
}



//-----------------------------------------------------------------------------
// Funktion: Vector3dMake
// Beschreibung: Erzeugt einen 3d-Vektor mit den angegebenen Koordinaten.
// Parameter: x, y, z
//-----------------------------------------------------------------------------
D3DXVECTOR3 Vector3dMake (float x, float y, float z)
{
	D3DXVECTOR3 retVec;

	retVec.x = x;
	retVec.y = y;
	retVec.z = z;
	return (retVec);
}



//-----------------------------------------------------------------------------
// Funktion: NormalVector3dMake
// Beschreibung: Erzeugt einen Normalvektor zu den angegebenen drei Punkten.
// Parameter: v1, v2, v3 (die jeweiligen Punkte)
//-----------------------------------------------------------------------------
D3DXVECTOR3 NormalVector3dMake (D3DXVECTOR3* v1, D3DXVECTOR3* v2, D3DXVECTOR3* v3)
{
	if (!v1 || !v2 || !v3)
		return (Vector3dMake (0, 0, 0));

	D3DXVECTOR3 retVec, VecA, VecB;

    D3DXVec3Subtract (&VecA, v2, v1);
	D3DXVec3Subtract (&VecB, v3, v2);
	D3DXVec3Cross (&retVec, &VecA, &VecB);
	D3DXVec3Normalize (&retVec, &retVec);
	return (retVec);
}



//-----------------------------------------------------------------------------
// Funktion: Rotate
// Beschreibung: Rotiert einen 3d-Vektor.
// Parameter: Vector (der Vektor, der rotiert werden soll), Angle (Winkel fr die
// Rotation um die x, y und z-Achse), Order (Die Reihenfolge, in der die Rotationen
// durchgefhrt werden sollen, beginnend bei 0. Wenn z.B. zuerst in y, dann in x
// und anschlieend in z rotiert werden soll, muss der Vektor so aussehen: x = 1,
// y = 0, z = 2. Wenn NULL bergeben wird, wird zuerst um x, dann um y und anschlieend
// um z rotiert.), Origin (der Mittelpunkt der Rotation)
//-----------------------------------------------------------------------------
D3DXVECTOR3 Vector3dRotate (D3DXVECTOR3 *Vector, D3DXVECTOR3 *Angle, D3DXVECTOR3 *Order, D3DXVECTOR3 *Origin)
{
	if (!Vector || !Angle || !Origin)
		return (Vector3dMake (0, 0, 0));

	D3DXMATRIX RMat, PMat, DMat;
	D3DXVECTOR3 retVec;

	//1. Erstellen der Rotations-Matrix
	RMat = RotationMatrixMake (Angle, Order);

	//2. Rotieren des Vektors
	D3DXMatrixIdentity (&PMat);

	PMat._41 = Vector->x - Origin->x;				//Position rotieren
	PMat._42 = Vector->y - Origin->y;
	PMat._43 = Vector->z - Origin->z;
	D3DXMatrixMultiply (&DMat, &PMat, &RMat);
	retVec.x = DMat._41 + Origin->x;
	retVec.y = DMat._42 + Origin->y;
	retVec.z = DMat._43 + Origin->z;

	return (retVec);
}



//-----------------------------------------------------------------------------
// Funktion: PlaneMake
// Beschreibung: Erzeugt eine 3d-Ebene die aus den drei angegebenen Punkten definiert wird.
// Parameter: p1, p2, p3 (die jeweiligen Punkte)
//-----------------------------------------------------------------------------
Revo3dPlane PlaneMake (D3DXVECTOR3 *p1, D3DXVECTOR3 *p2, D3DXVECTOR3 *p3)
{
	Revo3dPlane retPlane;

	retPlane.Points[0] = *p1;
	retPlane.Points[1] = *p2;
	retPlane.Points[2] = *p3;
	D3DXPlaneFromPoints (&retPlane.Plane, p1, p2, p3);
	retPlane.NormalVec = NormalVector3dMake (p1, p2, p3);
	return (retPlane);
}



//-----------------------------------------------------------------------------
// Funktion: VertexMake
// Beschreibung: Erzeugt einen Vertex
// Parameter: x, y, z, nx, ny, nz, tu, tv
//-----------------------------------------------------------------------------
Vertex VertexMake (float x, float y, float z, float nx, float ny, float nz, float tu, float tv)
{
	Vertex retVertex;

	retVertex.x = x;
	retVertex.y = y;
	retVertex.z = z;
	retVertex.nx = nx;
	retVertex.ny = ny;
	retVertex.nz = nz;
	retVertex.tu = tu;
	retVertex.tv = tv;
	return (retVertex);
}



//-----------------------------------------------------------------------------
// Funktion: TLVertexMake
// Beschreibung: Erzeugt einen TL-Vertex
// Parameter: x, y, tu, tv
//-----------------------------------------------------------------------------
TLVertex TLVertexMake (float x, float y, D3DCOLOR Color, float tu, float tv)
{
	TLVertex retVertex;

	retVertex.x = x;
	retVertex.y = y;
	retVertex.z = 0;
	retVertex.rhw = 1;
	retVertex.Color = Color;
	retVertex.tu = tu;
	retVertex.tv = tv;
	return (retVertex);
}



//-----------------------------------------------------------------------------
// Funktion: BillboardVertexMake
// Beschreibung: Erzeugt Vertexe, die ein Billboard beschreiben.
// Parameter: MidPos (der Mittelpunkt des Billboards, um den sich das Bitmap drehen soll),
// Width (die Breite des Bildes), Height (die Hhe des Bildes), CamPos (die Kamera-
// Position), AllowXRotation (Ob das Bild auch um die x-Achse rotieren darf. Wenn false,
// rotiert das Bild nur um die y-Achse, was z.B. fr Bume verwendet werden kann.),
// outV (Pointer auf ein Feld von mindestens 4 Vertices).
// Bemerkung: Das Feld muss als Triangle-Strip mit 2 Polygonen gerendert werden.
//-----------------------------------------------------------------------------
void BillboardVertexMake (D3DXVECTOR3 *MidPos, float Width, float Height, D3DXVECTOR3 *CamPos, bool AllowXRotation, Vertex *outV)
{
	if (!MidPos || !CamPos || !outV)
		return;
	if (MidPos->x == CamPos->x && MidPos->y == CamPos->y && MidPos->z == CamPos->z)
		return;

	int i;
	D3DXVECTOR3 Delta, DeltaCube, NVecW, NVecH;
	D3DXVECTOR3 Corner[4];
	float DeltaCubeSum, HypSum, sinx, cosx, siny, cosy;

	//Delta berechnen
	D3DXVec3Subtract (&Delta, MidPos, CamPos);
	DeltaCube.x = Delta.x * Delta.x;
	DeltaCube.y = Delta.y * Delta.y;
	DeltaCube.z = Delta.z * Delta.z;
	DeltaCubeSum = DeltaCube.x + DeltaCube.y + DeltaCube.z;
	//sin und cos der beiden Winkeln berechnen
	if (AllowXRotation)
	{
		sinx = Delta.y / (float) sqrt ((double) DeltaCubeSum);
		cosx = (float) sqrt ((double) (1 - DeltaCube.y / DeltaCubeSum));
	}
	if (DeltaCube.x + DeltaCube.z)
	{
		HypSum = (float) sqrt ((double) (DeltaCube.x + DeltaCube.z));
		siny = Delta.x / HypSum;
		cosy = Delta.z / HypSum;
	}
	else
	{
		siny = 0;
		cosy = 1;
	}
	//NVecW erstellen
	NVecW = Vector3dMake (cosy, 0, -siny);
	D3DXVec3Scale (&NVecW, &NVecW, Width / 2.0f);
	//NVecH erstellen
	if (AllowXRotation)
	{
		NVecH = Vector3dMake (-sinx * siny, cosx, -sinx * cosy);
		D3DXVec3Scale (&NVecH, &NVecH, Height / 2.0f);
	}
	else
		NVecH = Vector3dMake (0, Height / 2.0f, 0);
	//Eckpunkte berechnen
	D3DXVec3Subtract (&Corner[0], MidPos, &NVecW);
	D3DXVec3Add (&Corner[0], &Corner[0], &NVecH);
	D3DXVec3Add (&Corner[1], MidPos, &NVecW);
	D3DXVec3Add (&Corner[1], &Corner[1], &NVecH);
	D3DXVec3Subtract (&Corner[2], MidPos, &NVecW);
	D3DXVec3Subtract (&Corner[2], &Corner[2], &NVecH);
	D3DXVec3Add (&Corner[3], MidPos, &NVecW);
	D3DXVec3Subtract (&Corner[3], &Corner[3], &NVecH);
	//In die Vertices schreiben
	ZeroMemory (outV, sizeof (Vertex) * 4);
	D3DXVec3Subtract (&Delta, CamPos, MidPos);
	D3DXVec3Normalize (&Delta, &Delta);
	for (i = 0; i < 4; i++)
	{
		outV[i].x = Corner[i].x;
		outV[i].y = Corner[i].y;
		outV[i].z = Corner[i].z;
		outV[i].nx = Delta.x;
		outV[i].ny = Delta.y;
		outV[i].nz = Delta.z;
	}
	outV[1].tu = 1;
	outV[2].tv = 1;
	outV[3].tu = 1;
	outV[3].tv = 1;
}



//-----------------------------------------------------------------------------
// Funktion: ColorMake
// Beschreibung: Erzeugt eine Farbe ohne Transparenz.
// Parameter: r, g, b (die Farbanteile)
//-----------------------------------------------------------------------------
D3DCOLOR ColorMake (int r, int g, int b)
{
	return (D3DCOLOR_XRGB (r, g, b));
}



//-----------------------------------------------------------------------------
// Funktion: ColorAlphaMake
// Beschreibung: Erzeugt eine Farbe mit Transparenz.
// Parameter: a (der Transparenz-Anteil), r, g, b (die Farbanteile)
//-----------------------------------------------------------------------------
D3DCOLOR ColorAlphaMake (int a, int r, int g, int b)
{
	return (D3DCOLOR_ARGB (a, r, g, b));
}



//-----------------------------------------------------------------------------
// Funktion: TextureMake
// Beschreibung: Erzeugt eine Textur von einer Bitmap-Datei.
// Parameter: Eng (die Hauptengine), File (der Dateiname), ColorKey (der ColorKey,
// 0 = kein ColorKey), MipLevels (Wie viele MipMapping-Ebenen erstellt werden sollen.
// 0, um eine komplette MipMap zu erstellen.)
//-----------------------------------------------------------------------------
LPDIRECT3DTEXTURE8 TextureMake (Revo3dEngine *Eng, char *File, D3DCOLOR ColorKey, int MipLevels, bool EnableEdit)
{
	if (!Eng || !File)
		return (NULL);

	LPDIRECT3DTEXTURE8 retTexture;
	D3DFORMAT TexFormat;

	if (EnableEdit)
		TexFormat = D3DFMT_A8R8G8B8;
	else if (ColorKey)
		TexFormat = D3DFMT_A1R5G5B5;
	else
		TexFormat = D3DFMT_R5G6B5;
	if (FAILED (D3DXCreateTextureFromFileExA (Eng->pD3DDevice, File, D3DX_DEFAULT, D3DX_DEFAULT, MipLevels, 0, TexFormat, D3DPOOL_MANAGED, D3DX_FILTER_LINEAR, D3DX_FILTER_LINEAR, ColorKey, NULL, NULL, &retTexture)))
		return (NULL);
	return (retTexture);
}



//-----------------------------------------------------------------------------
// Funktion: MaterialMake
// Beschreibung: Erzeugt ein Direct3D-Material.
// Parameter: rReflect, gReflect, bReflect (die rgb-Werte der Farben, die reflektiert werden knnen),
// rOwn, gOwn, bOwn (die rgb-Werte des eigenen Lichtes), SpecPower (die Intensitt des Glanzlichtes),
// Transparency (die Transparenz)
// Bemerkung: Alle Parameter sind im Bereich von 0 bis 255.
//-----------------------------------------------------------------------------
D3DMATERIAL8 MaterialMake (int rReflect, int gReflect, int bReflect, int rOwn, int gOwn, int bOwn, float SpecPower, int Transparency)
{
	D3DMATERIAL8 retMat;
	D3DCOLORVALUE Reflect, Own;

	ZeroMemory (&Reflect, sizeof (D3DCOLORVALUE));
	Reflect.a = 1 - Transparency / (float) 255.0;
	Reflect.r = (float) rReflect / 255;
	Reflect.g = (float) gReflect / 255;
	Reflect.b = (float) bReflect / 255;
	ZeroMemory (&Own, sizeof (D3DCOLORVALUE));
	Own.a = 1;
	Own.r = (float) rOwn / 255;
	Own.g = (float) gOwn / 255;
	Own.b = (float) bOwn / 255;

	retMat.Ambient = Reflect;
	retMat.Ambient.a = 1;
	retMat.Diffuse = Reflect;
	retMat.Emissive = Own;
	retMat.Specular.a = 1;
	retMat.Specular.r = 1;
	retMat.Specular.g = 1;
	retMat.Specular.b = 1;
	retMat.Power = SpecPower;

	return (retMat);
}



//-----------------------------------------------------------------------------
// Funktion: LightMake
// Beschreibung: Erzeugt eine beliebige Lichtquelle.
// Parameter: Type (von welchem Typ das Licht ist), rEmit, gEmit, bEmit (die rgb-Werte
// des Lichtes, die ausgestrahlt werden), rSpecular, gSpecular, bSpecular (die rgb-Werte
// des Glanzlichtes, die ausgestrahlt werden), Position (die Position der Lichtquelle),
// Direction (die Richtung des ausgestrahlten Lichtes), Range (die maximale Reichweite des
// Lichtes), Attenuation (die Abnahme der Lichtintensitt mit der Entfernung, normalerweise 1),
// InnerCore, OuterCore (die beiden Winkel des inneren und ueren Lichtkegels).
//-----------------------------------------------------------------------------
D3DLIGHT8 LightMake (D3DLIGHTTYPE Type, int rEmit, int gEmit, int bEmit, int rSpecular, int gSpecular, int bSpecular, D3DVECTOR *Position, D3DVECTOR *Direction, float Range, float Attenuation, float InnerCore, float OuterCore)
{
	D3DLIGHT8 retLight;
	D3DCOLORVALUE Emit, Spec;

	ZeroMemory (&Emit, sizeof (D3DCOLORVALUE));
	Emit.a = 1;
	Emit.r = (float) rEmit / 25;
	Emit.g = (float) gEmit / 25;
	Emit.b = (float) bEmit / 25;
	ZeroMemory (&Spec, sizeof (D3DCOLORVALUE));
	Spec.a = 1;
	Spec.r = (float) rSpecular / 25;
	Spec.g = (float) gSpecular / 25;
	Spec.b = (float) bSpecular / 25;

	ZeroMemory (&retLight, sizeof (D3DLIGHT8));
	retLight.Type = Type;
	retLight.Ambient = Emit;
	retLight.Diffuse = Emit;
	retLight.Specular = Spec;
	if (Position)
		retLight.Position = *Position;
	if (Direction)
		retLight.Direction = *Direction;
	retLight.Range = Range;
	retLight.Falloff = 1;
	retLight.Attenuation1 = Attenuation;
	retLight.Theta = InnerCore;
	retLight.Phi = OuterCore;
	return (retLight);
}

//-----------------------------------------------------------------------------
// Funktion: PointLightMake
// Beschreibung: Erzeugt eine punktfrmige Lichtquelle, die in alle Richtungen Licht ausstrahlt.
// Parameter: siehe LightMake (teilweise)
//-----------------------------------------------------------------------------
D3DLIGHT8 PointLightMake (int rEmit, int gEmit, int bEmit, int Specular, D3DVECTOR *Position, float Range)
{
	return (LightMake (D3DLIGHT_POINT, rEmit, gEmit, bEmit, Specular, Specular, Specular, Position, NULL, Range, 1, 0, 0));
}

//-----------------------------------------------------------------------------
// Funktion: SpotLightMake
// Beschreibung: Erzeugt eine Lichtquelle, die mit einer Taschenlampe vergleichbar ist.
// Sie besitzt eine Position und leuchtet in eine bestimmte Richtung.
// Parameter: siehe LightMake (teilweise)
//-----------------------------------------------------------------------------
D3DLIGHT8 SpotLightMake (int rEmit, int gEmit, int bEmit, int Specular, D3DVECTOR *Position, D3DVECTOR *Direction, float Range, float InnerCore, float OuterCore)
{
	return (LightMake (D3DLIGHT_SPOT, rEmit, gEmit, bEmit, Specular, Specular, Specular, Position, Direction, Range, 1, InnerCore, OuterCore));
}

//-----------------------------------------------------------------------------
// Funktion: DirectionalLightMake
// Beschreibung: Erzeugt eine Lichtquelle, die die gesamte Welt beleuchtet und von einer bestimmten
// Richtung kommt.
// Parameter: siehe LightMake (teilweise)
//-----------------------------------------------------------------------------
D3DLIGHT8 DirectionalLightMake (int rEmit, int gEmit, int bEmit, int Specular, D3DVECTOR *Direction)
{
	return (LightMake (D3DLIGHT_DIRECTIONAL, rEmit, gEmit, bEmit, Specular, Specular, Specular, NULL, Direction, 0, 0, 0, 0));
}



//-----------------------------------------------------------------------------
// Funktion: FontMake
// Beschreibung: Erzeugt eine Schriftart.
// Parameter: FontName (der Name der Schriftart, zB "Arial", max. 32 Zeichen),
// Size (Gre), Bold (fett), Italic (kursiv), Underline (unterstrichen)
//-----------------------------------------------------------------------------
HFONT FontMake (char *FontName, int Size, bool Bold, bool Italic, bool Underline)
{
	int w;

	if (!Bold)
		w = 400;
	else
		w = 700;
	return (CreateFont (Size, 0, 0, 0, w, Italic, Underline, false, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, DEFAULT_PITCH, FontName));
}



//-----------------------------------------------------------------------------
// Funktion: ResolutionMake
// Beschreibung: Erzeugt eine Revo3dRes-Auflsungsbeschreibung.
// Parameter: x, y (die x- und y-Auflsung), bpp (Farbtiefe, entweder 16 oder 32 bpp)
//-----------------------------------------------------------------------------
Revo3dRes ResolutionMake (int x, int y, int bpp)
{
	Revo3dRes retRes;

	retRes.ResX = x;
	retRes.ResY = y;
	if (bpp != 32)
		retRes.Format = D3DFMT_R5G6B5;
	else
		retRes.Format = D3DFMT_X8R8G8B8;
	return (retRes);
}