//-----------------------------------------------------------------------------
// Datei: RevoInput.cpp
// Beschreibung: Enthlt Funktionen zum Verwenden der Tastatur und Maus.
//-----------------------------------------------------------------------------


#include <windows.h>
#include "revo3d.h"




RevoInput::RevoInput ()
{
	pDI = NULL;
	pKeyb = NULL;
	pMouse = NULL;
}

RevoInput::~RevoInput ()
{
	Cleanup ();
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert RevoInput.
// Parameter: hWnd (die hWnd der Hauptengine, mu Revo3dEngine.GethWnd sein
// falls sie verwendet wird), hInstance (der hInstance-Parameter der
// WinMain-Funktion), NoWinKeys (ob die Windows-Tastenkombinationen unterdrckt
// werden sollen, sofern mglich), NoWinMouse (ob der Windows-Mauscursor aus=
// geblendet werden soll).
//-----------------------------------------------------------------------------
bool RevoInput::Initialize (HWND hWnd, HINSTANCE hInstance, bool NoWinKeys, bool NoWinMouse)
{
	Cleanup ();

	if (FAILED (DirectInput8Create (hInstance, DIRECTINPUT_VERSION, IID_IDirectInput8, (void **) &pDI, NULL)))
		return (false);

	if (FAILED (pDI->CreateDevice (GUID_SysKeyboard, &pKeyb, NULL)))
		return (false);
	pKeyb->SetDataFormat (&c_dfDIKeyboard);
	if (NoWinKeys)
		pKeyb->SetCooperativeLevel (hWnd, DISCL_FOREGROUND | DISCL_EXCLUSIVE);
	else
		pKeyb->SetCooperativeLevel (hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
	pKeyb->Acquire ();

	if (FAILED (pDI->CreateDevice (GUID_SysMouse, &pMouse, NULL)))
		return (false);
	pMouse->SetDataFormat (&c_dfDIMouse);
	if (NoWinMouse)
		pMouse->SetCooperativeLevel (hWnd, DISCL_FOREGROUND | DISCL_EXCLUSIVE);
	else
		pMouse->SetCooperativeLevel (hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
	pMouse->Acquire ();

	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert RevoInput.
//-----------------------------------------------------------------------------
void RevoInput::Cleanup ()
{
	if (pMouse)
	{
		pMouse->Unacquire ();
		pMouse->Release ();
		pMouse = NULL;
	}
	if (pKeyb)
	{
		pKeyb->Unacquire ();
		pKeyb->Release ();
		pKeyb = NULL;
	}
	if (pDI)
	{
		pDI->Release ();
		pDI = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: ReadKeyboard
// Beschreibung: Liest alle aktuell gedrckten Tasten in ein BYTE-Feld ein.
// Parameter: AppendTo (Pointer auf einen String, in den die Texteingaben geschrieben
// werden. Dies ist daher ideal frs Einlesen eines Textes. Wird bei der bergabe von
// NULL ignoriert.), BufferSize (Die Gre des String-Buffers, d.h. die max. Lnge. Bei
// bergabe von 0 wird dieser Parameter ignoriert.)
// Rckgabewert: Pointer auf das BYTE-Feld (f[0] bis f[255]) oder NULL wenn ein
// Fehler aufgetreten ist.
//-----------------------------------------------------------------------------
BYTE *RevoInput::ReadKeyboard (char *AppendTo, int BufferSize)
{
	if (!pKeyb)
		return (NULL);

	int i;
	HRESULT hr;
	char RdSign, UpStr[2];
	bool ShiftDown;

	if (AppendTo)
		memcpy (KeybDown, KeybData, 256);
	hr = pKeyb->GetDeviceState (sizeof (KeybData), (void *) KeybData);
	if (hr == DIERR_INPUTLOST)
	{
		pKeyb->Acquire ();
		hr = pKeyb->GetDeviceState (sizeof (KeybData), (void *) KeybData);
		if (FAILED (hr))
			return (NULL);
	}
	if (AppendTo && !(BufferSize && (int) strlen (AppendTo) == BufferSize - 1))
	{
		//Eingabe in einen String schreiben
		if (KeybData[DIK_LSHIFT] || KeybData[DIK_RSHIFT])
			ShiftDown = true;
		else
			ShiftDown = false;
		for (i = 0; i < 256; i++)
		{
			if (KeybData[i] && !KeybDown[i])
			{
				//Diese Taste wurde jetzt neu gedrckt, mglicherweise an den String anhngen.
				RdSign = GetChar (i, ShiftDown);
				if (ShiftDown)
				{
					UpStr[0] = RdSign;
					UpStr[1] = '\0';
					strupr (UpStr);
					RdSign = UpStr[0];
				}
				if (KeybData[DIK_BACK])
				{
					if (strlen (AppendTo) != 0)
						AppendTo[strlen (AppendTo) - 1] = '\0';
				}
				else
					strncat (AppendTo, &RdSign, 1);
			}
		}
	}
	return (KeybData);
}



//-----------------------------------------------------------------------------
// Funktion: ReadMouse
// Beschreibung: Liest alle aktuell gedrckten Tasten der Maus und die Position
// der Maus ein.
// Rckgabewert: Die Beschreibung des aktuellen Maus-Status oder NULL wenn ein
// Fehler aufgetreten ist.
//-----------------------------------------------------------------------------
DIMOUSESTATE *RevoInput::ReadMouse ()
{
	if (!pMouse)
		return (NULL);

	HRESULT hr;

	hr = pMouse->GetDeviceState (sizeof (MouseData), (void *) &MouseData);
	if (hr == DIERR_INPUTLOST)
	{
		pMouse->Acquire ();
		hr = pMouse->GetDeviceState (sizeof (MouseData), (void *) &MouseData);
		if (FAILED (hr))
			return (NULL);
	}
	return (&MouseData);	
}





//*****Private Funktionen*****
char RevoInput::GetChar (int KeyCode, bool ShiftDown)
{
	if (ShiftDown)
	{
		if (KeyCode == DIK_0)
			return ('=');
		else if (KeyCode == DIK_1)
			return ('!');
		else if (KeyCode == DIK_2)
			return ('"');
		else if (KeyCode == DIK_3)
			return ('');
		else if (KeyCode == DIK_4)
			return ('$');
		else if (KeyCode == DIK_5)
			return ('%');
		//DIK_6 wre das &, dies wegen Shortcut-(Fehl)Funktion jedoch nicht zulassen!
		else if (KeyCode == DIK_7)
			return ('/');
		else if (KeyCode == DIK_8)
			return ('(');
		else if (KeyCode == DIK_9)
			return (')');
		else if (KeyCode == DIK_MINUS)
			return ('?');
		else if (KeyCode == DIK_SLASH)
			return ('_');
		else if (KeyCode == DIK_COMMA)
			return (';');
		else if (KeyCode == DIK_PERIOD)
			return (':');
		else if (KeyCode == DIK_OEM_102)
			return ('>');
	}
	if (KeyCode == DIK_A)
		return ('a');
	else if (KeyCode == DIK_B)
		return ('b');
	else if (KeyCode == DIK_C)
		return ('c');
	else if (KeyCode == DIK_D)
		return ('d');
	else if (KeyCode == DIK_E)
		return ('e');
	else if (KeyCode == DIK_F)
		return ('f');
	else if (KeyCode == DIK_G)
		return ('g');
	else if (KeyCode == DIK_H)
		return ('h');
	else if (KeyCode == DIK_I)
		return ('i');
	else if (KeyCode == DIK_J)
		return ('j');
	else if (KeyCode == DIK_K)
		return ('k');
	else if (KeyCode == DIK_L)
		return ('l');
	else if (KeyCode == DIK_M)
		return ('m');
	else if (KeyCode == DIK_N)
		return ('n');
	else if (KeyCode == DIK_O)
		return ('o');
	else if (KeyCode == DIK_P)
		return ('p');
	else if (KeyCode == DIK_Q)
		return ('q');
	else if (KeyCode == DIK_R)
		return ('r');
	else if (KeyCode == DIK_S)
		return ('s');
	else if (KeyCode == DIK_T)
		return ('t');
	else if (KeyCode == DIK_U)
		return ('u');
	else if (KeyCode == DIK_V)
		return ('v');
	else if (KeyCode == DIK_W)
		return ('w');
	else if (KeyCode == DIK_X)
		return ('x');
	else if (KeyCode == DIK_Y)
		return ('z');
	else if (KeyCode == DIK_Z)
		return ('y');
	else if (KeyCode == DIK_NUMPAD0 || KeyCode == DIK_0)
		return ('0');
	else if (KeyCode == DIK_NUMPAD1 || KeyCode == DIK_1)
		return ('1');
	else if (KeyCode == DIK_NUMPAD2 || KeyCode == DIK_2)
		return ('2');
	else if (KeyCode == DIK_NUMPAD3 || KeyCode == DIK_3)
		return ('3');
	else if (KeyCode == DIK_NUMPAD4 || KeyCode == DIK_4)
		return ('4');
	else if (KeyCode == DIK_NUMPAD5 || KeyCode == DIK_5)
		return ('5');
	else if (KeyCode == DIK_NUMPAD6 || KeyCode == DIK_6)
		return ('6');
	else if (KeyCode == DIK_NUMPAD7 || KeyCode == DIK_7)
		return ('7');
	else if (KeyCode == DIK_NUMPAD8 || KeyCode == DIK_8)
		return ('8');
	else if (KeyCode == DIK_NUMPAD9 || KeyCode == DIK_9)
		return ('9');
	else if (KeyCode == DIK_SPACE)
		return (' ');
	else if (KeyCode == DIK_COMMA || KeyCode == DIK_DECIMAL)
		return (',');
	else if (KeyCode == DIK_PERIOD)
		return ('.');
	else if (KeyCode == DIK_SUBTRACT)
		return ('-');
	else if (KeyCode == DIK_MINUS)
		return ('');
	else if (KeyCode == DIK_ADD)
		return ('+');
	else if (KeyCode == DIK_MULTIPLY)
		return ('*');
	else if (KeyCode == DIK_SLASH)
		return ('-');
	else if (KeyCode == DIK_OEM_102)
		return ('<');
	else
		return ('\0');
}