//-----------------------------------------------------------------------------
// Datei: RevoNetwork.cpp
// Beschreibung: Enthlt Funktionen zum Erstellen und Verwalten einer Netzwerkverbindung.
//-----------------------------------------------------------------------------


#define _WIN32_DCOM

#include <windows.h>
#include <stdio.h>
#include "revo3d.h"

static int CoInitCnt;
static int NrIPs;
static ServerDesc *pDesc;
static bool Connected;

UINT ReadSize = 0;
BYTE ReadBuf[131072];			//128 KByte Buffer zum Empfangen von Nachrichten



//Statische Variablen initsialisieren
DPNID *RevoClient::pClientIdents = NULL;
int RevoClient::NrClients = 0;
void (* RevoClient::fp) (BYTE *, int) = NULL;

RevoClient::RevoClient ()
{
	pClient = NULL;
	pAddrClient = NULL;
	pAddrServer = NULL;
	pClientIdents = NULL;
	NrClients = 0;
	if (CoInitCnt == 0)
		CoInitializeEx (NULL, COINIT_MULTITHREADED);
	CoInitCnt++;
}

RevoClient::~RevoClient ()
{
	Cleanup ();
	CoInitCnt--;
	if (CoInitCnt == 0)
		CoUninitialize ();
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert den RevoClient.
// Parameter: Identifier (Zahl, die das Programm identifiziert), Port (der zu
// verwendende Port), fp (Funktions-Pointer auf eine Callback-Funktion mit folgendem
// Typ: void Function (BYTE *, int))
// Bemerkung: Sowohl der Identifier als auch der Port muss beim Server und bei jedem Client der gleiche
// sein, um eine Verbindung herstellen zu knnen.
//-----------------------------------------------------------------------------
bool RevoClient::Initialize (unsigned int Identifier, unsigned int Port, void (* Callback) (BYTE *, int))
{
	Cleanup ();
	//Client erstellen
	if (FAILED (CoCreateInstance (CLSID_DirectPlay8Client, NULL, CLSCTX_ALL, IID_IDirectPlay8Client, (LPVOID *) &pClient)))
		return (false);
	if (FAILED (pClient->Initialize (NULL, ClientMsgHandler, 0)))
		return (false);
	//Client-Adresse erstellen
	if (FAILED (CoCreateInstance (CLSID_DirectPlay8Address, NULL, CLSCTX_ALL, IID_IDirectPlay8Address, (LPVOID *) &pAddrClient)))
		return (false);
	pAddrClient->SetSP (&CLSID_DP8SP_TCPIP);
	//Server-Adresse erstellen
	if (FAILED (CoCreateInstance (CLSID_DirectPlay8Address, NULL, CLSCTX_ALL, IID_IDirectPlay8Address, (LPVOID *) &pAddrServer)))
		return (false);
	pAddrServer->SetSP (&CLSID_DP8SP_TCPIP);
	pAddrServer->AddComponent (DPNA_KEY_PORT, &Port, sizeof (DWORD), DPNA_DATATYPE_DWORD);
	//GUID erstellen
	Ident = GuidMake (Identifier);
	//Anwendungs-Beschreibung erstellen
	ZeroMemory (&AppDesc, sizeof (DPN_APPLICATION_DESC));
	AppDesc.dwSize = sizeof (DPN_APPLICATION_DESC);
	AppDesc.guidApplication = Ident;
	//Funktions-Pointer sichern
	fp = Callback;

	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert den RevoClient.
//-----------------------------------------------------------------------------
void RevoClient::Cleanup ()
{
	int i;

	if (pAddrServer)
	{
		pAddrServer->Release ();
		pAddrServer = NULL;
	}
	if (pAddrClient)
	{
		pAddrClient->Release ();
		pAddrClient = NULL;
	}
	if (pClient)
	{
		pClient->Close (0);
		pClient->Release ();
		pClient = NULL;
	}
	if (pDesc)
	{
		for (i = 0; i < NrIPs; i++)
		{
			delete []pDesc[i].IP;
			delete []pDesc[i].SessionName;
		}
		free (pDesc);
		pDesc = NULL;
	}
	NrIPs = 0;
	if (pClientIdents)
	{
		free (pClientIdents);
		pClientIdents = NULL;
	}
	NrClients = 0;
	Connected = false;
}



//-----------------------------------------------------------------------------
// Funktion: GetServers
// Beschreibung: Listet alle offenen Server fr dieses Spiel auf und ermittelt deren IP-Adressen.
// Parameter: TimeOut (Zeit in Sekunden, bis die Operation gestoppt wird), outServers (die Adresse eines
// Pointers auf ein ServerDesc-Feld, in das die Server geschrieben werden).
// Rckgabewert: Die Anzahl der Server.
//-----------------------------------------------------------------------------
int RevoClient::GetServers (float TimeOut, ServerDesc **outServers)
{
	if (!pClient || !outServers)
		return (0);

	int i;
	DWORD dwTime;
	DPNHANDLE AsyncOp;
	LARGE_INTEGER TimerFrq, TimerBase, TimerCnt;

	//Server-Feld zurcksetzen
	if (pDesc)
	{
		for (i = 0; i < NrIPs; i++)
		{
			delete []pDesc[i].IP;
			delete []pDesc[i].SessionName;
		}
		free (pDesc);
		pDesc = NULL;
	}
	NrIPs = 0;

	if (TimeOut < 0.001)
		TimeOut = 0.001f;
	dwTime = (DWORD) (TimeOut * 1000.0f);
	if (FAILED (pClient->EnumHosts (&AppDesc, pAddrServer, pAddrClient, NULL, 0, 1, 0, dwTime, NULL, &AsyncOp, 0)))
		return (0);

	//Timeout abwarten
	QueryPerformanceFrequency (&TimerFrq);
	QueryPerformanceCounter (&TimerBase);
	do
		QueryPerformanceCounter (&TimerCnt);
	while ((TimerCnt.QuadPart - TimerBase.QuadPart) / TimerFrq.QuadPart < TimeOut);

	*outServers = pDesc;
	return (NrIPs);
}



//-----------------------------------------------------------------------------
// Funktion: Connect
// Beschreibung: Verbindet den Client mit einem Server.
// Parameter: IP (die IP-Adresse des Servers), TimeOut (Zeit in Sekunden, bis die Operation gestoppt wird).
//-----------------------------------------------------------------------------
bool RevoClient::Connect (char *IP, float TimeOut)
{
	if (!pClient || !pAddrServer || Connected || !IP)
		return (false);

	WCHAR *wIP = UnicodeStringMake (IP);
	DPNHANDLE AsyncOp;
	LARGE_INTEGER TimerFrq, TimerBase, TimerCnt;

	pAddrServer->AddComponent (DPNA_KEY_HOSTNAME, wIP, (wcslen (wIP) + 1) * sizeof (WCHAR), DPNA_DATATYPE_STRING);
	delete []wIP;
	Connected = false;
	if (FAILED (pClient->Connect (&AppDesc, pAddrServer, pAddrClient, NULL, NULL, NULL, 0, NULL, &AsyncOp, 0)))
		return (false);

	//Timeout abwarten
	QueryPerformanceFrequency (&TimerFrq);
	QueryPerformanceCounter (&TimerBase);
	do
	{
		if (Connected)
			break;
		QueryPerformanceCounter (&TimerCnt);
	}
	while ((TimerCnt.QuadPart - TimerBase.QuadPart) / TimerFrq.QuadPart < TimeOut);

	if (!Connected)
		pClient->CancelAsyncOperation (AsyncOp, 0);
	return (Connected);
}

//-----------------------------------------------------------------------------
// Funktion: Disconnect
// Beschreibung: Beendet die Verbindung des Clients mit dem Server.
//-----------------------------------------------------------------------------
void RevoClient::Disconnect ()
{
	if (!pClient || !Connected)
		return;
	pClient->Close (0);
	pClient->Release ();
	pClient = NULL;
	Connected = false;
	if (FAILED (CoCreateInstance (CLSID_DirectPlay8Client, NULL, CLSCTX_ALL, IID_IDirectPlay8Client, (LPVOID *) &pClient)))
		return;
	pClient->Initialize (NULL, ClientMsgHandler, 0);
}



//-----------------------------------------------------------------------------
// Funktion: GetPlayers
// Beschreibung: Gibt die Anzahl der Spieler und deren Identifiers zurck.
// Parameter: outIdentifiers (die Adresse eines Pointers auf ein Feld in dem die
// Identifiers fr jeden Spieler stehen)
// Rckgabewert: Die Anzahl der Spieler.
// Bemerkung: Der Server ist stets der erste Identifier. Der lokale Spieler ist
// ebenfalls enthalten.
//-----------------------------------------------------------------------------
int RevoClient::GetPlayers (unsigned long **outIdentifiers)
{
	*outIdentifiers = pClientIdents;
	return (NrClients);
}



//-----------------------------------------------------------------------------
// Funktion: SendData
// Beschreibung: Sendet Daten an den Server oder an einen anderen Client.
// Parameter: DestIdentifier (der Identifier des Zielcomputers, 0 um die Daten an alle zu senden), pData (Pointer auf die
// zu sendenden Daten), DataSize (die Gre der Daten in Bytes, auf die pData zeigt), TimeOut (das TimeOut fr die Nachricht,
// 0 fr kein TimeOut)
//-----------------------------------------------------------------------------
bool RevoClient::SendData (unsigned long DestIdentifier, BYTE *pData, unsigned int DataSize, float TimeOut)
{
	if (!pClient || !Connected || !pData)
		return (false);

	BYTE SendData[sizeof (DWORD) + 3 * sizeof (UINT) + NW_BUFSIZE];
	UINT DataDone, DataLeft;
	BUFFERDESC BufDesc;
	DPNHANDLE AsyncOp;
	DWORD dwTime = (DWORD) (TimeOut * 1000.0f);

	memcpy (SendData, &DestIdentifier, sizeof (DWORD));
	DataDone = 0;
	while (DataDone < DataSize)
	{
		DataLeft = DataSize - DataDone;
		if (DataLeft > NW_BUFSIZE)
			DataLeft = NW_BUFSIZE;
		memcpy (&SendData[sizeof (DWORD)], &DataDone, sizeof (UINT));						//Position im Zielbuffer
		memcpy (&SendData[sizeof (DWORD) + sizeof (UINT)], &DataLeft, sizeof (UINT));		//Lnge des aktuellen Blocks
		memcpy (&SendData[sizeof (DWORD) + 2 * sizeof (UINT)], &DataSize, sizeof (UINT));	//Lnge der gesamten Nachricht
		//Die Daten in Blcken verschicken.
		memcpy (&SendData[sizeof (DWORD) + 3 * sizeof (UINT)], &pData[DataDone], DataLeft);
		BufDesc.dwBufferSize = sizeof (DWORD) + 3 * sizeof (UINT) + DataLeft;
		BufDesc.pBufferData = SendData;
		if (FAILED (pClient->Send (&BufDesc, 1, dwTime, NULL, &AsyncOp, DPNSEND_NOLOOPBACK)))
			return (false);
		DataDone += NW_BUFSIZE;
	}
	return (true);
}



HRESULT WINAPI ClientMsgHandler (void *pvUserContext, DWORD dwMessageId, void *pMsgBuffer)
{
	if (dwMessageId == DPN_MSGID_ENUM_HOSTS_RESPONSE)
	{
		//Ein Server schickt Besttigung auf eine Client-Anfrage, ins Server-Feld eintragen.
		DPNMSG_ENUM_HOSTS_RESPONSE *GotMsg;
		DWORD NrSign = 1024;
		char RdBuf[1024], *pIPs, *pIPe;
		int slen;

		GotMsg = (DPNMSG_ENUM_HOSTS_RESPONSE *) pMsgBuffer;
		GotMsg->pAddressSender->GetURLA (RdBuf, &NrSign);
		//IP schreiben
		pIPs = strstr (RdBuf, "hostname=");
		pIPe = strstr (RdBuf, "port=");
		slen = pIPe - pIPs - 9;
		if (pDesc)
			pDesc = (ServerDesc *) realloc (pDesc, sizeof (ServerDesc) * (NrIPs + 1));
		else
			pDesc = (ServerDesc *) malloc (sizeof (ServerDesc));
		pDesc[NrIPs].IP = new char[slen];
		strncpy (pDesc[NrIPs].IP, pIPs + 9, slen - 1);
		pDesc[NrIPs].IP[slen - 1] = '\0';
		//Session-Name schreiben
		pDesc[NrIPs].SessionName = UnicodeStringTranslate (GotMsg->pApplicationDescription->pwszSessionName);
		NrIPs++;
	}
	else if (dwMessageId == DPN_MSGID_CONNECT_COMPLETE)
	{
		//Ein Verbindungsaufbau wird vom Server beantwortet.
		DPNMSG_CONNECT_COMPLETE *GotMsg;

		GotMsg = (DPNMSG_CONNECT_COMPLETE *) pMsgBuffer;
		if (SUCCEEDED (GotMsg->hResultCode))
			Connected = true;
	}
	else if (dwMessageId == DPN_MSGID_TERMINATE_SESSION)
		//Die Verbindung wurde unterbrochen oder beendet, dies dem User mitteilen.
		RevoClient::fp (NULL, 0);
	else if (dwMessageId == DPN_MSGID_RECEIVE)
	{
		//Es wurden beliebige Daten empfangen.
		int i;
		DPNMSG_RECEIVE *GotMsg;
		DWORD Offset = 1;
		UINT BlockPos, BlockLen, MsgLen;
		BYTE MsgType;

		GotMsg = (DPNMSG_RECEIVE *) pMsgBuffer;
		//DataSize = GotMsg->dwReceiveDataSize;
		MsgType = GotMsg->pReceiveData[0];
		if (MsgType == MSGTYPE_PLAYERLIST)
		{
			//Liste der Spieler wurde vom Server geschickt, d.h. lokal am Client neu anlegen
			memcpy (&RevoClient::NrClients, &GotMsg->pReceiveData[1], sizeof (int));
			Offset += sizeof (int);
			if (RevoClient::pClientIdents)
				free (RevoClient::pClientIdents);
			RevoClient::pClientIdents = (DPNID *) malloc (sizeof (DPNID) * RevoClient::NrClients);
			for (i = 0; i < RevoClient::NrClients; i++)
			{
				memcpy (&RevoClient::pClientIdents[i], &GotMsg->pReceiveData[Offset], sizeof (DPNID));
				Offset += sizeof (DPNID);
			}
		}
		else if (MsgType == MSGTYPE_DATA)
		{
			//Beliebige Daten wurden empfangen.
			//Die Nachricht zusammenfassen, bei Vollstndigkeit Callback-Funktion aufrufen.
			memcpy (&BlockPos, &GotMsg->pReceiveData[1], sizeof (UINT));
			memcpy (&BlockLen, &GotMsg->pReceiveData[1 + sizeof (UINT)], sizeof (UINT));
			memcpy (&MsgLen, &GotMsg->pReceiveData[1 + 2 * sizeof (UINT)], sizeof (UINT));
			memcpy (&ReadBuf[BlockPos], &GotMsg->pReceiveData[1 + 3 * sizeof (UINT)], BlockLen);
			if (BlockPos == 0)
				//Es beginnt bereits eine neue Nachricht, die vorige ist offenbar "timed-out"
				ReadSize = 0;
			ReadSize += BlockLen;
			if (ReadSize == MsgLen)
			{
				//Nachricht ist vollstndig.
				RevoClient::fp (ReadBuf, ReadSize);
				ReadSize = 0;
			}
		}
	}
	return (S_OK);
}










//Statische Variablen initsialisieren
IDirectPlay8Server *RevoServer::pServer = NULL;
bool RevoServer::AllowClientConn = true;
DPNID *RevoServer::pClientIdents = NULL;
int RevoServer::NrClients = 0;
void (* RevoServer::fp) (BYTE *, int) = NULL;

RevoServer::RevoServer ()
{
	pServer = NULL;
	pAddrServer = NULL;
	pClientIdents = NULL;
	NrClients = 0;
	if (CoInitCnt == 0)
		CoInitializeEx (NULL, COINIT_MULTITHREADED);
	CoInitCnt++;
}

RevoServer::~RevoServer ()
{
	Cleanup ();
	CoInitCnt--;
	if (CoInitCnt == 0)
		CoUninitialize ();
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert den RevoServer.
// Parameter: Identifier (Zahl, die das Programm identifiziert), Port (der zu
// verwendende Port), fp (Funktions-Pointer auf eine Callback-Funktion mit folgendem
// Typ: void Function (BYTE *, int))
// Bemerkung: Sowohl der Identifier als auch der Port muss beim Server und bei jedem Client der gleiche
// sein, um eine Verbindung herstellen zu knnen.
//-----------------------------------------------------------------------------
bool RevoServer::Initialize (unsigned int Identifier, unsigned int Port, void (* Callback) (BYTE *, int))
{
	DWORD dwPort = Port;

	Cleanup ();
	//Server erstellen
	if (FAILED (CoCreateInstance (CLSID_DirectPlay8Server, NULL, CLSCTX_ALL, IID_IDirectPlay8Server, (LPVOID *) &pServer)))
		return (false);
	if (FAILED (pServer->Initialize (NULL, ServerMsgHandler, 0)))
		return (false);
	//Server-Adresse erstellen
	if (FAILED (CoCreateInstance (CLSID_DirectPlay8Address, NULL, CLSCTX_ALL, IID_IDirectPlay8Address, (LPVOID *) &pAddrServer)))
		return (false);
	pAddrServer->SetSP (&CLSID_DP8SP_TCPIP);
	pAddrServer->AddComponent (DPNA_KEY_PORT, &dwPort, sizeof (DWORD), DPNA_DATATYPE_DWORD);
	//GUID erstellen
	Ident = GuidMake (Identifier);	
	//Anwendungs-Beschreibung erstellen
	ZeroMemory (&AppDesc, sizeof (DPN_APPLICATION_DESC));
	AppDesc.dwSize = sizeof (DPN_APPLICATION_DESC);
	AppDesc.guidApplication = Ident;
	AppDesc.dwFlags = DPNSESSION_CLIENT_SERVER;
	//Allgemeine Variablen setzen
	AllowClientConn = true;
	fp = Callback;

	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert den RevoServer.
//-----------------------------------------------------------------------------
void RevoServer::Cleanup ()
{
	if (pAddrServer)
	{
		pAddrServer->Release ();
		pAddrServer = NULL;
	}
	if (pServer)
	{
		pServer->Close (0);
		pServer->Release ();
		pServer = NULL;
	}
	if (pClientIdents)
	{
		free (pClientIdents);
		pClientIdents = NULL;
	}
	NrClients = 0;
}



//-----------------------------------------------------------------------------
// Funktion: Host
// Beschreibung: Erffnet eine Session bzw. startet den Server.
// Parameter: SessionName (Name der Session), MaxClients (die maximale Anzahl an Clients, die sich
// auf diesen Server verbinden knnen, 0 fr unbegrenzt).
//-----------------------------------------------------------------------------
bool RevoServer::Host (char *SessionName, int MaxClients)
{
	if (!pServer || !SessionName)
		return (false);

	WCHAR *wSession = UnicodeStringMake (SessionName);
	HRESULT hr;

	AppDesc.pwszSessionName = wSession;
	if (MaxClients > 0)
		AppDesc.dwMaxPlayers = MaxClients + 1;
	else
		AppDesc.dwMaxPlayers = 0;
	hr = pServer->Host (&AppDesc, &pAddrServer, 1, NULL, NULL, NULL, 0);
	delete []wSession;
	if (FAILED (hr))
		return (false);
	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: Unhost
// Beschreibung: Schliet die aktuelle Session mit allen zugehrigen Clients.
//-----------------------------------------------------------------------------
void RevoServer::Unhost ()
{
	if (!pServer)
		return;
	pServer->Close (0);
	pServer->Release ();
	pServer = NULL;
	if (FAILED (CoCreateInstance (CLSID_DirectPlay8Server, NULL, CLSCTX_ALL, IID_IDirectPlay8Server, (LPVOID *) &pServer)))
		return;
	pServer->Initialize (NULL, ServerMsgHandler, 0);
}



//-----------------------------------------------------------------------------
// Funktion: SetConnectionState
// Beschreibung: Legt fest, ob sich Clients mit diesem Server verbinden knnen.
// Parameter: AllowClientConnection (der Status)
//-----------------------------------------------------------------------------
void RevoServer::SetConnectionState (bool AllowClientConnection)
{
	AllowClientConn = AllowClientConnection;
}



//-----------------------------------------------------------------------------
// Funktion: GetPlayers
// Beschreibung: Gibt die Anzahl der Spieler und deren Identifiers zurck.
// Parameter: outIdentifiers (die Adresse eines Pointers auf ein Feld in dem die
// Identifiers fr jeden Spieler stehen)
// Rckgabewert: Die Anzahl der Spieler.
// Bemerkung: Der Server ist stets der erste Identifier.
//-----------------------------------------------------------------------------
int RevoServer::GetPlayers (unsigned long **outIdentifiers)
{
	*outIdentifiers = pClientIdents;
	return (NrClients);
}



//-----------------------------------------------------------------------------
// Funktion: SendData
// Beschreibung: Sendet Daten an einen beliebigen Client.
// Parameter: DestIdentifier (der Identifier des Zielclients, 0 um die Daten an alle zu senden), pData (Pointer auf die
// zu sendenden Daten), DataSize (die Gre der Daten in Bytes, auf die pData zeigt), TimeOut (das TimeOut fr die Nachricht,
// 0 fr kein TimeOut)
//-----------------------------------------------------------------------------
bool RevoServer::SendData (unsigned long DestIdentifier, BYTE *pData, unsigned int DataSize, float TimeOut)
{
	if (!pServer || !pData)
		return (false);

	BYTE SendData[1 + 3 * sizeof (UINT) + NW_BUFSIZE];
	UINT DataDone, DataLeft;
	BUFFERDESC BufDesc;
	DPNHANDLE AsyncOp;
	DWORD dwTime = (DWORD) (TimeOut * 1000.0f);

	SendData[0] = MSGTYPE_DATA;
	DataDone = 0;
	while (DataDone < DataSize)
	{
		DataLeft = DataSize - DataDone;
		if (DataLeft > NW_BUFSIZE)
			DataLeft = NW_BUFSIZE;
		memcpy (&SendData[1], &DataDone, sizeof (UINT));						//Position im Zielbuffer
		memcpy (&SendData[1 + sizeof (UINT)], &DataLeft, sizeof (UINT));		//Lnge des aktuellen Blocks
		memcpy (&SendData[1 + 2 * sizeof (UINT)], &DataSize, sizeof (UINT));	//Lnge der gesamten Nachricht
		//Die Daten in Blcken verschicken.
		memcpy (&SendData[1 + 3 * sizeof (UINT)], &pData[DataDone], DataLeft);
		BufDesc.dwBufferSize = 1 + 3 * sizeof (UINT) + DataLeft;
		BufDesc.pBufferData = SendData;
		if (FAILED (pServer->SendTo (DestIdentifier, &BufDesc, 1, dwTime, NULL, &AsyncOp, DPNSEND_NOLOOPBACK)))
			return (false);		
		DataDone += NW_BUFSIZE;
	}
	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: GetQueueSize
// Beschreibung: Gibt zurck, wie viele Nachrichten an den jeweiligen Client noch gesendet werden sollen
// und bis jetzt aber noch nicht gesendet wurden.
// Parameter: DestIdentifier (der Identifier des Zielclients)
//-----------------------------------------------------------------------------
int RevoServer::GetQueueSize (unsigned long DestIdentifier)
{
	DWORD NumMsg;

	pServer->GetSendQueueInfo (DestIdentifier, &NumMsg, NULL, 0);
	return (NumMsg);
}



void RevoServer::AddPlayer (DPNID Ident)
{
	if (pClientIdents)
		pClientIdents = (DPNID *) realloc (pClientIdents, sizeof (DPNID) * (NrClients + 1));
	else
		pClientIdents = (DPNID *) malloc (sizeof (DPNID));
	pClientIdents[NrClients] = Ident;
	NrClients++;
	SendPlayerList ();
}

void RevoServer::DeletePlayer (DPNID Ident)
{
	int i, j;
	
	for (i = 0; i < NrClients; i++)
	{
		if (pClientIdents[i] == Ident)
		{
			for (j = i; j < NrClients - 1; j++)
				pClientIdents[j] = pClientIdents[j + 1];
			break;
		}
	}
	NrClients--;
	pClientIdents = (DPNID *) realloc (pClientIdents, sizeof (DPNID) * NrClients);
	SendPlayerList ();
}

void RevoServer::SendPlayerList ()
{
	if (NrClients <= 1)
		return;

	int i;
	BUFFERDESC BufDesc;
	DPNHANDLE AsyncOp;
	BYTE *Buffer, *bp;
	UINT BufSize = 1 + sizeof (int) + NrClients * sizeof (DPNID);

	//Den Clients die Liste der Player neu schicken
	Buffer = new BYTE[BufSize];
	Buffer[0] = MSGTYPE_PLAYERLIST;
	bp = Buffer + 1;
	memcpy (bp, &NrClients, sizeof (int));
	bp += sizeof (int);
	for (i = 0; i < NrClients; i++)
	{
		memcpy (bp, &pClientIdents[i], sizeof (DPNID));
		bp += sizeof (DPNID);
	}
	BufDesc.dwBufferSize = BufSize;
	BufDesc.pBufferData = Buffer;
	pServer->SendTo (DPNID_ALL_PLAYERS_GROUP, &BufDesc, 1, 0, NULL, &AsyncOp, DPNSEND_NOLOOPBACK | DPNSEND_GUARANTEED);
	delete []Buffer;
}



HRESULT WINAPI ServerMsgHandler (void *pvUserContext, DWORD dwMessageId, void *pMsgBuffer)
{
	if (dwMessageId == DPN_MSGID_INDICATE_CONNECT)
	{
		//Client will sich einloggen
		if (RevoServer::AllowClientConn)
			return (S_OK);
		else
			return (E_FAIL);
	}
	else if (dwMessageId == DPN_MSGID_CREATE_PLAYER)
	{
		//Client hat sich erfolgreich verbunden, in die lokale Liste eintragen
		DPNMSG_CREATE_PLAYER *GotMsg;

		GotMsg = (DPNMSG_CREATE_PLAYER *) pMsgBuffer;
		RevoServer::AddPlayer (GotMsg->dpnidPlayer);
	}
	else if (dwMessageId == DPN_MSGID_DESTROY_PLAYER)
	{
		//Client beendet die Verbindung, aus der lokalen Liste austragen
		DPNMSG_DESTROY_PLAYER *GotMsg;

		GotMsg = (DPNMSG_DESTROY_PLAYER *) pMsgBuffer;
		RevoServer::DeletePlayer (GotMsg->dpnidPlayer);
		//Dies dem User mitteilen.
		RevoServer::fp (NULL, 0);
	}
	else if (dwMessageId == DPN_MSGID_RECEIVE)
	{
		//Daten wurden empfangen
		DPNMSG_RECEIVE *GotMsg;
		DWORD DestIdent;
		UINT BlockPos, BlockLen, MsgLen;

		GotMsg = (DPNMSG_RECEIVE *) pMsgBuffer;
		memcpy (&DestIdent, GotMsg->pReceiveData, sizeof (DWORD));
		memcpy (&BlockPos, &GotMsg->pReceiveData[sizeof (DWORD)], sizeof (UINT));
		memcpy (&BlockLen, &GotMsg->pReceiveData[sizeof (DWORD) + sizeof (UINT)], sizeof (UINT));
		memcpy (&MsgLen, &GotMsg->pReceiveData[sizeof (DWORD) + 2 * sizeof (UINT)], sizeof (UINT));
		
		memcpy (&ReadBuf[BlockPos], &GotMsg->pReceiveData[sizeof (DWORD) + 3 * sizeof (UINT)], BlockLen);
		if (BlockPos == 0)
			//Es beginnt bereits eine neue Nachricht, die vorige ist offenbar "timed-out"
			ReadSize = 0;
		ReadSize += BlockLen;
		if (ReadSize == MsgLen)
		{
			//Nachricht ist vollstndig.
			if (DestIdent != RevoServer::pClientIdents[0])
				//Daten weiterleiten
				RevoServer::SendData (DestIdent, ReadBuf, ReadSize, 0);
			if (DestIdent == 0 || DestIdent == RevoServer::pClientIdents[0])
				//Daten sind (auch) fr den Server selbst bestimmt
				RevoServer::fp (ReadBuf, ReadSize);
			ReadSize = 0;
		}					
	}
	return (S_OK);
}