#include <windows.h>
#include <stdio.h>
#include <time.h>
#include "revo3d.h"
#include "tron.h"


//REVO TRON VERSION 1.4
//Language: English


//3d Engine
Revo3dEngine Eng3d;
Revo3dCaps RdCaps;
int ResCnt, ActRes;
Revo3dRes *pRes;
Revo3dObject Shield;
RevoInput Inp;

//Sounds
RevoSoundEngine EngSound;
Revo3dSound BikeCrashSound, ShotCrashSound, FireShotSound;


//Multiplayer
RevoServer GameServer;
RevoClient GameClient;
TronMPSettings MPOptions;
TronMPBikeIdent MPBikeInfo[6];
int ClientReady;
bool FreeToRender;
//Steuervariablen zum Senden und Empfangen
BYTE RData[131072], SData[131072];				//128 KByte Buffer
float TimeOut;
ActionBuffer MPBuf;
RenderBuffer RBuf;



//TRON game variables
int PIdent, MopedAlive, LeaderScore;
int CamMode;
float NotRunTime, NotAliveTime, BlendTime, TimeFac;
bool GameRuns, GamePassed, MPGameBreak;
TronExplosion Explosions;

TronPlayer *pPlayer;
TronShots Shots;

TronSettings Options;


int WINAPI WinMain (HINSTANCE hInst, HINSTANCE, LPSTR, int)
{
	Revo2dTexture CursorTex, BusyTex, ButtonTex, BackgrTex;
	Revo2dText MenuText;
	FILE *fp;

	//Einstellungen von Datei laden
	if (!(fp = fopen ("settings\\config.dat", "r")))
	{
		if (!(fp = fopen ("settings\\default.dat", "r")))
		{
			MessageBox (NULL, "Couldn't load the game. Revo Tron will exit now.", "Revo Tron", MB_OK | MB_ICONERROR);
			return (0);
		}
	}
	fread (&Options, sizeof (TronSettings), 1, fp);
	fclose (fp);

	InitRandom ();
	if (!OneTimeDXInit (hInst, Options.Res.ResX, Options.Res.ResY, Options.Res.Format, Options.UseFSAA))
		return (0);

	//Multiplayer-Objekte initsialisieren
	GameServer.Initialize (TRON_MPVERS, 20000, SCallback);
	GameClient.Initialize (TRON_MPVERS, 20000, CCallback);

	//Objekte frs Men laden
	CursorTex.Initialize (&Eng3d);
	if (!CursorTex.LoadFromFile ("textures\\cursor.bmp", ColorMake (0, 0, 0), 1, false))
		return (0);
	BusyTex.Initialize (&Eng3d);
	if (!BusyTex.LoadFromFile ("textures\\busy.bmp", ColorMake (0, 0, 0), 1, false))
		return (0);
	ButtonTex.Initialize (&Eng3d);
	if (!ButtonTex.LoadFromFile ("textures\\button.bmp", 0, 1, false))
		return (0);
	BackgrTex.Initialize (&Eng3d);
	BackgrTex.LoadFromFile ("textures\\background.bmp", 0, 1, false);
	MenuText.Initialize (&Eng3d);
	MenuText.SetFormat (1, 1);
	MenuText.SetFont (FontMake ("Times New Roman", 20, false, false, false));

	ShowMenu (MENU_MAIN, Vector2dMake (Options.Res.ResX / 2.0f, Options.Res.ResY / 2.0f), &CursorTex, &BusyTex, &ButtonTex, &MenuText, &BackgrTex);

	//Einstellungen wieder auf Datei schreiben
	if (fp = fopen ("settings\\config.dat", "w"))
	{
		fwrite (&Options, sizeof (TronSettings), 1, fp);
		fclose (fp);
	}

	return (0);
}


bool OneTimeDXInit (HINSTANCE hInst, int ResX, int ResY, D3DFORMAT Format, int AntiAlias)
{
	int i;
	bool FoundRes;

	//*****DIRECT 3D*****
	if (!Eng3d.Initialize (0, "Revo Tron"))
	{
		ShowWindow (Eng3d.GethWnd (), SW_HIDE);
		MessageBox (NULL, "Couldn't find an usable 3d graphics card. Revo Tron will exit now.", "Revo Tron", MB_OK | MB_ICONERROR);
		return (false);
	}

	RdCaps = Eng3d.GetPossibleCaps (Format, false);
	if (RdCaps.MaxTextureSize.x < 256 || RdCaps.MaxTextureSize.y < 256)
	{
		ShowWindow (Eng3d.GethWnd (), SW_HIDE);
		MessageBox (NULL, "Couldn't find an usable 3d graphics card. Revo Tron will exit now.", "Revo Tron", MB_OK | MB_ICONERROR);
		return (false);
	}

	ResCnt = Eng3d.GetPossibleRes (false, false, true, &pRes);
	if (ResCnt == 0)
	{
		ShowWindow (Eng3d.GethWnd (), SW_HIDE);
		MessageBox (NULL, "Couldn't find an usable 3d graphics card. Revo Tron will exit now.", "Revo Tron", MB_OK | MB_ICONERROR);
		return (false);
	}

	//Suchen, ob die einzustellende Auflsung auch wirklich vorhanden ist
	FoundRes = false;
	for (i = 0; i < ResCnt; i++)
	{
		if (ResX == pRes[i].ResX && ResY == pRes[i].ResY && Format == pRes[i].Format && AntiAlias <= RdCaps.MaxAntiAlias)
		{
			FoundRes = true;
			break;
		}
	}
	if (FoundRes)
	{
		Options.Res = pRes[i];
		Options.UseFSAA = AntiAlias;
		ActRes = i;
	}
	else
	{
		Options.Res = pRes[0];
		Options.UseFSAA = 0;
		ActRes = 0;
	}


	//*****DIRECT SOUND*****
	if (!EngSound.Initialize (Eng3d.GethWnd ()))
	{
		ShowWindow (Eng3d.GethWnd (), SW_HIDE);
		MessageBox (NULL, "Couldn't find an usable sound card. There will be no sound output.", "Revo Tron", MB_OK | MB_ICONINFORMATION);
		ShowWindow (Eng3d.GethWnd (), SW_SHOW);
	}


	//*****DIRECT INPUT*****
	//Zum Debuggen hier die Parameter ndern!
	if (!Inp.Initialize (Eng3d.GethWnd (), hInst, true, true))
	{
		ShowWindow (Eng3d.GethWnd (), SW_HIDE);
		MessageBox (NULL, "Couldn't get access to the keyboard. Revo Tron will exit now.", "Revo Tron", MB_OK | MB_ICONERROR);
		return (false);
	}


	//Auflsung setzen
	if (!Eng3d.SetRes (&Options.Res, Options.UseFSAA, false))
	{
		Inp.Cleanup ();
		ShowWindow (Eng3d.GethWnd (), SW_HIDE);
		MessageBox (NULL, "Couldn't change the screen resolution. Revo Tron will exit now.", "Revo Tron", MB_OK | MB_ICONERROR);
		return (false);
	}

	return (true);
}

void GeometryInit (float SizeX, float SizeZ, float WallHeight, Vertex *outV)
{
	float tuX = (float) ((int) SizeX / 75);
	float tuZ = (float) ((int) SizeZ / 75);

	if (tuX == 0)
		tuX = 1;
	if (tuZ == 0)
		tuZ = 1;
	//Boden
	outV[0] = VertexMake (0, 0, SizeZ, 0, 1, 0, 0, 0);
	outV[1] = VertexMake (SizeX, 0, SizeZ, 0, 1, 0, SizeX / 5, 0);
	outV[2] = VertexMake (0, 0, 0, 0, 1, 0, 0, SizeZ / 5);
	outV[3] = VertexMake (SizeX, 0, 0, 0, 1, 0, SizeX / 5, SizeZ / 5);
	//Vorderwand
	outV[4] = VertexMake (SizeX, WallHeight, 0, 0, 0, 1, 0, 0);
	outV[5] = VertexMake (0, WallHeight, 0, 0, 0, 1, tuX, 0);
	outV[6] = VertexMake (SizeX, 0, 0, 0, 0, 1, 0, 0.25);
	outV[7] = VertexMake (0, 0, 0, 0, 0, 1, tuX, 0.25);
	//Rckwand
	outV[8] = VertexMake (0, WallHeight, SizeZ, 0, 0, -1, 0, 0.25);
	outV[9] = VertexMake (SizeX, WallHeight, SizeZ, 0, 0, -1, tuX, 0.25);
	outV[10] = VertexMake (0, 0, SizeZ, 0, 0, -1, 0, 0.5);
	outV[11] = VertexMake (SizeX, 0, SizeZ, 0, 0, -1, tuX, 0.5);
	//Linke Wand
	outV[12] = VertexMake (0, WallHeight, 0, 1, 0, 0, 0, 0.5);
	outV[13] = VertexMake (0, WallHeight, SizeZ, 1, 0, 0, tuZ, 0.5);
	outV[14] = VertexMake (0, 0, 0, 1, 0, 0, 0, 0.75);
	outV[15] = VertexMake (0, 0, SizeZ, 1, 0, 0, tuZ, 0.75);
	//Rechte Wand
	outV[16] = VertexMake (SizeX, WallHeight, SizeZ, -1, 0, 0, 0, 0.5);
	outV[17] = VertexMake (SizeX, WallHeight, 0, -1, 0, 0, tuZ, 0.5);
	outV[18] = VertexMake (SizeX, 0, SizeZ, -1, 0, 0, 0, 0.75);
	outV[19] = VertexMake (SizeX, 0, 0, -1, 0, 0, tuZ, 0.75);
}


void ShowMenu (int MenuID, D3DXVECTOR2 CursPos, Revo2dTexture *pCurs, Revo2dTexture *pBusy, Revo2dTexture *pButton, Revo2dText *pText, Revo2dTexture *pBackgr)
{
	int i;
	bool LDown = false, RDown = false, ESCDown = false, RetDown = false;
	bool Busy, EditName = false;
	D3DXVECTOR2 CursorPos = CursPos, ButtonPosL, ButtonPosR;
	DIMOUSESTATE *RdMouse;
	BYTE *RdKeyb;
	int MenuCnt, MenuHl, FadeIn;
	char MenuStr[MAX_MENUITEM][MAX_MENUTEXT];
	Revo2dPicture MenuCursor, MenuButton, MenuBackgr;
	int PrevRes, ActFSAA;

	ServerDesc *pServerDesc;
	int ServerCnt;
	float WaitTime;
	TronSettings SaveSettings;

	PrevRes = ActRes;
	ActFSAA = Options.UseFSAA;
	if (Options.PlayerName[0] == '\0')
		strcpy (Options.PlayerName, "Player");
	if (MenuID == MENU_MULTIPLAYER_CLIENT)
	{
		//Nach Servern suchen
		ServerCnt = GameClient.GetServers (3, &pServerDesc);
		if (ServerCnt == 0)
		{
			strcpy (&MenuStr[0][0], "No hosts found");
			strcpy (&MenuStr[1][0], "Back");
			MenuCnt = 2;
		}
		else
		{
			if (ServerCnt > MAX_MENUITEM - 1)
				ServerCnt = MAX_MENUITEM - 1;
			for (i = 0; i < ServerCnt; i++)
				sprintf (&MenuStr[i][0], "%s's game on host %s", pServerDesc[i].SessionName, pServerDesc[i].IP);
			strcpy (&MenuStr[ServerCnt][0], "Back");
			MenuCnt = ServerCnt + 1;
		}
	}
	else
		MenuCnt = CreateMenuText (MenuID, MenuStr, &pRes[ActRes], &ActFSAA, EditName);

	MenuButton.SetTransparency (125);

	while (true)
	{
		//Das Server-Men stndig aktualisieren (weil Clients sich inzwischen verbinden knnten)
		if (MenuID == MENU_MULTIPLAYER_SERVER)
		{
			//Clients anzeigen
			if (MPOptions.PlayerCnt + MPOptions.AIPlayerCnt > 4 || MPOptions.AIPlayerCnt > 2)
				MPOptions.AIPlayerCnt--;
			MPOptions.PlayerCnt = GameServer.GetPlayers (&MPOptions.pIdents);
			sprintf (&MenuStr[0][0], "Current human player count: %d", MPOptions.PlayerCnt);
			sprintf (&MenuStr[1][0], "Number of AI players: %d", MPOptions.AIPlayerCnt);
			if (Options.OpponentSkill == 0)
				strcpy (&MenuStr[2][0], "AI player skill: bad");
			else if (Options.OpponentSkill == 1)
				strcpy (&MenuStr[2][0], "AI player skill: medium");
			else if (Options.OpponentSkill == 2)
				strcpy (&MenuStr[2][0], "AI player skill: good");
			if (MPOptions.PlayerCnt > 1)
				strcpy (&MenuStr[3][0], "Start multiplayer game");
			else
				strcpy (&MenuStr[3][0], "Start multiplayer game - N/A");
			strcpy (&MenuStr[4][0], "Back");
			MenuCnt = 5;
		}
		//Beim Client-Men mglicherweise das Spiel starten
		if (MenuID == MENU_MULTIPLAYER_CLIENTWAIT && RData[0] == CLIENT_SETTINGS)
		{
			//**********CLIENT_SETTINGS empfangen*********
			MPOptions.PlayerCnt = GameClient.GetPlayers (&MPOptions.pIdents);
			//Die lokalen Gameplay-Optionen denen des Servers anpassen
			SaveSettings = Options;
			memcpy (&Options, &RData[1], sizeof (TronSettings));
			Options.Cam = SaveSettings.Cam;
			strcpy (Options.PlayerName, SaveSettings.PlayerName);
			Options.Res = SaveSettings.Res;
			Options.UseAnisotropic = SaveSettings.UseAnisotropic;
			Options.UseFSAA = SaveSettings.UseFSAA;
			Options.UseSounds = SaveSettings.UseSounds;
			Options.UseMusic = SaveSettings.UseMusic;
			Options.UseReflection = SaveSettings.UseReflection;
			Options.UseShadow = SaveSettings.UseShadow;
			Options.UseSpecular = SaveSettings.UseSpecular;
			Options.UseTransWalls = SaveSettings.UseTransWalls;
			for (i = 0; i < Options.OpponentCnt + 1; i++)					//Die Moped-Farben vom Server bernehmen
				memcpy (&MPBikeInfo[i].BikeType, &RData[1 + sizeof (TronSettings) + i * sizeof (int)], sizeof (int));
			//Client-IdentNummer speichern, die vom Server zugewiesen wird.
			memcpy (&PIdent, &RData[1 + sizeof (TronSettings) + (Options.OpponentCnt + 1) * sizeof (int)], sizeof (int));

			//**********SERVER_SETTINGSOK senden*********
			SData[0] = SERVER_SETTINGSOK;
			memcpy (&SData[1], &PIdent, sizeof (int));		//Client-IdentNummer
			memcpy (&SData[1 + sizeof (int)], Options.PlayerName, 64 * sizeof (char));
			GameClient.SendData (MPOptions.pIdents[0], SData, 1 + sizeof (int) + 64 * sizeof (char), 0);

			//**********auf CLIENT_LOAD warten*********
			Eng3d.StartTimer ();
			WaitTime = Eng3d.GetAbsTimeDiff ();
			while (RData[0] != CLIENT_LOAD)
			{
				Eng3d.HandleMessages ();
				if (Eng3d.GetAbsTimeDiff () - WaitTime > 3)
					return;
			}

			//Die Spielernamen werden in der Callback-Routine von CLIENT_LOAD ausgelesen!
			RunGame (true, false);

			//Lokale Optionen nach dem Multiplayer-Spiel wiederherstellen
			Options = SaveSettings;
			RData[0] = CONNECTION_OK;

			//Zum vorigen Men zurckkehren
			break;
		}

		//Maus-Cursor aktualisieren
		RdMouse = Inp.ReadMouse ();
		if (!RdMouse)
			break;
		if (EditName)
		{
			RdKeyb = Inp.ReadKeyboard (Options.PlayerName);
			if (strlen (Options.PlayerName) > 16)
				Options.PlayerName[16] = '\0';
			CreateMenuText (MenuID, MenuStr, NULL, NULL, true);
		}
		else
			RdKeyb = Inp.ReadKeyboard (NULL);
		if (!RdKeyb)
			break;
		if (RdMouse->lX || RdMouse->lY)
		{
			LDown = false;
			RDown = false;
			RetDown = false;			
		}
		CursorPos.x += RdMouse->lX * 1.5f;
		CursorPos.y += RdMouse->lY * 1.5f;
		if (CursorPos.x < 0)
			CursorPos.x = 0;
		if (CursorPos.x > (float) Options.Res.ResX)
			CursorPos.x = (float) Options.Res.ResX;
		if (CursorPos.y < 0)
			CursorPos.y = 0;
		if (CursorPos.y > (float) Options.Res.ResY)
			CursorPos.y = (float) Options.Res.ResY;
		MenuCursor.SetPosition ((int) CursorPos.x, (int) CursorPos.y, (int) CursorPos.x + 20, (int) CursorPos.y + 20);

		//Men-Markierung feststellen
		MenuHl = -1;
		for (i = 0; i < MenuCnt; i++)
		{
			if (CursorPos.x >= Options.Res.ResX / 2.0f - MENUITEM_WIDTH / 2.0f && CursorPos.x <= Options.Res.ResX / 2.0f + MENUITEM_WIDTH / 2.0f)
			{
				if (CursorPos.y >= Options.Res.ResY / 2.0f - (MenuCnt / 2.0f - i) * MENUITEM_HEIGHT && CursorPos.y <= Options.Res.ResY / 2.0f - (MenuCnt / 2.0f - i - 1) * MENUITEM_HEIGHT)
					MenuHl = i;
			}
		}
		//Abfragen auf Mausklick (bzw. ESC oder Enter) wegen der Sanduhr
		if (RdKeyb[DIK_ESCAPE] && MenuID != MENU_MAIN)
			ESCDown = true;
		if (RdKeyb[DIK_RETURN] || RdKeyb[DIK_NUMPADENTER])
			RetDown = true;
		if (RdMouse->rgbButtons[0])
			LDown = true;
		if (RdMouse->rgbButtons[1])
			RDown = true;
		if (((!RdMouse->rgbButtons[0] && LDown) || (!RdMouse->rgbButtons[1] && RDown) || (!RdKeyb[DIK_RETURN] && !RdKeyb[DIK_NUMPADENTER] && RetDown)) && MenuHl != -1)
		{
			if (MenuID == MENU_MAIN && MenuHl == MenuCnt - 1)
				Busy = true;
			else if (MenuID == MENU_MAIN && MenuHl == 0)
				Busy = true;
			else if (MenuID == MENU_MULTIPLAYER && MenuHl != MenuCnt - 1)
				Busy = true;
			else if (MenuID == MENU_MULTIPLAYER_SERVER && MenuHl != 0)
				Busy = true;
			else if (MenuID == MENU_MULTIPLAYER_CLIENT)
				Busy = true;
			else if (MenuID == MENU_MULTIPLAYER_CLIENTWAIT && MenuHl == 1)
				Busy = true;
			else if (MenuID == MENU_OPTIONS_GRAFICS && MenuHl == 3)
				Busy = true;
		}
		else
			Busy = false;

		MenuBackgr.SetPosition (0, 0, Options.Res.ResX, Options.Res.ResY);

		//Rendern
		Eng3d.RenderStart (NULL, NULL);
		Eng3d.RenderClear (true, ColorMake (0, 0, 255), false, false);
		Eng3d.Render2dPicture (&MenuBackgr, pBackgr);
		for (i = 0; i < MenuCnt; i++)
		{
			if (i == MenuHl)
				FadeIn = 16;
			else
				FadeIn = 32;
			ButtonPosL.x = Options.Res.ResX / 2.0f - MENUITEM_WIDTH / 2.0f;
			ButtonPosL.y = Options.Res.ResY / 2.0f - (MenuCnt / 2.0f - i) * MENUITEM_HEIGHT;
			ButtonPosR.x = ButtonPosL.x + MENUITEM_WIDTH;
			ButtonPosR.y = ButtonPosL.y + MENUITEM_HEIGHT;

			//Links
			MenuButton.SetPosition ((int) ButtonPosL.x, (int) ButtonPosL.y, (int) ButtonPosL.x + FadeIn, (int) ButtonPosR.y);
			MenuButton.SetPictureRange (0.01f, 0.51f, 0.49f, 0.99f);
			Eng3d.Render2dPicture (&MenuButton, pButton);
			//bergang links
			MenuButton.SetPosition ((int) ButtonPosL.x + FadeIn, (int) ButtonPosL.y, (int) ButtonPosL.x + FadeIn + 32, (int) ButtonPosR.y);
			MenuButton.SetPictureRange (0.01f, 0.01f, 0.49f, 0.49f);
			Eng3d.Render2dPicture (&MenuButton, pButton);
			//Mitte
			MenuButton.SetPosition ((int) ButtonPosL.x + FadeIn + 32, (int) ButtonPosL.y, (int) ButtonPosR.x - FadeIn - 32, (int) ButtonPosR.y);
			MenuButton.SetPictureRange (0.51f, 0.51f, 0.99f, 0.99f);
			Eng3d.Render2dPicture (&MenuButton, pButton);
			//bergang rechts
			MenuButton.SetPosition ((int) ButtonPosR.x - FadeIn - 32, (int) ButtonPosL.y, (int) ButtonPosR.x - FadeIn, (int) ButtonPosR.y);
			MenuButton.SetPictureRange (0.51f, 0.01f, 0.99f, 0.49f);
			Eng3d.Render2dPicture (&MenuButton, pButton);
			//Rechts
			MenuButton.SetPosition ((int) ButtonPosR.x - FadeIn, (int) ButtonPosL.y, (int) ButtonPosR.x, (int) ButtonPosR.y);
			MenuButton.SetPictureRange (0.01f, 0.51f, 0.49f, 0.99f);
			Eng3d.Render2dPicture (&MenuButton, pButton);
			//Text
			pText->SetPosition ((int) ButtonPosL.x, (int) ButtonPosL.y, (int) ButtonPosR.x, (int) ButtonPosR.y);
			pText->SetText (&MenuStr[i][0]);
			if (i == MenuHl)
				pText->SetColor (ColorMake (255, 255, 0));
			else
				pText->SetColor (ColorMake (255, 255, 255));
			Eng3d.Render2dText (pText);
		}
		if (!Busy)
			Eng3d.Render2dPicture (&MenuCursor, pCurs);
		else
			Eng3d.Render2dPicture (&MenuCursor, pBusy);
		Eng3d.RenderEnd ();
		Eng3d.RenderShow ();

		if (((!RdMouse->rgbButtons[0] && LDown) || (!RdMouse->rgbButtons[1] && RDown) || (!RdKeyb[DIK_RETURN] && !RdKeyb[DIK_NUMPADENTER] && RetDown)) && MenuHl != -1)
		{
			if (MenuHl == MenuCnt - 1 && MenuID != MENU_OPTIONS_GRAFICS)		//Zurck bzw. beenden
				break;
			if (RetDown)														//Enter entspricht Linksklick
				LDown = true;
			if (MenuID == MENU_MAIN)
			{
				//Hautmen
				if (MenuHl == 0)
				{
					//Einzelspieler
					PIdent = 0;
					RunGame (false, false);
				}
				else if (MenuHl == 1)
					//Mehrspieler
					ShowMenu (MENU_MULTIPLAYER, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
				else if (MenuHl == 2)
				{
					//Optionen
					ShowMenu (MENU_OPTIONS, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
					if (Options.PlayerName[0] == '\0')
						strcpy (Options.PlayerName, "Player");
				}
				else if (MenuHl == 3)
					//Informationen zum Spiel
					ShowInfo (pBackgr, false);
				else if (MenuHl == 4)
					//Credits
					ShowInfo (pBackgr, true);
			}
			else if (MenuID == MENU_MULTIPLAYER)
			{
				//Multiplayer-Men
				if (MenuHl == 0)
				{
					//Neues Spiel erstellen
					if (GameServer.Host (Options.PlayerName, 3))
					{
						MPOptions.AIPlayerCnt = 0;
						ShowMenu (MENU_MULTIPLAYER_SERVER, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
						GameServer.Unhost ();
					}
				}
				else if (MenuHl == 1)
					//An einem existierenden Spiel teilnehmen
					ShowMenu (MENU_MULTIPLAYER_CLIENT, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
			}
			else if (MenuID == MENU_MULTIPLAYER_SERVER)
			{
				//Server-Men
				if (MenuHl == 1)
				{
					//Anzahl Computergegner
					if (LDown)
					{
						MPOptions.AIPlayerCnt++;
						if (MPOptions.PlayerCnt + MPOptions.AIPlayerCnt > 4 || MPOptions.AIPlayerCnt > 2)
							MPOptions.AIPlayerCnt = 0;
					}
					else
					{
						MPOptions.AIPlayerCnt--;
						if (MPOptions.AIPlayerCnt < 0)
						{
							MPOptions.AIPlayerCnt = 4 - MPOptions.PlayerCnt;
							if (MPOptions.AIPlayerCnt > 2)
								MPOptions.AIPlayerCnt = 2;
						}
					}
				}
				else if (MenuHl == 2)
				{
					//Gegner-Intelligenz
					if (LDown)
					{
						Options.OpponentSkill++;
						Options.OpponentSkill %= 3;
					}
					else
					{
						Options.OpponentSkill--;
						if (Options.OpponentSkill < 0)
							Options.OpponentSkill = 2;
					}
				}
				else if (MenuHl == 3 && MPOptions.PlayerCnt > 1)
				{
					//Spiel starten
					//**********CLIENT_SETTINGS senden*********
					SData[0] = CLIENT_SETTINGS;
					SaveSettings = Options;
					Options.OpponentCnt = MPOptions.AIPlayerCnt + MPOptions.PlayerCnt - 1;
					memcpy (&SData[1], &Options, sizeof (TronSettings));			//Die lokalen Spieleinstellungen bertragen
					for (i = 0; i < Options.OpponentCnt + 1; i++)					//Die Moped-Farben bertragen
					{
						MPBikeInfo[i].BikeType = (int) GetRandomVal (0, 4) % 4;
						memcpy (&SData[1 + sizeof (TronSettings) + i * sizeof (int)], &MPBikeInfo[i].BikeType, sizeof (int));
					}
					ClientReady = 0;
					//Jetzt die Daten an jeden Client einzeln senden. Bei i=1 beginnen, da 0 der Server selbst ist.
					for (i = 1; i < MPOptions.PlayerCnt; i++)
					{
						//Die jeweilige Ident-Nummer fr den Client senden.
						memcpy (&SData[1 + sizeof (TronSettings) + (Options.OpponentCnt + 1) * sizeof (int)], &i, sizeof (int));
						GameServer.SendData (MPOptions.pIdents[i], SData, 1 + sizeof (TronSettings) + (Options.OpponentCnt + 1) * sizeof (int) + sizeof (int), 0);
					}
					//**********auf alle SERVER_SETTINGSOK warten*********
					Eng3d.StartTimer ();
					WaitTime = Eng3d.GetAbsTimeDiff ();
					while (ClientReady < MPOptions.PlayerCnt - 1)
					{
						Eng3d.HandleMessages ();
						if (Eng3d.GetAbsTimeDiff () - WaitTime > 3)
							return;
					}
					
					//Fehlende Namen ergnzen (Name des Spielers am Server, AI-Namen)
					strcpy (MPBikeInfo[0].Name, Options.PlayerName);
					for (i = MPOptions.PlayerCnt; i < Options.OpponentCnt + 1; i++)
					{
						if (i - MPOptions.PlayerCnt == 0)
							strcpy (MPBikeInfo[i].Name, "Bot Gates");
						else if (i - MPOptions.PlayerCnt == 1)
							strcpy (MPBikeInfo[i].Name, "Bot Elvis");
						else if (i - MPOptions.PlayerCnt == 2)
							strcpy (MPBikeInfo[i].Name, "Bot Bush");
						else
							strcpy (MPBikeInfo[i].Name, "Bot Britney");
					}

					//**********CLIENT_LOAD senden*********
					ClientReady = 0;
					SData[0] = CLIENT_LOAD;
					for (i = 0; i < Options.OpponentCnt + 1; i++)
						memcpy (&SData[1 + i * 64 * sizeof (char)], MPBikeInfo[i].Name, 64 * sizeof (char));
					GameServer.SendData (0, SData, 1 + (Options.OpponentCnt + 1) * 64 * sizeof (char), 0);

					//Spiel starten
					PIdent = 0;
					RunGame (true, true);
					
					//Lokale Optionen nach dem Multiplayer-Spiel wiederherstellen
					Options = SaveSettings;
					RData[0] = CONNECTION_OK;
				}
			}
			else if (MenuID == MENU_MULTIPLAYER_CLIENT)
			{
				//Client-Men
				if (ServerCnt > 0 && GameClient.Connect (pServerDesc[MenuHl].IP, 3))
				{
					ShowMenu (MENU_MULTIPLAYER_CLIENTWAIT, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
					GameClient.Disconnect ();
				}
			}
			else if (MenuID == MENU_OPTIONS)
			{
				//Optionen-Men
				if (MenuHl == 0)
					ShowMenu (MENU_OPTIONS_GRAFICS, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
				else if (MenuHl == 1)
					ShowMenu (MENU_OPTIONS_SOUNDS, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
				else if (MenuHl == 2)
					ShowMenu (MENU_OPTIONS_DETAILS, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
				else if (MenuHl == 3)
					ShowMenu (MENU_OPTIONS_GAMEPLAY, CursorPos, pCurs, pBusy, pButton, pText, pBackgr);
			}
			else if (MenuID == MENU_OPTIONS_GRAFICS)
			{
				//Grafik-Men
				if (MenuHl == 0)
				{
					//Auflsung ndern (bei gleichbleibender Farbtiefe)
					//Auflsung wird in der Temp-Variable ActRes gespeichert (um Grafikfehler zu vermeiden)
					i = ActRes;
					do
					{
						if (LDown)
							i++;
						else
							i--;
						if (i < 0)
							i = ResCnt - 1;
						i %= ResCnt;
					} while (pRes[i].Format != pRes[ActRes].Format);
					ActRes = i;
				}
				else if (MenuHl == 1)
				{
					//Farbtiefe ndern (bei gleichbleibender Auflsung)
					//Farbtiefe wird in der Temp-Variable ActRes gespeichert (um Grafikfehler zu vermeiden)
					for (i = 0; i < ResCnt; i++)
					{
						if (pRes[i].ResX == pRes[ActRes].ResX && pRes[i].ResY == pRes[ActRes].ResY && pRes[i].Format != pRes[ActRes].Format)
						{
							ActRes = i;
							RdCaps = Eng3d.GetPossibleCaps (pRes[ActRes].Format, false);
							break;
						}
					}
				}
				else if (MenuHl == 2)
				{
					//Anti-Aliasing ndern
					if (LDown)
					{
						if (ActFSAA == 0)
							ActFSAA = 1;
						ActFSAA *= 2;
						if (ActFSAA > RdCaps.MaxAntiAlias)
							ActFSAA = 0;
					}
					else
					{
						ActFSAA /= 2;
						if (ActFSAA == 1)
							ActFSAA = 0;
						else if (ActFSAA == 0)
							ActFSAA = RdCaps.MaxAntiAlias;
					}
				}
				else if (MenuHl == 3)
				{
					//bernehmen
					if (Eng3d.SetRes (&pRes[ActRes], ActFSAA, false))
					{
						Options.Res = pRes[ActRes];
						Options.UseFSAA = ActFSAA;
						break;
					}
					else
					{
						ActRes = PrevRes;
						ActFSAA = Options.UseFSAA;
					}
				}
				else if (MenuHl == 4)
				{
					//Zurck. Wird hier gesondert behandelt, da die Einstellungen wieder verworfen werden.
					ActRes = PrevRes;
					RdCaps = Eng3d.GetPossibleCaps (pRes[ActRes].Format, false);
					break;
				}
				CreateMenuText (MENU_OPTIONS_GRAFICS, MenuStr, &pRes[ActRes], &ActFSAA);
			}
			else if (MenuID == MENU_OPTIONS_SOUNDS)
			{
				if (MenuHl == 0)
					Options.UseSounds = !Options.UseSounds;
				else if (MenuHl == 1)
					Options.UseMusic = !Options.UseMusic;
				CreateMenuText (MENU_OPTIONS_SOUNDS, MenuStr);
			}
			else if (MenuID == MENU_OPTIONS_DETAILS)
			{
				if (MenuHl == 0)
					//Spiegeleffekte
					Options.UseReflection = !Options.UseReflection;
				else if (MenuHl == 1)
				{
					//Schatten
					if (LDown)
					{
						Options.UseShadow++;
						Options.UseShadow %= 3;
					}
					else
					{
						Options.UseShadow--;
						if (Options.UseShadow < 0)
							Options.UseShadow = 2;
					}
				}
				else if (MenuHl == 2)
					//Transparente Wnde
					Options.UseTransWalls = !Options.UseTransWalls;
				else if (MenuHl == 3)
					//Glanzlichter
					Options.UseSpecular = !Options.UseSpecular;
				else if (MenuHl == 4)
					//Anisotropischer Filter
					Options.UseAnisotropic = !Options.UseAnisotropic;
				CreateMenuText (MENU_OPTIONS_DETAILS, MenuStr);
			}
			else if (MenuID == MENU_OPTIONS_GAMEPLAY)
			{
				if (MenuHl != 7)
					EditName = false;
				if (MenuHl == 0)
					//Action-Modus
					Options.EnableAction = !Options.EnableAction;
				else if (MenuHl == 1)
				{
					//Kamera
					if (LDown)
					{
						Options.Cam++;
						Options.Cam %= 8;
					}
					else
					{
						Options.Cam--;
						if (Options.Cam < 0)
							Options.Cam = 7;
					}
				}
				else if (MenuHl == 2)
				{
					//Arenagre
					if (LDown)
					{
						Options.LandSize += 100;
						if (Options.LandSize > 500)
						{
							Options.LandSize = 100;
							if (Options.OpponentCnt > 4)
								Options.OpponentCnt = 4;
						}
					}
					else
					{
						Options.LandSize -= 100;
						if (Options.LandSize == 100 && Options.OpponentCnt > 4)
							Options.OpponentCnt = 4;
						else if (Options.LandSize < 100)
							Options.LandSize = 500;
					}
				}
				else if (MenuHl == 3)
				{
					if (LDown)
					{
						//Moped-Geschwindigkeit
						Options.MopedSpeed += 14;
						if (Options.MopedSpeed == 126)
							Options.MopedSpeed = 14;
					}
					else
					{
						Options.MopedSpeed -= 14;
						if (Options.MopedSpeed == 0)
							Options.MopedSpeed = 112;
					}
				}
				else if (MenuHl == 4)
				{
					//Anzahl Gegner
					if (LDown)
					{
						Options.OpponentCnt++;
						if (Options.OpponentCnt > 4 && Options.LandSize == 100)
							Options.OpponentCnt = 1;
						else if (Options.OpponentCnt > 5)
							Options.OpponentCnt = 1;
					}
					else
					{
						Options.OpponentCnt--;
						if (Options.OpponentCnt == 0)
						{
							if (Options.LandSize == 100)
								Options.OpponentCnt = 4;
							else
								Options.OpponentCnt = 5;
						}
					}
				}
				else if (MenuHl == 5)
				{
					//Gegner-Intelligenz
					if (LDown)
					{
						Options.OpponentSkill++;
						Options.OpponentSkill %= 3;
					}
					else
					{
						Options.OpponentSkill--;
						if (Options.OpponentSkill < 0)
							Options.OpponentSkill = 2;
					}
				}
				else if (MenuHl == 6)
				{
					//Punkte zum Sieg
					if (LDown)
					{
						Options.FragLimit += 5;
						if (Options.FragLimit > 100)
							Options.FragLimit = 5;
					}
					else
					{
						Options.FragLimit -= 5;
						if (Options.FragLimit < 5)
							Options.FragLimit = 100;
					}
				}
				else if (MenuHl == 7)
					//Name des Spielers
					EditName = !EditName;
				CreateMenuText (MENU_OPTIONS_GAMEPLAY, MenuStr, NULL, NULL, EditName);
			}
			LDown = false;
			RDown = false;
			RetDown = false;
		}
		if (!RdKeyb[DIK_ESCAPE] && ESCDown)
		{
			if (MenuID == MENU_OPTIONS_GRAFICS)
			{
				ActRes = PrevRes;
				RdCaps = Eng3d.GetPossibleCaps (pRes[ActRes].Format, false);
			}
			break;
		}
	}
}

void ShowInfo (Revo2dTexture *pBackgr, bool Credits)
{
	int i;
	char f[128];
	Revo2dText Text;
	Revo2dPicture MenuBackgr;
	BYTE *RdKeyb;
	bool KeyDown;

	MenuBackgr.SetPosition (0, 0, Options.Res.ResX, Options.Res.ResY);
	Text.Initialize (&Eng3d);
	Text.SetFont (FontMake ("Arial", 40, true, false, false));
	Text.SetColor (ColorMake (0, 255, 0));
	Text.SetFormat (1, 1);
	//Hintergrundbild
	Eng3d.RenderStart (NULL, NULL);
	Eng3d.RenderClear (true, ColorMake (0, 0, 255), false, false);
	Eng3d.Render2dPicture (&MenuBackgr, pBackgr);
	//berschrift
	Text.SetPosition (0, 0, Options.Res.ResX, 40);
	if (Credits)
		Text.SetText ("CREDITS");
	else
	{
		sprintf (f, "REVO TRON v %d.%d", TRON_MVERS, TRON_SVERS);
		Text.SetText (f);
	}
	Eng3d.Render2dText (&Text);
	//Restlicher Text
	Text.SetFont (FontMake ("Arial", 25, true, false, false));
	Text.SetColor (ColorMake (255, 255, 255));
	for (i = 0; i < 12; i++)
	{
		Text.SetPosition (0, 60 + i * 30, Options.Res.ResX, 60 + (i + 1) * 30);
		if (Credits)
		{
			if (i == 0)
				Text.SetText ("Completly programming work:");
			else if (i == 1)
				Text.SetText ("Mathias Kunter (mathiaskunter@yahoo.de)");
			else if (i == 2 || i == 5 || i == 8)
				Text.SetText ("");
			else if (i == 3)
				Text.SetText ("Designer of the 3d objects:");
			else if (i == 4)
				Text.SetText ("Wolfgang Gelbmann (wolf_gelb@hotmail.com)");
			else if (i == 6)
				Text.SetText ("Designer of the floor and wall textures:");
			else if (i == 7)
				Text.SetText ("Tyler Esselstrom (hazard369@aol.com)");
			else if (i == 9)
				Text.SetText ("Beta testers:");
			else if (i == 10)
				Text.SetText ("Timon Kunter, True, Wolfgang Unger");
			else
				Text.SetText ("");
		}
		else
		{
			if (i == 0)
				Text.SetText ("Website: http://revotron.tripod.com");
			else if (i == 1 || i == 6)
				Text.SetText ("");
			else if (i == 2)
				Text.SetText ("Key configuration:");
			else if (i == 3)
				Text.SetText ("Left and right: Changes the direction");
			else if (i == 4)
				Text.SetText ("Space: In action mode, use an item");
			else if (i == 5)
				Text.SetText ("C: Changes the camera view ingame");
			else if (i == 7)
				Text.SetText ("Action mode: This is a new feature which was invented");
			else if (i == 8)
				Text.SetText ("by Revo Tron. You'll get an item every 10 to 15 seconds.");
			else if (i == 9)
				Text.SetText ("You can see this on the top-right corner of your screen.");
			else if (i == 10)
				Text.SetText ("If you have an item, you can use it with the space key.");
			else
				Text.SetText ("");
		}
		Eng3d.Render2dText (&Text);
	}
	Eng3d.RenderEnd ();
	Eng3d.RenderShow ();
	//Auf Tastendruck-Besttigung warten
	KeyDown = false;
	while (!KeyDown)
	{
		RdKeyb = Inp.ReadKeyboard ();
		for (i = 0; i < 256; i++)
		{
			if (RdKeyb[i])
			{
				KeyDown = true;
				break;
			}
		}
	}
	KeyDown = true;
	while (KeyDown)
	{
		KeyDown = false;
		RdKeyb = Inp.ReadKeyboard ();
		for (i = 0; i < 256; i++)
		{
			if (RdKeyb[i])
			{
				KeyDown = true;
				break;
			}
		}
	}
}

int CreateMenuText (int MenuID, char MenuStr[MAX_MENUITEM][MAX_MENUTEXT], Revo3dRes *pNR, int *pNF, bool EditName)
{
	if (MenuID == MENU_MAIN)
	{
		strcpy (&MenuStr[0][0], "Singleplayer game");
		strcpy (&MenuStr[1][0], "Multiplayer game");
		strcpy (&MenuStr[2][0], "Options");
		strcpy (&MenuStr[3][0], "Game information");
		strcpy (&MenuStr[4][0], "Credits");
		strcpy (&MenuStr[5][0], "Exit");
		return (6);
	}
	else if (MenuID == MENU_MULTIPLAYER)
	{
		strcpy (&MenuStr[0][0], "Host a game");
		strcpy (&MenuStr[1][0], "Connect to a hosted game");
		strcpy (&MenuStr[2][0], "Back");
		return (3);
	}
	else if (MenuID == MENU_MULTIPLAYER_CLIENTWAIT)
	{
		strcpy (&MenuStr[0][0], "Waiting for server...");
		strcpy (&MenuStr[1][0], "Back");
		return (2);
	}
	else if (MenuID == MENU_OPTIONS)
	{
		strcpy (&MenuStr[0][0], "Graphic options");
		strcpy (&MenuStr[1][0], "Sound options");
		strcpy (&MenuStr[2][0], "Detail options");
		strcpy (&MenuStr[3][0], "Game settings");
		strcpy (&MenuStr[4][0], "Back");
		return (5);
	}
	else if (MenuID == MENU_OPTIONS_GRAFICS)
	{
		sprintf (&MenuStr[0][0], "Resolution: %d x %d", pNR->ResX, pNR->ResY);
		sprintf (&MenuStr[1][0], "Color depth: %d bit", Eng3d.GetFormatBPP (pNR->Format));
		if (*pNF)
			sprintf (&MenuStr[2][0], "Anti-Aliasing: %d x", *pNF);
		else
			strcpy (&MenuStr[2][0], "Anti-Aliasing: off");
		strcpy (&MenuStr[3][0], "Accept");
		strcpy (&MenuStr[4][0], "Back");
		return (5);
	}
	else if (MenuID == MENU_OPTIONS_SOUNDS)
	{
		if (Options.UseSounds)
			strcpy (&MenuStr[0][0], "Sounds: on");
		else
			strcpy (&MenuStr[0][0], "Sounds: off");
		strcpy (&MenuStr[1][0], "Back");
		return (2);
	}
	else if (MenuID == MENU_OPTIONS_DETAILS)
	{
		if (Options.UseReflection)
			strcpy (&MenuStr[0][0], "Mirror effects: on");
		else
			strcpy (&MenuStr[0][0], "Mirror effects: off");
		if (Options.UseShadow == 0)
			strcpy (&MenuStr[1][0], "Shadows: off");
		else if (Options.UseShadow == 1)
		{
			if (RdCaps.ShadowNoTransparencyAviable)
				strcpy (&MenuStr[1][0], "Shadows: on, without transparency");
			else
				strcpy (&MenuStr[1][0], "Shadows: on, without transparency - N/A");
		}
		else
		{
			if (RdCaps.ShadowAviable)
				strcpy (&MenuStr[1][0], "Shadows: on, with transparency");
			else
				sprintf (&MenuStr[1][0], "Shadows: on, with transparency - N/A in %d bit color depth", Eng3d.GetFormatBPP (Options.Res.Format));
		}
		if (Options.UseTransWalls)
			strcpy (&MenuStr[2][0], "Transparent walls: on");
		else
			strcpy (&MenuStr[2][0], "Transparent walls: off");
		if (Options.UseSpecular)
		{
			if (RdCaps.SpecularAviable)
				strcpy (&MenuStr[3][0], "Light reflections: on");
			else
				strcpy (&MenuStr[3][0], "Light reflections: on - N/A");
		}
		else
			strcpy (&MenuStr[3][0], "Light reflections: off");
		if (Options.UseAnisotropic)
		{
			if (RdCaps.AnisotropicAviable)
				strcpy (&MenuStr[4][0], "Anisotropic texture filtering: on");
			else
				strcpy (&MenuStr[4][0], "Anisotropic texture filtering: on - N/A");
		}
		else
			strcpy (&MenuStr[4][0], "Anisotropic texture filtering: off");
		strcpy (&MenuStr[5][0], "Back");
		return (6);
	}
	else if (MenuID == MENU_OPTIONS_GAMEPLAY)
	{
		if (Options.EnableAction)
			strcpy (&MenuStr[0][0], "Action mode: on");
		else
			strcpy (&MenuStr[0][0], "Action mode: off");
		if (Options.Cam == 0)
			strcpy (&MenuStr[1][0], "Camera: behind, near");
		else if (Options.Cam == 1)
			strcpy (&MenuStr[1][0], "Camera: behind, medium distance");
		else if (Options.Cam == 2)
			strcpy (&MenuStr[1][0], "Camera: behind, far distance");
		else if (Options.Cam == 3)
			strcpy (&MenuStr[1][0], "Camera: classic top-down");
		else if (Options.Cam == 4)
			strcpy (&MenuStr[1][0], "Camera: modern top-down, low");
		else if (Options.Cam == 5)
			strcpy (&MenuStr[1][0], "Camera: modern top-down, high");
		else if (Options.Cam == 6)
			strcpy (&MenuStr[1][0], "Camera: onboard");
		else if (Options.Cam == 7)
			strcpy (&MenuStr[1][0], "Camera: in front of");
		sprintf (&MenuStr[2][0], "Arena size: %d x %d meters", (int) Options.LandSize, (int) Options.LandSize);
		sprintf (&MenuStr[3][0], "Speed: %d km/h", (int) (Options.MopedSpeed / 14) * 50);
		sprintf (&MenuStr[4][0], "Bot count: %d", Options.OpponentCnt);
		if (Options.OpponentSkill == 0)
			strcpy (&MenuStr[5][0], "Bot skill: bad");
		else if (Options.OpponentSkill == 1)
			strcpy (&MenuStr[5][0], "Bot skill: medium");
		else if (Options.OpponentSkill == 2)
			strcpy (&MenuStr[5][0], "Bot skill: good");
		sprintf (&MenuStr[6][0], "Point limit: %d", Options.FragLimit);
		if (EditName)
			sprintf (&MenuStr[7][0], "Player's name: %s_", Options.PlayerName);
		else
			sprintf (&MenuStr[7][0], "Player's name: %s", Options.PlayerName);
		strcpy (&MenuStr[8][0], "Back");		
		return (9);
	}
	return (0);
}

//Frs Gameplay relevant: RelTimeDiff, Abweichung der grnen Schsse (zufllig)
void MasterGameController (float RelTimeDiff, bool Multiplayer)
{
	int i, CrashCnt, ActionCnt;
	int Ident[100], ActIdent;
	BYTE Action[100], ActAction;
	
	//**************Schsse handeln***************
	Shots.DoEvents (RelTimeDiff, Multiplayer);

	//**************Schilde der Mopeds handeln************
	for (i = 0; i < Options.OpponentCnt + 1; i++)
		pPlayer[i].HandleShield ();

	//*******************Mopeds handeln (bewegen, lenken, Reifen drehen etc.)**************
	if (GameRuns)
	{
		//Eigener Spieler lenkt in der RunGame-Routine!
		for (i = 0; i < Options.OpponentCnt + 1; i++)
		{
			if (Multiplayer && i < MPOptions.PlayerCnt)
				continue;			
			if (pPlayer[i].IsAlive () && i != PIdent)
				//AI lenkt hier
				pPlayer[i].AI (RelTimeDiff, &Shots);
		}
		if (Multiplayer)
		{
			//Clients beim Multiplayer-Spiel werden hier gelenkt.
			ActionCnt = MPBuf.GetActionCnt ();
			MPBuf.GetActions (Ident, Action);
			for (i = 0; i < ActionCnt; i++)
			{
				ActIdent = Ident[i];
				if (ActIdent < 0 || ActIdent > Options.OpponentCnt)
					//Es ist etwas faul im Buffer!
					continue;
				if (pPlayer[ActIdent].IsAlive () && ActIdent != PIdent)
				{
					ActAction = Action[i];
					if (ActAction == 0)
						//Links lenken
						pPlayer[ActIdent].ChangeDir (0);
					else if (ActAction == 1)
						//Rechts lenken
						pPlayer[ActIdent].ChangeDir (1);
					else if (ActAction == 2)
						//Item verwenden
						pPlayer[ActIdent].FireItem (&Shots);
				}
			}
		}
		//DoEvents muss nach AI aufgerufen werden, da ein mgliches
		//ChangeDir vorher verarbeitet werden muss
		for (i = 0; i < Options.OpponentCnt + 1; i++)
			pPlayer[i].DoEvents (RelTimeDiff, Options.MopedSpeed);
	}
	else
	{
		if (NotRunTime < 3)
		{
			for (i = 0; i < Options.OpponentCnt + 1; i++)
				pPlayer[i].DoEvents (RelTimeDiff, 3 - NotRunTime);
		}
		else
		{
			for (i = 0; i < Options.OpponentCnt + 1; i++)
				pPlayer[i].DoEvents (RelTimeDiff, 0);
		}
	}

	//******************Mopeds auf Kollission abfragen**************
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (pPlayer[i].IsAlive () && !pPlayer[i].CanPlace ())
			pPlayer[i].WillCrash = true;
	}
	//Anzahl der Spieler ermitteln, die jetzt zerstrt werden.
	//Muss extra durchgefhrt werden aufgrund von mglichen Moped-Zusammenstssen.
	CrashCnt = 0;
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (pPlayer[i].IsAlive () && pPlayer[i].WillCrash)
			CrashCnt++;
	}
	//Punkte vergeben fr jeden Spieler, der jetzt noch "alive" ist.
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (pPlayer[i].IsAlive ())
		{
			if (pPlayer[i].WillCrash)
			{
				pPlayer[i].Frags += CrashCnt - 1;
				pPlayer[i].Crash ();
				if (Options.UseSounds)
				{
					//Sound abspielen
					BikeCrashSound.Set3dPosition (&Vector3dMake (pPlayer[i].GetPos ().x, 0, pPlayer[i].GetPos ().y));
					BikeCrashSound.SetPlayPosition (0);
					EngSound.Update3dSounds ();
					EngSound.Play3dSound (&BikeCrashSound, false);
				}
				if (Multiplayer)
				{
					SData[0] = CLIENT_PLAYSOUND;
					SData[1] = 0;
					memcpy (&SData[2], &pPlayer[i].GetPos (), sizeof (D3DXVECTOR2));
					GameServer.SendData (0, SData, 2 + sizeof (D3DXVECTOR2), 0);
				}
				//Explosion erstellen
				Explosions.CreateExplosion (&pPlayer[i].GetPos (), false, Multiplayer);
				MopedAlive--;
			}
			else
				pPlayer[i].Frags += CrashCnt;
		}
	}	


	//**************************Den Fhrenden feststellen*************************
	LeaderScore = DoRanking ();
	if (LeaderScore >= Options.FragLimit)
	{
		for (i = 0; i < Options.OpponentCnt + 1; i++)
			pPlayer[i].Crash ();
		GamePassed = true;
		TimeFac = 1;
	}
	
	//*******************Allgemeines*********************
	NotRunTime += RelTimeDiff;
	if (!GameRuns && NotRunTime >= 5)
		GameRuns = true;
	if (NotAliveTime >= 3 && !Multiplayer)
		TimeFac = 3;
	if (MopedAlive < 2)
	{
		GamePassed = true;
		TimeFac = 1;
	}
}

void RunGame (bool Multiplayer, bool IsServer)
{
	int i, j, x, y;
	int fpsRendered, OppShow, RankLine, RDataSize;
	float RelTimeDiff, fpsT, CamBlendTime, TransErg, SendT, WaitTime;
	char f[128];
	BYTE *RdKeyb, pRData[131072];
	BYTE *pMem;
	bool RDown, LDown, CDown, ESCDown, SUse, ReturnDown;
	bool StartOK, FirstInit, GameQuitQuest;
	
	TronWallDesc WallDesc;
	CameraHandler Cam;

	Revo3dPolygons GameLand;
	Vertex GameV[20];
	Revo2dTexture TexFloor, TexWall, TexMopedWall, TexFire;
	Revo2dPicture PicItem;
	Revo2dTexture TexShotGreen, TexShotRed, TexSpeedup, TexShotBlack, TexShield;
	Revo2dText fps, TextRank, TextLost;
	Revo3dPlane FloorPlane;

	RECT TexRect;
	D3DLIGHT8 PointLight;
	D3DMATERIAL8 FloorMat;
	D3DXVECTOR2 ActPos, CamDir, PrevDir, CamDiff, CamVec;				//Fr Kamera (2d)
	D3DXVECTOR3 NPos, NDest, CamDir3d;



	//********************EINMALIGES LADEN********************
	//Texturen und Pictures laden
	TexFloor.Initialize (&Eng3d);
	if (!TexFloor.LoadFromFile ("textures\\floor.bmp", 0, 5, false))
		return;
	TexWall.Initialize (&Eng3d);
	if (!TexWall.LoadFromFile ("textures\\wall.bmp", 0, 0, false))
		return;
	TexMopedWall.Initialize (&Eng3d);
	if (!TexMopedWall.LoadFromFile ("textures\\mopedwall.bmp", 0, 1, false))
		return;
	TexFire.Initialize (&Eng3d);
	if (!TexFire.LoadFromFile ("textures\\fire.bmp", 0, 1, true))
		return;
	if (TexFire.EditStart ())
	{
		for (x = 0; x < 64; x++)
		{
			for (y = 0; y < 64; y++)
			{
				TransErg = 255 - 8 * (float) sqrt ((32 - x) * (32 - x) + (32 - y) * (32 - y));
				if (TransErg > 0)
					TexFire.Edit (&Vector2dMake ((float) x, (float) y), (int) TransErg, -1, -1, -1);
				else
					TexFire.Edit (&Vector2dMake ((float) x, (float) y), 0, -1, -1, -1);
			}
		}
		TexFire.EditEnd ();
	}
	TexShotGreen.Initialize (&Eng3d);
	if (!TexShotGreen.LoadFromFile ("textures\\shotgreen.bmp", 0, 1, false))
		return;
	TexShotRed.Initialize (&Eng3d);
	if (!TexShotRed.LoadFromFile ("textures\\shotred.bmp", 0, 1, false))
		return;
	TexSpeedup.Initialize (&Eng3d);
	if (!TexSpeedup.LoadFromFile ("textures\\speedup.bmp", ColorMake (0, 0, 0), 1, false))
		return;
	TexShotBlack.Initialize (&Eng3d);
	if (!TexShotBlack.LoadFromFile ("textures\\shotblack.bmp", 0, 1, false))
		return;
	TexShield.Initialize (&Eng3d);
	if (!TexShield.LoadFromFile ("textures\\shield.bmp", ColorMake (0, 0, 0), 1, false))
		return;
	PicItem.SetPosition (Options.Res.ResX - 64, 0, Options.Res.ResX, 64);

	//Landschaft laden
	GeometryInit (Options.LandSize, Options.LandSize, 15, GameV);
	GameLand.Initialize (&Eng3d, false);
	GameLand.SetVertexData (GameV, 0, 20);
	GameLand.SetMaterial (&Eng3d.GetDefaultMaterial ());
	GameLand.SetPolyFormat (TRIANGLESTRIP);
	GameLand.SetVisibility (FRONTSIDE);

	//Schild erstellen
	Shield.Initialize (&Eng3d);
	if (!Shield.LoadFromFile ("objects\\shield.x", 1))
		return;
	Shield.SetTransparency (100);

	//Textfelder erstellen
	fps.Initialize (&Eng3d);
	fps.SetFont (FontMake ("Arial", 20, false, false, false));
	fps.SetColor (ColorMake (255, 255, 255));
	fps.SetPosition (0, 0, 400, 20);
	fps.SetText ("");
	TextRank.Initialize (&Eng3d);
	TextRank.SetFont (FontMake ("Arial", 20, true, false, false));
	TextRank.SetColor (ColorMake (255, 255, 255));
	TextLost.Initialize (&Eng3d);
	TextLost.SetFont (FontMake ("Arial", 50, true, false, false));
	TextLost.SetPosition (0, 0, Options.Res.ResX, Options.Res.ResY);
	TextLost.SetFormat (1, 1);
	TextLost.SetColor (ColorMake (255, 255, 255));
	TextLost.SetText ("Connection lost...");

	//Wnde vorbereiten
	WallDesc.Initialize (100 * (Options.OpponentCnt + 1));

	//Explosionen vorbereiten
	Explosions.Initialize (30, &TexFire);

	//Speicher fr Spieler-Objekte erstellen
	pPlayer = new TronPlayer[Options.OpponentCnt + 1];

	//Schsse vorbereiten
	Shots.Initialize (10, &WallDesc, &TexRect, &Explosions, pPlayer);

	//Lichtquellen erstellen
	PointLight = PointLightMake (1000, 1000, 1000, 255, &Vector3dMake (Options.LandSize / 2, 30, Options.LandSize / 2), 1000);
	Eng3d.SetLight (0, &PointLight);
	PointLight = PointLightMake (0, 0, 0, 75, &Vector3dMake (0, 0, 0), 1000);

	//Planes erstellen
	FloorPlane = PlaneMake (&Vector3dMake (0, 0, 0), &Vector3dMake (0, 0, 1), &Vector3dMake (1, 0, 0));
	FloorMat = MaterialMake (255, 255, 255, 0, 0, 0, 0, 75);

	//Anisotropischer Filter
	if (Options.UseAnisotropic)
		Eng3d.SetAnisotropicState (true);

	//Sounds
	BikeCrashSound.Initialize (&EngSound);
	BikeCrashSound.LoadFromFile ("sounds\\bikecrash.wav");
	BikeCrashSound.SetDistances (0, 1);
	ShotCrashSound.Initialize (&EngSound);
	ShotCrashSound.LoadFromFile ("sounds\\wallcrash.wav");
	ShotCrashSound.SetDistances (0, 1);
	FireShotSound.Initialize (&EngSound);
	FireShotSound.LoadFromFile ("sounds\\shot.wav");
	FireShotSound.SetDistances (0, 5);


	//Sonstige Variablen, die einmalig gesetzt werden
	FirstInit = true;
	LeaderScore = 0;

	//Spieler erstellen
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (!pPlayer[i].Initialize ((int) GetRandomVal (0, 4) % 4, &WallDesc, i, Multiplayer, IsServer, &TexMopedWall))
		{
			delete []pPlayer;
			return;
		}
	}

	if (Multiplayer)
	{
		if (!IsServer)
		{
			//**********SERVER_LOADOK senden*********
			SData[0] = SERVER_LOADOK;
			GameClient.SendData (MPOptions.pIdents[0], SData, 1, 0);
		}
		else
		{
			//**********auf alle SERVER_LOADOK warten*********
			Eng3d.StartTimer ();
			WaitTime = Eng3d.GetAbsTimeDiff ();
			while (ClientReady < MPOptions.PlayerCnt - 1)
			{
				Eng3d.HandleMessages ();
				if (Eng3d.GetAbsTimeDiff () - WaitTime > 3)
				{
					delete []pPlayer;
					return;
				}
			}
		}
	}

	while (LeaderScore < Options.FragLimit)
	{
		//*********************INIT-VORGANG BEI JEDEM NEUEN SPIEL*****************
		if (!FirstInit)
		{
			Shots.Delete ();				//Eventuelle Schsse aus dem vorigen Spiel lschen
			Explosions.DoEvents (60);		//Eventuelle Explosionen aus dem vorigen Spiel lschen
			Cam.DoEvents (60);
			for (i = 0; i < Options.OpponentCnt + 1; i++)
			{
				pPlayer[i].Crash ();		//Crash-Routine durchfhren (falls nicht schon frher geschehen), damit der Motor-Sound gestoppt wird
				pPlayer[i].Delete ();		//Die Revo3dPolygons der Wand-Beschreibung jedes Mopeds zurcksetzen
			}
			WallDesc.Delete ();				//Die globale Wand-Beschreibung des vorigen Spiels lschen
		}
		//Neue Positionen fr die Mopeds ermitteln, falls nicht Client eines Multiplayer-Spiels.
		if (!Multiplayer || IsServer)
		{
			for (i = 0; i < Options.OpponentCnt + 1; i++)
				pPlayer[i].Reset ();
			//Prfen, ob die Startaufstellung in Ordnung ist (die Mopeds drfen am Anfang nicht zu nahe beisammen sein)
			do
			{
				StartOK = true;
				for (i = 0; i < Options.OpponentCnt + 1; i++)
				{
					for (j = 0; j < Options.OpponentCnt + 1; j++)
					{
						if (i != j)
						{
							if (fabs (pPlayer[j].GetPos ().x - pPlayer[i].GetPos ().x) < 20 && fabs (pPlayer[j].GetPos ().y - pPlayer[i].GetPos ().y) < 20)
							{
								StartOK = false;
								break;
							}
						}
					}
				}
				if (!StartOK)
				{
					//Position neu generieren
					for (i = 0; i < Options.OpponentCnt + 1; i++)
						pPlayer[i].GeneratePos ();
				}
			} while (!StartOK);
		}
		if (Multiplayer)
		{
			if (IsServer)
			{
				if (!FirstInit)
				{
					//**********CLIENT_START senden*********
					//Wird im 1. Spiel nicht bentigt, nur in den weiteren.
					SData[0] = CLIENT_START;
					GameServer.SendData (0, SData, 1, 0);
				}
			}
			else
				FreeToRender = false;
		}

		//Variablen setzen
		FirstInit = false;
		CamDir = pPlayer[PIdent].GetDir ();
		CamDir3d = Vector3dMake (0, 0, 0);
		fpsRendered = 0;
		OppShow = -1;
		MopedAlive = Options.OpponentCnt + 1;
		fpsT = 0;
		fps.SetText ("");
		CamBlendTime = 0;
		NotRunTime = 0;
		NotAliveTime = 0;
		BlendTime = 0;
		RDown = false;
		LDown = false;
		CDown = false;
		ESCDown = false;
		ReturnDown = false;
		GameRuns = false;
		GamePassed = false;
		CamMode = 0;
		TimeFac = 1;
		SendT = 0;
		GameQuitQuest = false;
		MPGameBreak = false;

		//Multiplayer-Buffer reseten.
		MPBuf.Reset ();
		RBuf.Reset ();

		Eng3d.SetLookDistance (1000);

		//**************************HAUPTSCHLEIFE FR 1 SPIEL*******************
		TimeOut = 0;
		Eng3d.StartTimer ();
		while (true)
		{
			Eng3d.HandleMessages ();
			
			//Text (fps etc.)
			RelTimeDiff = Eng3d.GetRelTimeDiff ();
			fpsT += RelTimeDiff;
			TimeOut += RelTimeDiff;
			SendT += RelTimeDiff;
			if (fpsT >= 0.5)
			{
				sprintf (f, "%.2f fps", fpsRendered / fpsT);
				fps.SetText (f);
				fpsT = 0;
				fpsRendered = 0;
			}
			fpsRendered++;
			RelTimeDiff *= TimeFac;

			//********************Tastatur und Maus abfragen***********************
			Inp.ReadMouse ();
			RdKeyb = Inp.ReadKeyboard ();
			if (!RdKeyb)
				return;
			//Links-Taste
			if (RdKeyb[DIK_LEFT] && !LDown && pPlayer[PIdent].IsAlive () && GameRuns)
			{
				if (Multiplayer && !IsServer)
				{
					//**********SERVER_EXEC senden*********
					SData[0] = SERVER_EXEC;
					memcpy (&SData[1], &PIdent, sizeof (int));
					SData[1 + sizeof (int)] = 0;
					GameClient.SendData (MPOptions.pIdents[0], SData, 2 + sizeof (int), 0);
					D3DXVec2Subtract (&CamDiff, &pPlayer[PIdent].GetChangeDir (0), &CamDir);
				}
				else
				{
					pPlayer[PIdent].ChangeDir (0);
					D3DXVec2Subtract (&CamDiff, &pPlayer[PIdent].GetDir (), &CamDir);
				}
				PrevDir = CamDir;				
				CamBlendTime = 0.2f;
				LDown = true;
			}
			else if (!RdKeyb[DIK_LEFT])
				LDown = false;
			//Rechts-Taste
			if (RdKeyb[DIK_RIGHT] && !RDown && pPlayer[PIdent].IsAlive () && GameRuns)
			{
				if (Multiplayer && !IsServer)
				{
					//**********SERVER_EXEC senden*********
					SData[0] = SERVER_EXEC;
					memcpy (&SData[1], &PIdent, sizeof (int));
					SData[1 + sizeof (int)] = 1;
					GameClient.SendData (MPOptions.pIdents[0], SData, 2 + sizeof (int), 0);
					D3DXVec2Subtract (&CamDiff, &pPlayer[PIdent].GetChangeDir (1), &CamDir);
				}
				else
				{
					pPlayer[PIdent].ChangeDir (1);
					D3DXVec2Subtract (&CamDiff, &pPlayer[PIdent].GetDir (), &CamDir);
				}
				PrevDir = CamDir;
				CamBlendTime = 0.2f;
				RDown = true;
			}
			else if (!RdKeyb[DIK_RIGHT])
				RDown = false;
			//Leertaste (Item verwenden)
			if (RdKeyb[DIK_SPACE] && pPlayer[PIdent].ItemAviable != -1 && pPlayer[PIdent].IsAlive () && GameRuns)
			{
				if (Multiplayer && !IsServer)
				{
					//**********SERVER_EXEC senden*********
					SData[0] = SERVER_EXEC;
					memcpy (&SData[1], &PIdent, sizeof (int));
					SData[1 + sizeof (int)] = 2;
					GameClient.SendData (MPOptions.pIdents[0], SData, 2 + sizeof (int), 0);
				}
				else
					pPlayer[PIdent].FireItem (&Shots);
			}
			//c-Taste (Kamerawechsel)
			if (RdKeyb[DIK_C] && !CDown && pPlayer[PIdent].IsAlive () && GameRuns)
			{
				Options.Cam++;
				Options.Cam %= 8;
				if (Options.Cam <= 3 || Options.Cam >= 6)
					Eng3d.SetLookDistance (1000, D3DX_PI / 2.5);
				else
					Eng3d.SetLookDistance (1000);
				CDown = true;
				Cam.BlendPosition (&Vector3dMake (0, 0, 0), 0);
			}
			else if (!RdKeyb[DIK_C])
				CDown = false;
			//ESC-Taste
			if (RdKeyb[DIK_ESCAPE])
				ESCDown = true;
			else if (!RdKeyb[DIK_ESCAPE] && ESCDown)
			{
				GameQuitQuest = !GameQuitQuest;
				ESCDown = false;
			}
			//Return (Spiel beenden, sobald es fertig ist)
			if (RdKeyb[DIK_RETURN] || RdKeyb[DIK_NUMPADENTER])
				ReturnDown = true;
			else if (!RdKeyb[DIK_RETURN] && !RdKeyb[DIK_NUMPADENTER] && ReturnDown)
			{
				if (GameQuitQuest)
				{
					if (Multiplayer)
					{
						//Den anderen benachrichtigen, dass die Verbindung getrennt wird.
						SData[0] = CONNECTION_LOST;
						if (IsServer)
							GameServer.SendData (0, SData, 1, 0);
						else
							GameClient.SendData (0, SData, 1, 0);
					}
					delete []pPlayer;
					return;
				}
				if (GamePassed && (!Multiplayer || IsServer))
					//CLIENT_START wird in der usseren Schleife an alle Clients gesendet.
					break;
				ReturnDown = false;
			}
			if (MPGameBreak)
				//Client eines Multiplayer-Spiels, beginnen des nchsten Spiels
				break;


			////////////////////G A M E   C O N T R O L L E R////////////////////
			if (!pPlayer[PIdent].IsAlive ())
				NotAliveTime += RelTimeDiff;

			if (!Multiplayer || RData[0] != CONNECTION_LOST)
			{
				Explosions.DoEvents (RelTimeDiff);
				if (!Multiplayer || IsServer)
					MasterGameController (RelTimeDiff, Multiplayer);
			}

			
			////////////////////N E T W O R K I N G///////////////////////////
			if (Multiplayer && RData[0] != CONNECTION_LOST)
			{
				if (IsServer && SendT >= 0.02f)
				{
					//Um das Netzwerk nicht zu berlasten, mit maximal 50 fps
					//die Daten zu den Clients bertragen.
					SData[0] = CLIENT_GAMESTATE;
					//Spiel-Variablen
					pMem = &SData[1];
					memcpy (pMem, &MopedAlive, sizeof (int));
					pMem += sizeof (int);
					memcpy (pMem, &LeaderScore, sizeof (int));
					pMem += sizeof (int);
					memcpy (pMem, &NotRunTime, sizeof (float));
					pMem += sizeof (float);
					memcpy (pMem, &GameRuns, sizeof (bool));
					pMem += sizeof (bool);
					memcpy (pMem, &GamePassed, sizeof (bool));
					pMem += sizeof (bool);
					//Player
					for (i = 0; i < Options.OpponentCnt + 1; i++)
						pPlayer[i].WriteData (&pMem);
					//Shots
					Shots.WriteData (&pMem);
					//Walls
					WallDesc.WriteData (&pMem);
					
					GameServer.SendData (0, SData, pMem - SData, 0.02f);
					SendT = 0;
				}
				else if (!IsServer && RBuf.NewData)
				{
					//Wir sind Client und haben neue Render-Daten bekommen.					
					//Render-Buffer auslesen
					RDataSize = RBuf.GetDataSize ();
					if (RDataSize != 0)
					{
						RBuf.GetData (pRData);
						
						//******************************Buffer auslesen******************************
						//Spiel-Variablen
						pMem = pRData;
						memcpy (&MopedAlive, pMem, sizeof (int));
						pMem += sizeof (int);
						memcpy (&LeaderScore, pMem, sizeof (int));
						pMem += sizeof (int);
						memcpy (&NotRunTime, pMem, sizeof (float));
						pMem += sizeof (float);
						memcpy (&GameRuns, pMem, sizeof (bool));
						pMem += sizeof (bool);
						memcpy (&GamePassed, pMem, sizeof (bool));
						pMem += sizeof (bool);
						//Player
						for (i = 0; i < Options.OpponentCnt + 1; i++)
							pPlayer[i].ReadData (&pMem);
						//Shots
						Shots.ReadData (&pMem);
						//Walls
						WallDesc.ReadData (&pMem);

						FreeToRender = true;
					}
					RBuf.NewData = false;
				}
				if (!IsServer)
				{
					if (TimeOut > 3)
						RData[0] = CONNECTION_LOST;
					else if (!FreeToRender)
						//Der Server hat uns (dem Client) noch keine Render-Daten zur Verfgung gestellt.
						//Dies kann hchstens ein paar Frames lang ganz zu Beginn eines Spiels passieren.
						continue;					
				}
			}
			if (RData[0] == CONNECTION_LOST && Multiplayer && !IsServer)
				GameClient.Disconnect ();


			////////////////////R E N D E R E R////////////////////
			//******************Kamera einstellen****************
			ActPos = pPlayer[PIdent].GetPos ();
			
			if (CamMode == 0 && NotRunTime >= 3)
				CamMode = 1;
			else if (CamMode == 2 && NotRunTime >= 5)
				CamMode = 3;			
			else if (CamMode == 3 && !pPlayer[PIdent].IsAlive ())
				CamMode = 4;
			else if (CamMode == 5 && NotAliveTime >= 3)
				CamMode = 6;
			else if (CamMode == 7 && BlendTime >= 1.5f)
				CamMode = 6;
			if (CamMode == 0)
			{
				//Spiel luft noch nicht, nahe Kamera am Moped zum Zeigen der Grafik.
				Cam.SetDest (&Vector3dMake (ActPos.x, 0, ActPos.y));
				CamVec = NormalVector2dMake (&Vector2dMake (0, 0), &pPlayer[PIdent].GetDir ());					
				Cam.SetPos (&Vector3dMake (ActPos.x - CamVec.x * 7, 6 - NotRunTime, ActPos.y - CamVec.y * 7));
			}
			else if (CamMode == 1)
			{
				//Spiel luft noch nicht, berblenden der nahen Kamera zur Spiel-Kamera.
				GetCamDesc (ActPos, pPlayer[PIdent].GetDir (), &NPos, &NDest);
				Cam.BlendPosition (&NPos, 2);
				Cam.BlendDestination (&NDest, 2);
				if (Options.Cam <= 3 || Options.Cam >= 6)
					Cam.BlendAngle (D3DX_PI / 4, D3DX_PI / 2.5, 2);
				CamMode = 2;
			}
			//CamMode == 2 wird zum berblenden verwendet.
			else if (CamMode == 3)
			{
				//Spiel luft ganz normal, entsprechend den Optionen die Kamera einstellen.
				//Kamera-Richtung beim Lenken vom Moped anpassen
				if (CamBlendTime > 0)
				{
					//Kamerarichtung anpassen
					D3DXVec2Scale (&CamDir, &CamDiff, 1 - CamBlendTime / 0.2f);
					D3DXVec2Add (&CamDir, &PrevDir, &CamDir);
					CamBlendTime -= RelTimeDiff;
				}
				else
					//Kamerarichtung vom Moped verwenden
					CamDir = pPlayer[PIdent].GetDir ();
				GetCamDesc (ActPos, CamDir, &NPos, &NDest);
				if ((Options.Cam == 4 || Options.Cam == 5) && !CDown)
				{
					//Wenn nicht gerade Kamerawechsel (c-Taste), auf neue Position blenden.
					//Dies bewirkt, dass die Kamera nicht extrem schnell dreht.
					Cam.BlendPosition (&NPos, 0.25f);
					Cam.SetDest (&NDest);
				}
				else
				{
					//Sofort auf neue Position setzen
					Cam.SetPos (&NPos);
					Cam.SetDest (&NDest);
				}
			}
			else if (CamMode == 4)
			{
				//Der Spieler ist gerade gestorben, Kamera blenden.
				Cam.BlendPosition (&Vector3dMake (Options.LandSize / 2, 40, Options.LandSize / 2), 3);
				if (Options.Cam <= 3 || Options.Cam >= 6)
					Cam.BlendAngle (D3DX_PI / 2.5, D3DX_PI / 4, 3);
				CamMode = 5;
			}
			//CamMode == 5 wird zum berblenden verwendet.
			else if (CamMode == 6)
			{
				//Kamera am Schluss auf ein anderes Moped richten
				if (OppShow == -1 || !pPlayer[OppShow].IsAlive ())
				{
					//Neues Moped anvisieren
					for (i = 0; i < Options.OpponentCnt + 1; i++)
					{
						if (i != OppShow && pPlayer[i].IsAlive ())
						{
							if (OppShow != -1)
							{
								BlendTime = 0;
								CamMode = 7;
							}
							OppShow = i;
							break;
						}
					}
				}
				else if (OppShow != -1)
				{
					if (TimeFac == 1 && !Multiplayer)
					{
						//Spezial-Kamera am Schluss verwenden, zum Zeigen der Grafik.
						D3DXVec3Subtract (&NDest, &Vector3dMake (pPlayer[OppShow].GetPos ().x - pPlayer[OppShow].GetDir ().x * 5, GetRandomVal (10, 30), pPlayer[OppShow].GetPos ().y - pPlayer[OppShow].GetDir ().y * 5), Cam.GetPos ());
						D3DXVec3Normalize (&NDest, &NDest);
						D3DXVec3Lerp (&CamDir3d, &CamDir3d, &NDest, RelTimeDiff * 3);		//= Trgheit von 1/3 Sekunde
						D3DXVec3Scale (&CamDir3d, &CamDir3d, Options.MopedSpeed / 1.25f * RelTimeDiff);
						D3DXVec3Add (&NPos, Cam.GetPos (), &CamDir3d);
						D3DXVec3Normalize (&CamDir3d, &CamDir3d);
						if (NPos.y < 10)
							NPos.y = 10;
						Cam.SetPos (&NPos);
					}
					Cam.SetDest (&Vector3dMake (pPlayer[OppShow].GetPos ().x, 0, pPlayer[OppShow].GetPos ().y));
				}
			}
			else if (CamMode == 7)
				//Statische Kamera, wenn ein Gegner explodiert.
				BlendTime += RelTimeDiff;
			Cam.DoEvents (RelTimeDiff);

			//****************Soundposition einstellen**************
			EngSound.SetListener3dPosition (Cam.GetPos (), Cam.GetDest ());
			EngSound.Update3dSounds ();

			//****************Lichtpositionen einstellen*************
			PointLight.Position = *Cam.GetPos ();
			Eng3d.SetLight (1, &PointLight);

			//****************Rendern*************
			Eng3d.RenderStart (NULL, NULL);
			if (Options.UseShadow == 2)
				Eng3d.RenderClear (true, ColorMake (0, 0, 0), true, true);
			else
				Eng3d.RenderClear (true, ColorMake (0, 0, 0), true, false);
			if ((Options.Cam <= 3 || Options.Cam >= 6) && Cam.IsAngleBlended ())
				Eng3d.SetLookDistance (1000, Cam.GetAngle ());
			Cam.SetCamera ();
			//Spiegelung rendern, falls eingestellt
			if (Options.UseReflection)
			{
				Eng3d.SetReflectionState (&FloorPlane, false);
				//Wnde rendern
				Eng3d.SetGlobalLight (ColorMake (200, 200, 200));
				Eng3d.SetLightState (0, false);
				Eng3d.SetLightState (1, false);
				Eng3d.SetZState (false, false);
				GameLand.SetTexture (&TexWall);
				Eng3d.Render3dPolygons (&GameLand, 4, 2);
				Eng3d.Render3dPolygons (&GameLand, 8, 2);
				Eng3d.Render3dPolygons (&GameLand, 12, 2);
				Eng3d.Render3dPolygons (&GameLand, 16, 2);
				Eng3d.SetZState (true, true);
				//Mopeds rendern
				Eng3d.SetGlobalLight (ColorMake (0, 0, 0));
				Eng3d.SetLightState (0, true);
				Eng3d.SetLightState (1, true);
				Eng3d.SetSpecularState (Options.UseSpecular);
				for (i = 0; i < Options.OpponentCnt + 1; i++)
					pPlayer[i].Render (false, &Shield);
				Eng3d.SetLightState (1, false);
				Eng3d.SetSpecularState (false);
				//Wnde von den Mopeds rendern
				for (i = 0; i < Options.OpponentCnt + 1; i++)
					pPlayer[i].RenderWalls (true);
				Eng3d.SetLightState (0, false);
				//Schsse rendern
				Eng3d.SetGlobalLight (ColorMake (255, 255, 255));
				Shots.Render (false);
				Eng3d.SetReflectionState (NULL);
			}
			Eng3d.SetLightState (0, false);
			Eng3d.SetLightState (1, false);
			//Bodenflche rendern
			if (Options.UseShadow == 0)							//sonst wrde man den Schatten unterhalb der Landschaft auf der Mauer sehen
				Eng3d.SetZState (false, false);
			GameLand.SetTexture (&TexFloor);
			if (Options.UseReflection)
			{
				//transparenten Boden rendern
				Eng3d.SetGlobalLight (ColorMake (255, 255, 255));
				GameLand.SetMaterial (&FloorMat);
				Eng3d.Render3dPolygons (&GameLand, 0, 2);
				Eng3d.SetGlobalLight (ColorMake (200, 200, 200));
				GameLand.SetMaterial (&Eng3d.GetDefaultMaterial ());
			}
			else
			{
				Eng3d.SetGlobalLight (ColorMake (200, 200, 200));
				Eng3d.Render3dPolygons (&GameLand, 0, 2);
			}
			//Mauern von der Landschaft rendern
			GameLand.SetTexture (&TexWall);
			Eng3d.Render3dPolygons (&GameLand, 4, 2);
			Eng3d.Render3dPolygons (&GameLand, 8, 2);
			Eng3d.Render3dPolygons (&GameLand, 12, 2);
			Eng3d.Render3dPolygons (&GameLand, 16, 2);
			//Schatten rendern (am Boden und an den Mauern)
			if (Options.UseShadow)
			{
				Eng3d.SetZState (true, false);
				//Schatten am Boden
				if (Options.UseShadow == 1)
					SUse = Eng3d.SetShadowState (&FloorPlane, &Vector3dMake (CamVec.x * 10, 9, CamVec.y * 10), 255, DIRECTIONALLIGHT, false);
				else if (Options.UseShadow == 2)
					SUse = Eng3d.SetShadowState (&FloorPlane, &Vector3dMake (CamVec.x * 10, 9, CamVec.y * 10), 100, DIRECTIONALLIGHT, false);
				if (SUse)
				{
					for (i = 0; i < Options.OpponentCnt + 1; i++)
					{
						pPlayer[i].Render (true, NULL);
						pPlayer[i].RenderWalls (false);
					}
				}
				Eng3d.SetShadowState (NULL, NULL, 0);
			}
			Eng3d.SetZState (true, true);
			//Schsse rendern
			Shots.Render (true);
			//Mopeds rendern
			Eng3d.SetSpecularState (Options.UseSpecular);
			Eng3d.SetGlobalLight (ColorMake (0, 0, 0));
			Eng3d.SetLightState (0, true);
			Eng3d.SetLightState (1, Options.Cam != 6);
			pPlayer[PIdent].Render (false, &Shield);
			Eng3d.SetLightState (1, true);
			for (i = 0; i < Options.OpponentCnt + 1; i++)
			{
				if (i != PIdent)
					pPlayer[i].Render (false, &Shield);
			}
			//Wnde von den Mopeds rendern
			Eng3d.SetLightState (1, false);
			Eng3d.SetSpecularState (false);
			for (i = 0; i < Options.OpponentCnt + 1; i++)
				pPlayer[i].RenderWalls (false);
			Eng3d.SetLightState (0, false);
			//Explosionen rendern
			Eng3d.SetGlobalLight (ColorMake (255, 255, 255));
			Explosions.Render (Cam.GetPos ());
			if (Options.EnableAction && pPlayer[PIdent].IsAlive () && pPlayer[PIdent].ItemAviable != -1)
			{
				//Item-Bild rendern
				if (pPlayer[PIdent].ItemAviable == 0)
					Eng3d.Render2dPicture (&PicItem, &TexSpeedup);
				else if (pPlayer[PIdent].ItemAviable == 1)
					Eng3d.Render2dPicture (&PicItem, &TexShotGreen);
				else if (pPlayer[PIdent].ItemAviable == 2)
					Eng3d.Render2dPicture (&PicItem, &TexShotRed);
				else if (pPlayer[PIdent].ItemAviable == 3)
					Eng3d.Render2dPicture (&PicItem, &TexShotBlack);
				else if (pPlayer[PIdent].ItemAviable == 4)
					Eng3d.Render2dPicture (&PicItem, &TexShield);
			}
			//Textausgaben rendern
			Eng3d.Render2dText (&fps);
			if (GameQuitQuest)
			{
				TextRank.SetFormat (1, 1);
				TextRank.SetPosition (0, Options.Res.ResY - 20, Options.Res.ResX, Options.Res.ResY);
				TextRank.SetColor (ColorMake (255, 255, 255));
				TextRank.SetText ("Really quit? Return = yes, ESC = no");
				Eng3d.Render2dText (&TextRank);
			}
			if (GamePassed)
			{
				//Ranking-Tabelle rendern, falls der Spieler schon gestorben ist.
				TextRank.SetFormat (1, 1);
				RankLine = 0;
				if (LeaderScore >= Options.FragLimit)
				{
					TextRank.SetPosition (0, 20, Options.Res.ResX, 50);
					TextRank.SetColor (ColorMake (255, 0, 0));
					TextRank.SetText ("FINAL RESULTS");
					Eng3d.Render2dText (&TextRank);
				}
				for (i = 0; i < Options.OpponentCnt + 1; i++)
				{
					for (j = 0; j < Options.OpponentCnt + 1; j++)
					{
						if (pPlayer[j].Rank == i + 1)
						{
							TextRank.SetPosition (Options.Res.ResX / 5, 100 + RankLine * 30, Options.Res.ResX * 4 / 5, 100 + (RankLine + 1) * 30);
							if (j == PIdent)
								TextRank.SetColor (ColorMake (255, 255, 0));
							else
								TextRank.SetColor (ColorMake (255, 255, 255));
							sprintf (f, "%d.: %s with %d points", pPlayer[j].Rank, pPlayer[j].Name, pPlayer[j].Frags);
							TextRank.SetText (f);
							Eng3d.Render2dText (&TextRank);
							RankLine++;
						}
					}
				}
				TextRank.SetColor (ColorMake (255, 255, 255));
				if (!GameQuitQuest)
				{
					//Dies nur dann rendern, wenn hier nicht bereits die Beenden-Frage steht.
					TextRank.SetPosition (0, Options.Res.ResY - 20, Options.Res.ResX, Options.Res.ResY);
					if (!Multiplayer || IsServer)
						TextRank.SetText ("Press return to continue");
					else
						TextRank.SetText ("Waiting for server...");
					Eng3d.Render2dText (&TextRank);
				}
			}
			else
			{
				if (pPlayer[PIdent].Rank == 1)
				{
					int SecondScore = -1;

					for (i = 0; i < Options.OpponentCnt + 1; i++)
					{
						if (i != PIdent && pPlayer[i].Frags > SecondScore)
							SecondScore = pPlayer[i].Frags;
					}
					if (SecondScore == pPlayer[PIdent].Frags)
						sprintf (f, "Points: %d (+0)", pPlayer[PIdent].Frags);
					else
						sprintf (f, "Points: %d (%d)", pPlayer[PIdent].Frags, SecondScore - LeaderScore);
				}
				else
					sprintf (f, "Points: %d (+%d)", pPlayer[PIdent].Frags, LeaderScore - pPlayer[PIdent].Frags);
				TextRank.SetPosition (0, Options.Res.ResY - 20, Options.Res.ResX, Options.Res.ResY);
				TextRank.SetFormat (0, 1);
				TextRank.SetText (f);
				Eng3d.Render2dText (&TextRank);
			}
			if (Multiplayer && RData[0] == CONNECTION_LOST)
				Eng3d.Render2dText (&TextLost);
			Eng3d.RenderEnd ();
			Eng3d.RenderShow ();
		}
	}
	delete []pPlayer;
}



//Sortiert das Feld nach der Anzahl der Frags.
int DoRanking ()
{
	int i, j, min, MaxScore;
	int ActRank = 0, AddRank = 0;
	TronPlayer *pSwap, **pField;

	pField = new TronPlayer *[Options.OpponentCnt + 1];
	for (i = 0; i < Options.OpponentCnt + 1; i++)
		pField[i] = &pPlayer[i];
	//Absteigend sortieren (durch direkte Auswahl)
	for (i = 0; i < Options.OpponentCnt; i++)
	{
		min = i;
		for (j = i + 1; j < Options.OpponentCnt + 1; j++)
		{
			if (pField[min]->Frags < pField[j]->Frags)
				min = j;
		}
		if (min != i)
		{
			//Vertauschen
			pSwap = pField[min];
			pField[min] = pField[i];
			pField[i] = pSwap;
		}
	}
	MaxScore = pField[0]->Frags;
	//Jetzt die Rankings in die Tron-Player schreiben
	for (i = 0; i <= Options.OpponentCnt; i++)
	{
		AddRank++;
		if (i == 0 || pField[i]->Frags < pField[i - 1]->Frags)
		{
			//Hat weniger Frags als der vorige, daher Ranking erhhen
			ActRank += AddRank;
			AddRank = 0;
		}
		pField[i]->Rank = ActRank;
	}
	delete []pField;
	return (MaxScore);
}


CameraHandler::CameraHandler ()
{
	PosTime = 0;
	DestTime = 0;
	AngleTime = 0;
}

void CameraHandler::BlendPosition (D3DXVECTOR3 *NewPos, float Time)
{
	D3DXVec3Subtract (&BlendPos, NewPos, &CamPos);
	D3DXVec3Scale (&BlendPos, &BlendPos, 1 / Time);
	PosTime = Time;
}

void CameraHandler::BlendDestination (D3DXVECTOR3 *NewDest, float Time)
{
	D3DXVec3Subtract (&BlendDest, NewDest, &CamDest);
	D3DXVec3Scale (&BlendDest, &BlendDest, 1 / Time);
	DestTime = Time;
}

void CameraHandler::BlendAngle (float ActAngle, float NewAngle, float Time)
{
	CamAng = ActAngle;
	BlendAng = (NewAngle - ActAngle) / Time;
	AngleTime = Time;
}

void CameraHandler::DoEvents (float RelTimeDiff)
{
	float BlendTime;
	D3DXVECTOR3 AddVec;
	
	if (PosTime > 0)
	{
		//Kamera-Position blenden
		if (RelTimeDiff > PosTime)
			BlendTime = PosTime;
		else
			BlendTime = RelTimeDiff;
		D3DXVec3Scale (&AddVec, &BlendPos, BlendTime);
		D3DXVec3Add (&CamPos, &CamPos, &AddVec);
		PosTime -= BlendTime;
	}
	if (DestTime > 0)
	{
		//Kamera-Destination blenden
		if (RelTimeDiff > DestTime)
			BlendTime = DestTime;
		else
			BlendTime = RelTimeDiff;
		D3DXVec3Scale (&AddVec, &BlendDest, BlendTime);
		D3DXVec3Add (&CamDest, &CamDest, &AddVec);
		DestTime -= BlendTime;
	}
	if (AngleTime > 0)
	{
		//Kamera-Winkel blenden
		if (RelTimeDiff > AngleTime)
			BlendTime = AngleTime;
		else
			BlendTime = RelTimeDiff;
		CamAng += BlendAng * BlendTime;
		AngleTime -= BlendTime;
	}
}

void CameraHandler::SetCamera ()
{
	D3DXVECTOR3 SetPos;
	
	SetPos = CamPos;
	if (SetPos.x < 1)
		SetPos.x = 1;
	if (SetPos.x > Options.LandSize - 1)
		SetPos.x = Options.LandSize - 1;
	if (SetPos.z < 1)
		SetPos.z = 1;
	if (SetPos.z > Options.LandSize - 1)
		SetPos.z = Options.LandSize - 1;
	Eng3d.SetCamera (&SetPos, &CamDest, 0);
}



ActionBuffer::ActionBuffer ()
{
	Reset ();
}

void ActionBuffer::Reset ()
{
	ActionCnt = 0;
}

void ActionBuffer::GetActions (int *DestIdents, BYTE *DestActions)
{
	memcpy (DestIdents, Idents, ActionCnt * sizeof (int));
	memcpy (DestActions, Actions, ActionCnt * sizeof (BYTE));
	ActionCnt = 0;
}

void ActionBuffer::AddAction (int Ident, BYTE Action)
{
	if (ActionCnt == 100)
		return;
	Idents[ActionCnt] = Ident;
	Actions[ActionCnt] = Action;
	ActionCnt++;
}



RenderBuffer::RenderBuffer ()
{
	Reset ();
}

void RenderBuffer::Reset ()
{
	DataSize = 0;
	NewData = false;
}

void RenderBuffer::GetData (BYTE *Dest)
{
	memcpy (Dest, pRenderData, DataSize);
}

void RenderBuffer::SetData (BYTE *Src, int Size)
{
	if (NewData)
		return;
	memcpy (pRenderData, Src, Size);
	DataSize = Size;
	NewData = true;
}



//Callback-Routine fr den Server
void SCallback (BYTE *pData, int DataSize)
{
	int RdInt;
	BYTE RdByte;

	if (!pData)
	{
		//Die Verbindung wurde bzw ist getrennt.
		RData[0] = CONNECTION_LOST;
		return;
	}
	if (pData[0] == SERVER_SETTINGSOK)
	{
		//MP-Einstellungen wurden von einem Client angenommen
		//Client-Bikenummer lesen
		memcpy (&RdInt, &pData[1], sizeof (int));
		//Client-Name dieser Nummer schreiben
		memcpy (MPBikeInfo[RdInt].Name, &pData[1 + sizeof (int)], 64 * sizeof (char));
		ClientReady++;
	}
	else if (pData[0] == SERVER_LOADOK)
		//Ladevorgang am Client wurde abgeschlossen
		ClientReady++;
	else if (pData[0] == SERVER_EXEC)
	{
		//Ein Client will eine Aktion durchfhren.
		//Dies ist entweder Lenken oder Verwenden eines Items im Action-Modus.
		//Client-Bikenummer lesen
		memcpy (&RdInt, &pData[1], sizeof (int));
		//Aktions-Nummer lesen
		RdByte = pData[1 + sizeof (int)];
		MPBuf.AddAction (RdInt, RdByte);
	}
	memcpy (RData, pData, DataSize);
}

//Callback-Routine fr den Client
void CCallback (BYTE *pData, int DataSize)
{
	int i;
	D3DXVECTOR2 RdVec;

	if (!pData)
	{
		//Die Verbindung wurde bzw ist getrennt.
		RData[0] = CONNECTION_LOST;
		return;
	}
	if (pData[0] == CLIENT_LOAD)
	{
		//Spieler-Namen auslesen
		for (i = 0; i < Options.OpponentCnt + 1; i++)
			memcpy (MPBikeInfo[i].Name, &pData[1 + i * 64 * sizeof (char)], 64 * sizeof (char));
	}
	else if (pData[0] == CLIENT_START)
		MPGameBreak = true;
	else if (pData[0] == CLIENT_GAMESTATE)
	{
		//Render-Daten empfangen
		RBuf.SetData (&pData[1], DataSize - 1);
		TimeOut = 0;
	}
	else if (pData[0] == CLIENT_PLAYSOUND)
	{
		if (Options.UseSounds)
		{
			//Sound abspielen.
			memcpy (&RdVec, &pData[2], sizeof (D3DXVECTOR2));
			if (pData[1] == 0)
			{
				//Bike-Crash
				BikeCrashSound.Set3dPosition (&Vector3dMake (RdVec.x, 0, RdVec.y));
				BikeCrashSound.SetPlayPosition (0);
				EngSound.Update3dSounds ();
				EngSound.Play3dSound (&BikeCrashSound, false);
			}
			else if (pData[1] == 1)
			{
				//Shot-Crash
				ShotCrashSound.Set3dPosition (&Vector3dMake (RdVec.x, 0, RdVec.y));
				ShotCrashSound.SetPlayPosition (0);
				EngSound.Update3dSounds ();
				EngSound.Play3dSound (&ShotCrashSound, false);
			}
			else if (pData[1] == 2)
			{
				//Fire shot
				FireShotSound.Set3dPosition (&Vector3dMake (RdVec.x, 0, RdVec.y));
				FireShotSound.SetPlayPosition (0);
				EngSound.Update3dSounds ();
				EngSound.Play3dSound (&FireShotSound, false);
			}
		}
	}
	else if (pData[0] == CLIENT_EXPLOSION)
	{
		//Explosion erstellen
		memcpy (&RdVec, &pData[2], sizeof (D3DXVECTOR2));
		if (pData[1])
			//kleine Explosion
			Explosions.CreateExplosion (&RdVec, true, false);
		else
			//groe Explosion
			Explosions.CreateExplosion (&RdVec, false, false);
	}

	if (pData[0] == CLIENT_GAMESTATE && !GamePassed)
		//Solange das Spiel luft, nur das 1. Byte kopieren, da das Auslesen der Daten
		//ber den Render-Buffer funktioniert.
		RData[0] = CLIENT_GAMESTATE;
	else if (pData[0] != CLIENT_GAMESTATE)
		//Nachricht in den lokalen Speicher kopieren.
		memcpy (RData, pData, DataSize);
}


void GetCamDesc (D3DXVECTOR2 ActPos, D3DXVECTOR2 CamDir, D3DXVECTOR3 *CamPos, D3DXVECTOR3 *CamDest)
{
	D3DXVECTOR2 CamVec;

	if (Options.Cam == 0)
	{
		*CamDest = Vector3dMake (ActPos.x + CamDir.x * 30, 0, ActPos.y + CamDir.y * 30);
		*CamPos = Vector3dMake (ActPos.x - CamDir.x * 10, 6, ActPos.y - CamDir.y * 10);				//nahe Back-Kamera
	}
	else if (Options.Cam == 1)
	{
		*CamDest = Vector3dMake (ActPos.x + CamDir.x * 30, 0, ActPos.y + CamDir.y * 30);
		*CamPos = Vector3dMake (ActPos.x - CamDir.x * 20, 10, ActPos.y - CamDir.y * 20);			//weite Back-Kamera
	}
	else if (Options.Cam == 2)
	{
		*CamDest = Vector3dMake (ActPos.x + CamDir.x * 30, 0, ActPos.y + CamDir.y * 30);
		*CamPos = Vector3dMake (ActPos.x - CamDir.x * 30, 15, ActPos.y - CamDir.y * 30);			//sehr weite Back-Kamera
	}
	else if (Options.Cam == 3)
	{
		*CamDest = Vector3dMake (ActPos.x + CamDir.x * 20, 0, ActPos.y + CamDir.y * 20);
		*CamPos = Vector3dMake (ActPos.x + CamDir.x * 15, 75, ActPos.y + CamDir.y * 15);			//klassische Top-Down
	}
	else if (Options.Cam == 4)
	{																								//niedrige dynamische Top-Down
		CamVec = Vector2dMake (Options.LandSize / 2 - ActPos.x, Options.LandSize / 2 - ActPos.y);
		D3DXVec2Normalize (&CamVec, &CamVec);
		D3DXVec2Scale (&CamVec, &CamVec, 100);
		*CamDest = Vector3dMake (ActPos.x, 0, ActPos.y);
		*CamPos = Vector3dMake (ActPos.x + CamVec.x, 50, ActPos.y + CamVec.y);
	}
	else if (Options.Cam == 5)
	{																								//hohe dynamische Top-Down
		CamVec = Vector2dMake (Options.LandSize / 2 - ActPos.x, Options.LandSize / 2 - ActPos.y);
		D3DXVec2Normalize (&CamVec, &CamVec);
		D3DXVec2Scale (&CamVec, &CamVec, 100);
		*CamDest = Vector3dMake (ActPos.x, 0, ActPos.y);
		*CamPos = Vector3dMake (ActPos.x + CamVec.x, 100, ActPos.y + CamVec.y);
	}
	else if (Options.Cam == 6)
	{
		*CamDest = Vector3dMake (ActPos.x, 0, ActPos.y);
		*CamPos = Vector3dMake (ActPos.x - CamDir.x * 2, 2, ActPos.y - CamDir.y * 2);				//Auf Moped
		D3DXVec3Add (CamDest, CamPos, &Vector3dMake (CamDir.x * 15, 0, CamDir.y * 15));
	}
	else if (Options.Cam == 7)
	{
		*CamDest = Vector3dMake (ActPos.x, 0, ActPos.y);
		*CamPos = Vector3dMake (ActPos.x + CamDir.x * 3, 1, ActPos.y + CamDir.y * 3);				//Stostange
		D3DXVec3Add (CamDest, CamPos, &Vector3dMake (CamDir.x, 0, CamDir.y));
	}
}

void InitRandom ()
{
	int i;
	unsigned long tm = (unsigned long) time (NULL);

	for (i = 0; i < 3; i++)
		tm *= tm;
	srand (tm);
}

float GetRandomVal (float Min, float Max)
{
	float val;

	val = rand () / (float) RAND_MAX;
	val *= Max - Min;
	val += Min;
	return (val);
}

bool InArea (float Min, float Max, float Val)
{
	if ((Val > Min && Val < Max) || (Val > Max && Val < Min))
		return (true);
	return (false);
}

bool InAreaDelta (float Min, float Max, float Val, float Delta)
{
	if (InArea (Min - Delta, Max + Delta, Val) || InArea (Min + Delta, Max - Delta, Val))
		return (true);
	return (false);
}

bool InLine (float v11, float v12, float v21, float v22)
{
	if (InArea (v11, v12, v21) || InArea (v11, v12, v22))
		return (true);
	if (InArea (v21, v22, v11) || InArea (v21, v22, v12))
		return (true);
	return (false);
}





//**********************TRON PLAYER***************
bool TronPlayer::Initialize (int MopedNr, TronWallDesc *pWallDesc, int Ident, bool Multiplayer, bool IsServer, Revo2dTexture *pWallTex)
{
	int DirVal;
	char File[128];

	//Moped
	Moped.Initialize (&Eng3d);
	Wheel.Initialize (&Eng3d);
	SWheel.Initialize (&Eng3d);
	//3d-Daten laden
	if (Multiplayer)
		//Die Mopeds von MPBikeInfo verwenden.
		MopedNr = MPBikeInfo[Ident].BikeType;
	sprintf (File, "objects\\moped%d.x", MopedNr);
	if (!Moped.LoadFromFile (File, 1))
		return (false);
	if (!Wheel.LoadFromFile ("objects\\wheel.x", 1))
		return (false);
	if (!SWheel.LoadFromFile ("objects\\swheel.x", 1))
		return (false);
	Moped.SetSpecPower (10);
	Wheel.SetSpecPower (10);
	if (!Multiplayer || IsServer)
	{
		DirVal = (int) GetRandomVal (0, 4);
		ZeroMemory (&Dir, sizeof (D3DXVECTOR2));
		if (DirVal == 0)
		{
			Dir.x = 1;
			YAngle = D3DX_PI * 1.5f;
		}
		else if (DirVal == 1)
		{
			Dir.x = -1;
			YAngle = D3DX_PI / 2.0f;
		}
		else if (DirVal == 2)
		{
			Dir.y = 1;
			YAngle = D3DX_PI;
		}
		else
		{
			Dir.y = -1;
			YAngle = 0;
		}
		RotationLeft = 0;
		Moped.Rotate (&Vector3dMake (0, YAngle, 0), NULL, &Vector3dMake (0, 0, 0));
		Wheel.Rotate (&Vector3dMake (0, YAngle, 0), NULL, &Vector3dMake (0, 0, 0));
	}
	//Wnde des Mopeds
	this->pWallDesc = pWallDesc;
	if (MopedNr == 0)
		Wall.Initialize (200, 0, 0, 200, pWallTex);
	else if (MopedNr == 1)
		Wall.Initialize (200, 200, 0, 0, pWallTex);
	else if (MopedNr == 2)
		Wall.Initialize (200, 50, 50, 50, pWallTex);
	else
		Wall.Initialize (200, 200, 200, 0, pWallTex);
	//Name
	if (Multiplayer)
		//Die Namen von MPBikeInfo verwenden.
		strcpy (Name, MPBikeInfo[Ident].Name);
	else
	{
		if (Ident == 0)
			strcpy (Name, Options.PlayerName);
		else if (Ident == 1)
			strcpy (Name, "Gates");
		else if (Ident == 2)
			strcpy (Name, "Elvis");
		else if (Ident == 3)
			strcpy (Name, "Bush");
		else if (Ident == 4)
			strcpy (Name, "Britney");
		else if (Ident == 5)
			strcpy (Name, "Zidane");
	}
	Frags = 0;
	Reset ();
	this->Ident = Ident;
	MPSrv = Multiplayer && IsServer;
	return (true);
}

void TronPlayer::Reset ()
{
	//Items
	ItemAviable = -1;
	ItemTime = GetRandomVal (10, 15);

	Wall.AddWall ();
	WallDescIndex = pWallDesc->AddWall (Ident);

	//Allgemein
	GeneratePos ();
	Wall.UpdateWall (&Pos, &Pos);
	pWallDesc->UpdateWall (&Pos, &Pos, WallDescIndex);
	Alive = true;
	WillCrash = false;
	SpeedTime = 0;
	ShieldTime = 0;
	Rank = 1;
	AIReact = 0;
	AILastCh = 1;
	AILastHook = 1;
}

void TronPlayer::GeneratePos ()
{
	Pos.x = GetRandomVal (20, Options.LandSize - 20);
	Pos.y = GetRandomVal (20, Options.LandSize - 20);
	PrevPos = Pos;
	ActWallPos.x = Pos.x - Dir.x * 1.6f;
	ActWallPos.y = Pos.y - Dir.y * 1.6f;
	LastWallPos = ActWallPos;
	LastOWPos = ActWallPos;
}

void TronPlayer::Crash ()
{
	if (!Alive)
		return;
	DoEvents (30, 0);		//Rotation fertig drehen lassen, etc.
	pWallDesc->DeletePlayer (Ident);
	Alive = false;
}

void TronPlayer::Delete ()
{
	Wall.Delete ();
}

D3DXVECTOR2 TronPlayer::GetChangeDir (int TurnDir)
{
	if (TurnDir == 0)
	{
		//Links
		if (Dir.y)
			return (Vector2dMake (-Dir.y, 0));
		else
			return (Vector2dMake (0, Dir.x));
	}
	else
	{
		//Rechts
		if (Dir.y)
			return (Vector2dMake (Dir.y, 0));
		else
			return (Vector2dMake (0, -Dir.x));
	}
}

void TronPlayer::ChangeDir (int TurnDir)
{
	D3DXVECTOR3 NVec;

	if (!Alive)
		return;	
	if (TurnDir == 0)
	{
		//Links
		if (Dir.y)
		{
			Dir.x = -Dir.y;
			Dir.y = 0;
		}
		else
		{
			Dir.y = Dir.x;
			Dir.x = 0;
		}
		RotationLeft -= D3DX_PI / 2.0f;
		if (RotationLeft < 2 * -D3DX_PI)
			RotationLeft += 2 * D3DX_PI;
	}
	else
	{
		//Rechts
		if (Dir.y)
		{
			Dir.x = Dir.y;
			Dir.y = 0;
		}
		else
		{
			Dir.y = -Dir.x;
			Dir.x = 0;
		}
		RotationLeft += D3DX_PI / 2.0f;
		if (RotationLeft > 2 * D3DX_PI)
			RotationLeft -= 2 * D3DX_PI;
	}
	//Letze Wand bis zum Moped "ziehen"
	Wall.UpdateWall (&LastOWPos, &Pos);
	pWallDesc->UpdateWall (&LastWallPos, &Pos, WallDescIndex);
	//Neue Wand erstellen
	Wall.AddWall ();
	WallDescIndex = pWallDesc->AddWall (Ident);
	Wall.UpdateWall (&Pos, &Pos);
	pWallDesc->UpdateWall (&Pos, &Pos, WallDescIndex);
	//Variablen updaten
	LastWallPos = Pos;
	LastOWPos = Pos;
	ActWallPos.x = LastWallPos.x - Dir.x * 1.6f;
	ActWallPos.y = LastWallPos.y - Dir.y * 1.6f;
}

void TronPlayer::AI (float RelTimeDiff, TronShots *pShots)
{
	int i, NDir, WIndex;
	float AIRTime, WayLen, MopedSpeed = Options.MopedSpeed;
	D3DXVECTOR2 DirVec;
	TronPlayer **pField;

	if (!Alive || AIReact > 0)
		return;
	if (SpeedTime > 0)
		MopedSpeed *= 1.5f;
	else if (ShieldTime > 0)
		MopedSpeed *= 1.25f;
	//Zielsuchenden Schuss feuern oder Schild verwenden.
	if (Options.EnableAction && ItemAviable >= 2)
		FireItem (pShots);
	//Normalen Schuss feuern, falls vorhanden
	WayLen = pWallDesc->GetWP (&Pos, &Dir);
	if (Options.EnableAction && ItemAviable == 1 && WayLen > 3 * MopedSpeed)
		FireItem (pShots);
	//TronPlayer-Feld erstellen
	pField = new TronPlayer *[Options.OpponentCnt + 1];
	for (i = 0; i < Options.OpponentCnt + 1; i++)
		pField[i] = &pPlayer[i];
	//TRON AI III
	AIRTime = 0.25f - Options.OpponentSkill * 0.1f;
	if (AIChDir (pField, &NDir, RelTimeDiff))
	{
		//Es wird empfohlen zu lenken.
		if (NDir != -1)
		{
			//Die empfohlene Richtung ist bereits bekannt. Prfen, ob sie auch mglich bzw. sinnvoll ist.
			DirVec = NormalVector2dMake (&Vector2dMake (0, 0), &Dir);		//Rechts
			if (NDir == 0)
				D3DXVec2Scale (&DirVec, &DirVec, -1);						//Links
			if (pWallDesc->GetWP (&Pos, &DirVec) - MopedSpeed * (AIRTime + 0.05f) > 0)
			{
				ChangeDir (NDir);
				AIReact = AIRTime;
				AILastCh = 2;			//verhindern, dass dieser Zug durch zuflliges Lenken wieder verhindert wird
			}
		}
		else
		{
			//Die Richtung wird erst bestimmt. Mglicherweise auch geradeaus weiterfahren.
			for (i = 0; i <= Options.OpponentCnt; i++)
			{
				WIndex = pWallDesc->AddWall (-1);
				pWallDesc->UpdateWall (&pField[i]->LastWallPos, &pField[i]->Pos, WIndex);
			}
			WayLen = AIGetWayLen (Pos, Dir, Options.OpponentSkill + 3, &NDir);
			//Speedup verwenden, falls sinnvoll
			if (Options.EnableAction && ItemAviable == 0 && WayLen > 5 * MopedSpeed)
				FireItem (pShots);
			pWallDesc->DeletePlayer (-1);
			if (NDir != -1)
			{
				ChangeDir (NDir);
				AIReact = AIRTime;
			}
		}
	}
	else if (AILastCh == 0)
	{
		//Ab einem bestimmten Zeitabstand "selbststndig" lenken
		//Die Richtung wird erst bestimmt. Mglicherweise auch geradeaus weiterfahren.
		for (i = 0; i <= Options.OpponentCnt; i++)
		{
			WIndex = pWallDesc->AddWall (-1);
			pWallDesc->UpdateWall (&pField[i]->LastWallPos, &pField[i]->Pos, WIndex);
		}
		WayLen = AIGetWayLen (Pos, Dir, Options.OpponentSkill + 3, &NDir);
		//Speedup verwenden, falls sinnvoll
		if (Options.EnableAction && ItemAviable == 0 && WayLen > 5 * MopedSpeed)
			FireItem (pShots);
		pWallDesc->DeletePlayer (-1);
		if (NDir != -1)
		{
			ChangeDir (NDir);
			AIReact = AIRTime;
			AILastCh = 0.5f;
		}
	}
	delete []pField;
}


//TRON AI III
float TronPlayer::AIGetWayLen (D3DXVECTOR2 Pos, D3DXVECTOR2 LastDir, int d, int *pDir)
{
	int i, WIndex;
	float WayLen[3], AIRTime, MopedSpeed = Options.MopedSpeed;
	D3DXVECTOR2 DirVec[3], AddVec, NPos;

	if (d == 0)
		return (0);
	DirVec[0] = LastDir;													//Geradeaus
	DirVec[1] = NormalVector2dMake (&Vector2dMake (0, 0), &LastDir);		//Rechts
	D3DXVec2Scale (&DirVec[2], &DirVec[1], -1);								//Links
	AIRTime = 0.3f - Options.OpponentSkill * 0.1f;
	if (SpeedTime > 0)
		MopedSpeed *= 1.5f;
	else if (ShieldTime > 0)
		MopedSpeed *= 1.25f;
	for (i = 0; i < 3; i++)
	{
		//Die drei mglichen Richtungen auf die lngste Strecke abfragen
		WayLen[i] = pWallDesc->GetWP (&Pos, &DirVec[i]) - MopedSpeed * AIRTime;
		if (WayLen[i] > 0)
		{
			D3DXVec2Scale (&AddVec, &DirVec[i], WayLen[i]);
			D3DXVec2Add (&NPos, &Pos, &AddVec);
			//Temporre Wand erstellen
			WIndex = pWallDesc->AddWall (-1);
			pWallDesc->UpdateWall (&Pos, &NPos, WIndex);
			//Vom nchsten Punkt ausgehend berechnen
			WayLen[i] += AIGetWayLen (NPos, DirVec[i], d - 1, pDir);
			//Temporre Wand wieder lschen
			pWallDesc->DeleteWall (WIndex);
		}
		else
			WayLen[i] = 0;
	}
	//Den lngsten Weg herausfinden
	if (!WayLen[0] && !WayLen[1] && !WayLen[2])
	{
		//Es schaut schlecht aus. Geradeaus weiterfahren, es bleibt keine Mglichkeit mehr brig.
		*pDir = -1;
		return (0);
	}
	if (WayLen[0] > WayLen[1] && WayLen[0] > WayLen[2])
	{
		*pDir = -1;				//Geradeaus
		return (WayLen[0]);
	}
	else if (WayLen[1] > WayLen[0] && WayLen[1] > WayLen[2])
	{
		*pDir = 1;				//Rechts
		return (WayLen[1]);
	}
	else
	{
		*pDir = 0;				//Links
		return (WayLen[2]);
	}
}


bool TronPlayer::AIChDir (TronPlayer **pField, int *pDir, float RelTimeDiff)
{
	int i;
	float Result, dx, dy, MopedSpeed = Options.MopedSpeed;
	D3DXVECTOR2 TestPos;

	if (SpeedTime > 0)
		MopedSpeed *= 1.5f;
	else if (ShieldTime > 0)
		MopedSpeed *= 1.25f;
	//Flle, bei denen dringend gelenkt werden sollte
	*pDir = -1;
	//1. Bei drohendem Anstoss an ein Hindernis
	TestPos.x = Pos.x + Dir.x * (2 * MopedSpeed * RelTimeDiff + 5 - Options.OpponentSkill * 2.0f);
	TestPos.y = Pos.y + Dir.y * (2 * MopedSpeed * RelTimeDiff + 5 - Options.OpponentSkill * 2.0f);
	if (TestPos.x < 0 || TestPos.x > Options.LandSize || TestPos.y < 0 || TestPos.y > Options.LandSize)
		return (true);
	if (ShieldTime != 0)
		return (false);
	if (!pWallDesc->CanPlace (&PrevPos, &TestPos, WallDescIndex))
		return (true);
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (pField[i] != this && !CanMopedPlace (&TestPos, pField[i]))
			return (true);
	}
	//2. Beim Entgegenkommen eines anderen Mopeds Richtung ndern (direkte Kollission vermeiden)
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (pField[i] != this && pField[i]->Alive && pField[i]->Dir.x == -Dir.x && pField[i]->Dir.y == -Dir.y)
		{
			//Dieses Moped kommt entgegen, mglicherweise lenken
			if (Dir.x && fabs (pField[i]->Pos.x - Pos.x) < 20 && fabs (pField[i]->Pos.y - Pos.y) < 5)
				return (true);
			else if (Dir.y && fabs (pField[i]->Pos.x - Pos.x) < 5 && fabs (pField[i]->Pos.y - Pos.y) < 20)
				return (true);
		}
	}
	//3. Ist ein Gegner (knapp) dahinter, Haken machen (Weg abschneiden)
	if (Options.OpponentSkill > 0)
	{
		for (i = 0; i < Options.OpponentCnt + 1; i++)
		{
			if (pField[i] != this && pField[i]->Alive && AILastHook == 0)
			{
				if ((pField[i]->Dir.x == Dir.x && pField[i]->Dir.y == Dir.y) || (pField[i]->Dir.x == -Dir.x && pField[i]->Dir.y == -Dir.y))
				{
					//Dieses Moped fhrt entweder in die gleiche Richtung wie wir oder kommt uns entgegen.
					//0 = links, 1 = rechts
					AILastHook = 1;
					dx = (float) fabs (Pos.x - pField[i]->Pos.x);
					dy = (float) fabs (Pos.y - pField[i]->Pos.y);
					if (Dir.x && dx < 75 && dx > dy)
					{
						Result = Dir.x * (pField[i]->Pos.y - Pos.y);
						if (Result > 0)
							*pDir = 0;
						else
							*pDir = 1;
						return (true);
					}
					else if (Dir.y && dy < 75 && dy > dx)
					{
						Result = Dir.y * (pField[i]->Pos.x - Pos.x);
						if (Result > 0)
							*pDir = 1;
						else
							*pDir = 0;
						return (true);
					}
				}
			}
		}
	}
	return (false);
}

void TronPlayer::DoEvents (float RelTimeDiff, float MopedSpeed)
{
	D3DXVECTOR2 AddDir, WheelRot, CalcVec;
	float sinAngle, cosAngle, WLen;

	if (!Alive)
	{
		Wall.TransIncr (RelTimeDiff);
		return;
	}
	//*************Items********************
	if (Options.EnableAction)
	{
		SpeedTime -= RelTimeDiff;
		if (SpeedTime > 0)
			MopedSpeed *= 1.5f;
		else
			SpeedTime = 0;
		ShieldTime -= RelTimeDiff;
		if (ShieldTime > 0)
			MopedSpeed *= 1.25f;
		else
			ShieldTime = 0;
		if (ItemAviable == -1)
		{
			ItemTime -= RelTimeDiff;
			if (ItemTime < 0)
			{
				ItemTime = GetRandomVal (10, 15);
				ItemAviable = (int) GetRandomVal (0, 100) % 100;
				if (ItemAviable < 25)			//0 bis 24
					ItemAviable = 0;			//Speedup: 25 %
				else if (ItemAviable < 50)		//25 bis 49
					ItemAviable = 1;			//Normaler Schuss: 25 %
				else if (ItemAviable < 80)		//50 bis 79
					ItemAviable = 2;			//Zielsucher: 30 %
				else if (ItemAviable < 82)		//80 bis 81
					ItemAviable = 3;			//Superschuss: 2 %
				else							//82 bis 99
					ItemAviable = 4;			//Schild: 18 %
			}
		}
	}
	//*************Moped bewegen**************
	AddDir.x = Dir.x * MopedSpeed * RelTimeDiff;
	AddDir.y = Dir.y * MopedSpeed * RelTimeDiff;
	PrevPos = Pos;
	D3DXVec2Add (&Pos, &Pos, &AddDir);
	D3DXVec2Add (&ActWallPos, &ActWallPos, &AddDir);
	Moped.SetPosition (&Vector3dMake (Pos.x, 0, Pos.y));
	//*******************AI***************
	if (MopedSpeed > 3)
	{
		//Spiel luft bereits
		AIReact -= RelTimeDiff;
		if (AIReact < 0)
			AIReact = 0;
		AILastCh -= RelTimeDiff;
		if (AILastCh < 0)
			AILastCh = 0;
		AILastHook -= RelTimeDiff;
		if (AILastHook < 0)
			AILastHook = 0;
	}
	//*************Moped rotieren, falls ntig**************
	if (RotationLeft > 0)
	{
		//Rechts-Rotation
		Moped.Rotate (&Vector3dMake (0, RelTimeDiff * 10, 0), NULL, &Vector3dMake (Pos.x, 0, Pos.y));
		Wheel.Rotate (&Vector3dMake (0, RelTimeDiff * 10, 0), NULL, Wheel.GetPosition ());
		RotationLeft -= RelTimeDiff * 10;
		YAngle += RelTimeDiff * 10;
		if (RotationLeft < 0)
		{
			Moped.Rotate (&Vector3dMake (0, RotationLeft, 0), NULL, &Vector3dMake (Pos.x, 0, Pos.y));
			Wheel.Rotate (&Vector3dMake (0, RotationLeft, 0), NULL, Wheel.GetPosition ());
			YAngle += RotationLeft;
			RotationLeft = 0;
		}
	}
	else if (RotationLeft < 0)
	{
		//Links-Rotation
		Moped.Rotate (&Vector3dMake (0, -RelTimeDiff * 10, 0), NULL, &Vector3dMake (Pos.x, 0, Pos.y));
		Wheel.Rotate (&Vector3dMake (0, -RelTimeDiff * 10, 0), NULL, Wheel.GetPosition ());
		RotationLeft += RelTimeDiff * 10;
		YAngle -= RelTimeDiff * 10;
		if (RotationLeft > 0)
		{
			Moped.Rotate (&Vector3dMake (0, RotationLeft, 0), NULL, &Vector3dMake (Pos.x, 0, Pos.y));
			Wheel.Rotate (&Vector3dMake (0, RotationLeft, 0), NULL, Wheel.GetPosition ());
			YAngle += RotationLeft;
			RotationLeft = 0;
		}
	}
	else
	{
		//*************Rder rotieren**************
		ZeroMemory (&WheelRot, sizeof (D3DXVECTOR2));
		if (Dir.x == 1)
			WheelRot.y = -MopedSpeed * RelTimeDiff * 2;
		else if (Dir.x == -1)
			WheelRot.y = MopedSpeed * RelTimeDiff * 2;
		else if (Dir.y == 1)
			WheelRot.x = MopedSpeed * RelTimeDiff * 2;
		else
			WheelRot.x = -MopedSpeed * RelTimeDiff * 2;
		Wheel.Rotate (&Vector3dMake (WheelRot.x, 0, WheelRot.y), NULL, Wheel.GetPosition ());
	}
	//*************Reifen-Positionen ausrechnen**************
	if (RotationLeft)
	{
		sinAngle = (float) sin (YAngle);
		cosAngle = (float) cos (YAngle);
		//Hinterreifen
		BackRPos = Vector3dMake (Pos.x + 1.35f * sinAngle, 0.45f, Pos.y + 1.35f * cosAngle);
		//Vorderreifen
		FrontRPos = Vector3dMake (Pos.x - 1.21f * sinAngle, 0.5, Pos.y - 1.21f * cosAngle);
	}
	else
	{
		//Hinterreifen
		BackRPos = Vector3dMake (Pos.x - Dir.x * 1.35f, 0.45f, Pos.y - Dir.y * 1.35f);
		//Vorderreifen
		FrontRPos = Vector3dMake (Pos.x + Dir.x * 1.21f, 0.5, Pos.y + Dir.y * 1.21f);
	}
	//Wnde aktualisieren
	if (ActWallPos.x * Dir.x > LastWallPos.x * Dir.x || ActWallPos.y * Dir.y > LastWallPos.y * Dir.y)
	{
		D3DXVec2Subtract (&CalcVec, &LastOWPos, &ActWallPos);
		WLen = D3DXVec2Length (&CalcVec);
		if (WLen > 15)
		{
			//Wand optisch teilen (bei einer Lnge von 15)
			D3DXVec2Scale (&CalcVec, &Dir, 15);
			D3DXVec2Add (&CalcVec, &LastOWPos, &CalcVec);
			Wall.UpdateWall (&LastOWPos, &CalcVec);
			Wall.AddWall ();
			LastOWPos = CalcVec;
		}
		Wall.UpdateWall (&LastOWPos, &ActWallPos);
		pWallDesc->UpdateWall (&LastWallPos, &ActWallPos, WallDescIndex);
	}
	else
	{
		Wall.UpdateWall (&LastOWPos, &LastOWPos);
		pWallDesc->UpdateWall (&LastWallPos, &LastWallPos, WallDescIndex);
	}
}

void TronPlayer::HandleShield ()
{
	if (ShieldTime == 0 || !Alive)
		return;

	D3DXVECTOR2 CrashPos;
	int WIndex;
	bool bx;

	//Dieses Moped verwendet ein Schild und kann daher andere Moped-Wnde zerstren.
	D3DXVec2Add (&CrashPos, &Pos, &Dir);
	if (!pWallDesc->CanPlace (&PrevPos, &CrashPos, -1, &bx, &WIndex))
	{
		pWallDesc->DestroyWall (WIndex, bx, &Pos);
		Explosions.CreateExplosion (&Pos, true, MPSrv);
		if (Options.UseSounds)
		{
			ShotCrashSound.Set3dPosition (&Vector3dMake (Pos.x, 0, Pos.y));
			ShotCrashSound.SetPlayPosition (0);
			EngSound.Update3dSounds ();
			EngSound.Play3dSound (&ShotCrashSound, false);
		}
		if (MPSrv)
		{
			SData[0] = CLIENT_PLAYSOUND;
			SData[1] = 1;
			memcpy (&SData[2], &Pos, sizeof (D3DXVECTOR2));
			GameServer.SendData (0, SData, 2 + sizeof (D3DXVECTOR2), 0);
		}
	}
}

bool TronPlayer::CanPlace ()
{
	int i;
	D3DXVECTOR2 CrashPos;

	if (!Alive)
		return (true);	
	D3DXVec2Add (&CrashPos, &Pos, &Dir);
	//Auf Wnde von anderen Mopeds berprfen
	if (ShieldTime == 0 && !pWallDesc->CanPlace (&PrevPos, &CrashPos, WallDescIndex))
		return (false);
	//Auf die Wnde der Landschaft berprfen
	if (CrashPos.x < 0 || CrashPos.x > Options.LandSize || CrashPos.y < 0 || CrashPos.y > Options.LandSize)
		return (false);
	//Auf andere Mopeds berprfen (direkte Kollission)
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (&pPlayer[i] != this && !CanMopedPlace (&CrashPos, &pPlayer[i]))
			return (false);
	}
	return (true);
}

bool TronPlayer::CanMopedPlace (D3DXVECTOR2 *CrashPos, TronPlayer *pOpp)
{
	if (!Alive || !pOpp->Alive)
		return (true);

	D3DXVECTOR2 MopedPos[2];

	MopedPos[0] = Vector2dMake (pOpp->Pos.x - pOpp->Dir.x * 1.85f, pOpp->Pos.y - pOpp->Dir.y * 1.85f);
	MopedPos[1] = Vector2dMake (pOpp->Pos.x + pOpp->Dir.x * 1.71f, pOpp->Pos.y + pOpp->Dir.y * 1.71f);
	if (InLine (PrevPos.x, CrashPos->x, MopedPos[0].x, MopedPos[1].x))
	{
		//x-Bereich berlappend
		if (InLine (PrevPos.y, CrashPos->y, MopedPos[0].y, MopedPos[1].y))
		{
			//x- und y-Bereich berlappend, Crash
			//Der andere explodiert, ausser er hat ein Schild und wir keines.
			if (pOpp->ShieldTime == 0 || ShieldTime > 0)
				pOpp->WillCrash = true;
			//Wir explodieren, ausser wir haben ein Schild und er keines.
			if (pOpp->ShieldTime > 0 || ShieldTime == 0)
				return (false);
		}
	}
	return (true);
}

void TronPlayer::WallCrash ()
{
	//Letzte Mauer bis zum Moped ziehen
	Wall.UpdateWall (&LastOWPos, &Pos);
	pWallDesc->UpdateWall (&LastWallPos, &Pos, WallDescIndex);
	//Neue Mauer-Beschreibung generieren
	WallDescIndex = pWallDesc->AddWall (Ident);
	pWallDesc->UpdateWall (&Pos, &Pos, WallDescIndex);
	LastWallPos = Pos;
	LastOWPos = Pos;
	ActWallPos.x = LastWallPos.x - Dir.x * 1.6f;
	ActWallPos.y = LastWallPos.y - Dir.y * 1.6f;	
}

void TronPlayer::FireItem (TronShots *pShots)
{
	if (ItemAviable == 0)
		SpeedTime = 5;
	else if (ItemAviable == 1)
		pShots->FireShot (&Pos, &Dir, 0, this, MPSrv);
	else if (ItemAviable == 2)
		pShots->FireShot (&Pos, &Dir, 1, this, MPSrv);
	else if (ItemAviable == 3)
		pShots->FireShot (&Pos, &Dir, 2, this, MPSrv);
	else if (ItemAviable == 4)
		ShieldTime = 5;
	ItemAviable = -1;
}

void TronPlayer::Render (bool Shadow, Revo3dObject *pShield)
{
	if (!Alive)
		return;

	int i;
	bool RenderShield;

	//Moped rendern
	Eng3d.Render3dObject (&Moped);
	if (!Shadow)
	{
		//Rder rendern
		Wheel.SetPosition (&BackRPos);
		Eng3d.Render3dObject (&Wheel);
		Wheel.SetPosition (&FrontRPos);
		Eng3d.Render3dObject (&Wheel);
		//Schild rendern, falls vorhanden
		if (ShieldTime > 0)
		{
			RenderShield = true;
			if (ShieldTime < 1.5)
			{
				//Schild blinken lassen
				for (i = 1; i <= 13; i += 2)
				{
					if (ShieldTime > i * 0.1f && ShieldTime < (i + 1) * 0.1f)
					{
						RenderShield = false;
						break;
					}
				}
			}
			if (RenderShield)
			{
				pShield->SetPosition (&Vector3dMake (Pos.x, 0, Pos.y));
				Eng3d.Render3dObject (pShield);
			}
		}
	}
	else
	{
		//Schatten-Rder rendern
		SWheel.SetRotationMatrix (Wheel.GetRotationMatrix ());
		SWheel.SetPosition (&BackRPos);
		Eng3d.Render3dObject (&SWheel);
		SWheel.SetPosition (&FrontRPos);
		Eng3d.Render3dObject (&SWheel);
	}
}

void TronPlayer::RenderWalls (bool Mirrored)
{
	if (!Alive)
	{
		Wall.RenderTrans ();
		return;
	}
	if (!Options.UseTransWalls || Mirrored)
		Wall.Render ();
	else
		Wall.RenderTrans ();
}

void TronPlayer::WriteData (BYTE **pMem)
{
	memcpy (*pMem, &Pos, sizeof (D3DXVECTOR2));
	*pMem += sizeof (D3DXVECTOR2);
	memcpy (*pMem, &Dir, sizeof (D3DXVECTOR2));
	*pMem += sizeof (D3DXVECTOR2);
	memcpy (*pMem, &BackRPos, sizeof (D3DXVECTOR3));
	*pMem += sizeof (D3DXVECTOR3);
	memcpy (*pMem, &FrontRPos, sizeof (D3DXVECTOR3));
	*pMem += sizeof (D3DXVECTOR3);
	memcpy (*pMem, &Alive, sizeof (bool));
	*pMem += sizeof (bool);
	memcpy (*pMem, &Frags, sizeof (int));
	*pMem += sizeof (int);
	memcpy (*pMem, &ItemAviable, sizeof (int));
	*pMem += sizeof (int);
	memcpy (*pMem, &Rank, sizeof (int));
	*pMem += sizeof (int);
	memcpy (*pMem, &ShieldTime, sizeof (float));
	*pMem += sizeof (float);
	memcpy (*pMem, Moped.GetRotationMatrix (), sizeof (D3DXMATRIX));
	*pMem += sizeof (D3DXMATRIX);
	memcpy (*pMem, Wheel.GetRotationMatrix (), sizeof (D3DXMATRIX));
	*pMem += sizeof (D3DXMATRIX);
	memcpy (*pMem, Wall.GetTransMat (), sizeof (D3DMATERIAL8));
	*pMem += sizeof (D3DMATERIAL8);
	memcpy (*pMem, Wall.GetSolidMat (), sizeof (D3DMATERIAL8));
	*pMem += sizeof (D3DMATERIAL8);
	memcpy (*pMem, &Wall.Trans, sizeof (float));
	*pMem += sizeof (float);
}

void TronPlayer::ReadData (BYTE **pMem)
{
	D3DXMATRIX RdMatrix;

	memcpy (&Pos, *pMem, sizeof (D3DXVECTOR2));
	*pMem += sizeof (D3DXVECTOR2);
	Moped.SetPosition (&Vector3dMake (Pos.x, 0, Pos.y));
	memcpy (&Dir, *pMem, sizeof (D3DXVECTOR2));
	*pMem += sizeof (D3DXVECTOR2);
	memcpy (&BackRPos, *pMem, sizeof (D3DXVECTOR3));
	*pMem += sizeof (D3DXVECTOR3);
	memcpy (&FrontRPos, *pMem, sizeof (D3DXVECTOR3));
	*pMem += sizeof (D3DXVECTOR3);
	memcpy (&Alive, *pMem, sizeof (bool));
	*pMem += sizeof (bool);
	memcpy (&Frags, *pMem, sizeof (int));
	*pMem += sizeof (int);
	memcpy (&ItemAviable, *pMem, sizeof (int));
	*pMem += sizeof (int);
	memcpy (&Rank, *pMem, sizeof (int));
	*pMem += sizeof (int);
	memcpy (&ShieldTime, *pMem, sizeof (float));
	*pMem += sizeof (float);
	memcpy (&RdMatrix, *pMem, sizeof (D3DXMATRIX));
	*pMem += sizeof (D3DXMATRIX);
	Moped.SetRotationMatrix (&RdMatrix);
	memcpy (&RdMatrix, *pMem, sizeof (D3DXMATRIX));
	*pMem += sizeof (D3DXMATRIX);
	Wheel.SetRotationMatrix (&RdMatrix);
	memcpy (Wall.GetTransMat (), *pMem, sizeof (D3DMATERIAL8));
	*pMem += sizeof (D3DMATERIAL8);
	memcpy (Wall.GetSolidMat (), *pMem, sizeof (D3DMATERIAL8));
	*pMem += sizeof (D3DMATERIAL8);
	memcpy (&Wall.Trans, *pMem, sizeof (float));
	*pMem += sizeof (float);
}




//*******************TRON WALLS*********************
void TronWalls::Initialize (int NrStaticWalls, int r, int g, int b, Revo2dTexture *pTex)
{
	WallCnt = NrStaticWalls;
	WallUsed = 0;
	Walls.Initialize (&Eng3d, true);
	Walls.SetVisibility (BOTHSIDES);
	Walls.SetPolyFormat (TRIANGLELIST);
	Walls.SetTexture (pTex);
	WallsTop.Initialize (&Eng3d, true);
	WallsTop.SetVisibility (BOTHSIDES);
	WallsTop.SetPolyFormat (TRIANGLELIST);
	SolidMat = MaterialMake (r, g, b, 0, 0, 0, 0, 0);
	TransMat = MaterialMake (r, g, b, 0, 0, 0, 0, 100);
	this->r = r;
	this->g = g;
	this->b = b;
	if (Options.UseTransWalls)
		Trans = 100;
	else
		Trans = 0;
}

void TronWalls::Delete ()
{
	WallUsed = 0;
	TransMat = MaterialMake (r, g, b, 0, 0, 0, 0, 100);
	if (Options.UseTransWalls)
		Trans = 100;
	else
		Trans = 0;
}

void TronWalls::Rebuild (void *pWDesc, int Player, bool Multiplayer)
{
	int i, DescUsed;
	TronWall *pWalls;
	TronWallDesc *pWallDesc;
	D3DXVECTOR2 WallPos, WallVec, NWallVec;

	pWallDesc = (TronWallDesc *) pWDesc;
	DescUsed = pWallDesc->GetWallUsed ();
	pWalls = pWallDesc->GetWallDesc ();
	WallUsed = 0;
	for (i = 0; i < DescUsed; i++)
	{
		if (pWalls[i].Used && pWalls[i].From == Player)
		{
			//Diese Wand hinzufgen
			WallPos = pWalls[i].Pos1;
			D3DXVec2Subtract (&WallVec, &pWalls[i].Pos2, &WallPos);
			D3DXVec2Normalize (&NWallVec, &WallVec);
			while (D3DXVec2Length (&WallVec) > 15)
			{
				AddWall ();
				UpdateWall (&WallPos, &Vector2dMake (WallPos.x + NWallVec.x * 15, WallPos.y + NWallVec.y * 15));
				WallPos = Vector2dMake (WallPos.x + NWallVec.x * 15, WallPos.y + NWallVec.y * 15);
				D3DXVec2Subtract (&WallVec, &pWalls[i].Pos2, &WallPos);
			}
			//Jetzt noch die letzte Wand von dieser Gerade hinzufgen
			AddWall ();
			UpdateWall (&WallPos, &pWalls[i].Pos2);
		}
	}
	if (!Multiplayer)
	{
		//Die aktuelle Wand neu erstellen, die immer upgedated wird
		AddWall ();
	}
}

void TronWalls::AddWall ()
{
	if (WallUsed == WallCnt)
	{
		//Speicher fr neue Wnde erstellen
		WallCnt *= 2;
		Walls.ChangeBufferSize (WallCnt * 6, 0);
		WallsTop.ChangeBufferSize (WallCnt * 6, 0);
	}
	WallUsed++;
}

void TronWalls::UpdateWall (D3DXVECTOR2 *Pos1, D3DXVECTOR2 *Pos2)
{
	if (WallUsed == 0)
		return;

	Vertex ActVertex[6], TopVertex[6];
	D3DXVECTOR2 CalcVec, NVec;
	float TexRep;
	int WallNr = WallUsed - 1;

	D3DXVec2Subtract (&CalcVec, Pos1, Pos2);
	TexRep = D3DXVec2Length (&CalcVec) / 15.0f;
	//Vertex-Daten erstellen
	ActVertex[0] = VertexMake (Pos1->x, 0, Pos1->y, 0, 1, 0, 0, 1);
	ActVertex[1] = VertexMake (Pos2->x, 2, Pos2->y, 0, 1, 0, TexRep, 0);
	ActVertex[2] = VertexMake (Pos2->x, 0, Pos2->y, 0, 1, 0, TexRep, 1);
	ActVertex[3] = ActVertex[0];
	ActVertex[4] = VertexMake (Pos1->x, 2, Pos1->y, 0, 1, 0, 0, 0);
	ActVertex[5] = ActVertex[1];
	//Vertex-Top-Daten erstellen
	NVec = NormalVector2dMake (Pos1, Pos2);
	D3DXVec2Scale (&NVec, &NVec, 0.1f);
	TopVertex[0] = VertexMake (Pos1->x - NVec.x, 2, Pos1->y - NVec.y, 0, 1, 0, 0, 0);
	TopVertex[1] = VertexMake (Pos1->x + NVec.x, 2, Pos1->y + NVec.y, 0, 1, 0, 0, 0);
	TopVertex[2] = VertexMake (Pos2->x + NVec.x, 2, Pos2->y + NVec.y, 0, 1, 0, 0, 0);
	TopVertex[3] = TopVertex[0];
	TopVertex[4] = VertexMake (Pos2->x - NVec.x, 2, Pos2->y - NVec.y, 0, 1, 0, 0, 0);
	TopVertex[5] = TopVertex[2];
	//Vertex-Daten an die Wand zuweisen
	Walls.SetVertexData (ActVertex, WallNr * 6, 6);
	WallsTop.SetVertexData (TopVertex, WallNr * 6, 6);
}

void TronWalls::TransIncr (float RelTimeDiff)
{
	if (Trans == 255)
		return;
	Trans += RelTimeDiff * 200;
	if (Trans > 255)
		Trans = 255;
	TransMat = MaterialMake (r, g, b, 0, 0, 0, 0, (int) Trans);
}

void TronWalls::Render ()
{
	Walls.SetMaterial (&SolidMat);
	Eng3d.Render3dPolygons (&Walls, 0, WallUsed * 2);
	if (Options.Cam != 6)
	{
		WallsTop.SetMaterial (&SolidMat);
		Eng3d.Render3dPolygons (&WallsTop, 0, WallUsed * 2);
	}
}

void TronWalls::RenderTrans ()
{
	if (Trans == 255)
		return;
	Walls.SetMaterial (&TransMat);
	Eng3d.Render3dPolygons (&Walls, 0, WallUsed * 2);
	if (Options.Cam != 6)
	{
		WallsTop.SetMaterial (&TransMat);
		Eng3d.Render3dPolygons (&WallsTop, 0, WallUsed * 2);
	}
}




//*******************TRON WALL-DESC*********************
void TronWallDesc::Initialize (int NrStaticWalls)
{
	WallCnt = NrStaticWalls;
	WallUsed = 0;
	pWallDesc = (TronWall *) malloc (WallCnt * sizeof (TronWall));
}

TronWallDesc::TronWallDesc ()
{
	pWallDesc = NULL;
}

TronWallDesc::~TronWallDesc ()
{
	if (TronWallDesc)
		free (pWallDesc);
}

int TronWallDesc::AddWall (int From)
{
	int i, Index = -1;

	for (i = 0; i < WallUsed; i++)
	{
		if (!pWallDesc[i].Used)
		{
			Index = i;
			break;
		}
	}
	if (Index == -1)
	{
		if (WallUsed == WallCnt)
		{
			//Speicher fr neue Wnde erstellen
			WallCnt *= 2;
			pWallDesc = (TronWall *) realloc (pWallDesc, WallCnt * sizeof (TronWall));
		}
		Index = WallUsed;
		WallUsed++;
	}
	pWallDesc[Index].Used = true;
	pWallDesc[Index].From = From;
	return (Index);
}

void TronWallDesc::UpdateWall (D3DXVECTOR2 *Pos1, D3DXVECTOR2 *Pos2, int WallIndex)
{
	//Wand-"Beschreibung" erneuern
	pWallDesc[WallIndex].Pos1 = *Pos1;
	pWallDesc[WallIndex].Pos2 = *Pos2;
}

void TronWallDesc::DeleteWall (int WallIndex)
{
	pWallDesc[WallIndex].Used = false;
}

bool TronWallDesc::CanPlace (D3DXVECTOR2 *Pos1, D3DXVECTOR2 *Pos2, int OwnWallIndex, bool *xWall, int *pWIndex)
{
	int i;
	bool xW, xD, Hit;
	D3DXVECTOR2 NVec;

	if (Pos1->y == Pos2->y)
		xD = true;
	else
		xD = false;
	for (i = 0; i < WallUsed; i++)
	{
		if (pWallDesc[i].Used && i != OwnWallIndex)
		{
			if (pWallDesc[i].Pos1.y == pWallDesc[i].Pos2.y)
				xW = true;
			else
				xW = false;
			//Prfen, ob die Mauer selbst geschnitten wird
			if (xD != xW || xWall)
			{
				//Diese Wand ist quer zur Prf-Richtung, daher beachten.
				Hit = false;
				if (xW)
				{
					//x-Wand, zuerst y prfen.
					if (InArea (Pos1->y, Pos2->y, pWallDesc[i].Pos1.y))
					{
						if (InArea (pWallDesc[i].Pos1.x, pWallDesc[i].Pos2.x, Pos1->x))
							Hit = true;
					}
				}
				else
				{
					//y-Wand, zuerst x prfen.
					if (InArea (Pos1->x, Pos2->x, pWallDesc[i].Pos1.x))
					{
						if (InArea (pWallDesc[i].Pos1.y, pWallDesc[i].Pos2.y, Pos1->y))
							Hit = true;
					}
				}
				if (Hit)
				{
					//x und y berlappend, Crash
					if (xWall)
						*xWall = xW;
					if (pWIndex)
						*pWIndex = i;
					return (false);
				}
			}
			//Prfen, ob das Moped seitlich zu nahe an die Wand heranfhrt
			if (xD == xW && !xWall)
			{
				//Diese Wand ist parallel zur Prf-Richtung, daher beachten.
				NVec = NormalVector2dMake (&pWallDesc[i].Pos1, &pWallDesc[i].Pos2);
				D3DXVec2Scale (&NVec, &NVec, 0.5);
				if (xW)
				{
					//x-Wand, zuerst y prfen.
					if (InArea (pWallDesc[i].Pos1.y - NVec.y, pWallDesc[i].Pos1.y + NVec.y, Pos1->y))
					{
						if (InArea (Pos1->x, Pos2->x, pWallDesc[i].Pos1.x))
							return (false);
						if (InArea (Pos1->x, Pos2->x, pWallDesc[i].Pos2.x))
							return (false);
					}
				}
				else
				{
					//y-Wand, zuerst x prfen.
					if (InArea (pWallDesc[i].Pos1.x - NVec.x, pWallDesc[i].Pos1.x + NVec.x, Pos1->x))
					{
						if (InArea (Pos1->y, Pos2->y, pWallDesc[i].Pos1.y))
							return (false);
						if (InArea (Pos1->y, Pos2->y, pWallDesc[i].Pos2.y))
							return (false);
					}
				}
			}
		}
	}
	return (true);
}

void TronWallDesc::DeletePlayer (int Player)
{
	int i;

	for (i = 0; i < WallUsed; i++)
	{
		if (pWallDesc[i].From == Player)
			pWallDesc[i].Used = false;
	}
}

void TronWallDesc::DestroyWall (int WallIndex, bool xWall, D3DXVECTOR2 *pPos)
{
	int i, WIndex;
	D3DXVECTOR2 WallPos2, WallDir;
	float WallCut[2], WallCoord;
	bool CutValid[2];
	TronPlayer *pCP;

	//Die aktuellen Wnde von den Spielern neu generieren (sonst wird beim nchsten Update-Aufruf die Wand wieder hergestellt)
	pCP = &pPlayer[pWallDesc[WallIndex].From];
	pCP->WallCrash ();
	//Die Wand-"Richtung" ausrechnen
	D3DXVec2Subtract (&WallDir, &pWallDesc[WallIndex].Pos2, &pWallDesc[WallIndex].Pos1);
	D3DXVec2Normalize (&WallDir, &WallDir);
	//Wandbeschreibung neu erstellen
	if (xWall)
	{
		WallCut[0] = pPos->x - 5 * WallDir.x;
		WallCut[1] = pPos->x + 5 * WallDir.x;
		WallCoord = pWallDesc[WallIndex].Pos1.y;
		for (i = 0; i < 2; i++)
		{
			if (InArea (pWallDesc[WallIndex].Pos1.x, pWallDesc[WallIndex].Pos2.x, WallCut[i]))
				CutValid[i] = true;
			else
				CutValid[i] = false;
		}
		if (CutValid[0] && CutValid[1])
		{
			//Allgemeiner Fall: Wand wird in der Mitte geteilt und es entstehen 2 neue Wnde
			WallPos2 = pWallDesc[WallIndex].Pos2;
			UpdateWall (&pWallDesc[WallIndex].Pos1, &Vector2dMake (WallCut[0], WallCoord), WallIndex);
			WIndex = AddWall (pWallDesc[WallIndex].From);
			UpdateWall (&Vector2dMake (WallCut[1], WallCoord), &WallPos2, WIndex);
		}
		else if (!CutValid[0] && CutValid[1])
			UpdateWall (&Vector2dMake (WallCut[1], WallCoord), &pWallDesc[WallIndex].Pos2, WallIndex);
		else if (CutValid[0] && !CutValid[1])
			UpdateWall (&pWallDesc[WallIndex].Pos1, &Vector2dMake (WallCut[0], WallCoord), WallIndex);
		else
			DeleteWall (WallIndex);
	}
	else
	{
		WallCut[0] = pPos->y - 5 * WallDir.y;
		WallCut[1] = pPos->y + 5 * WallDir.y;
		WallCoord = pWallDesc[WallIndex].Pos1.x;
		for (i = 0; i < 2; i++)
		{
			if (InArea (pWallDesc[WallIndex].Pos1.y, pWallDesc[WallIndex].Pos2.y, WallCut[i]))
				CutValid[i] = true;
			else
				CutValid[i] = false;
		}
		if (CutValid[0] && CutValid[1])
		{
			//Allgemeiner Fall: Wand wird in der Mitte geteilt und es entstehen 2 neue Wnde
			WallPos2 = pWallDesc[WallIndex].Pos2;
			UpdateWall (&pWallDesc[WallIndex].Pos1, &Vector2dMake (WallCoord, WallCut[0]), WallIndex);
			WIndex = AddWall (pWallDesc[WallIndex].From);
			UpdateWall (&Vector2dMake (WallCoord, WallCut[1]), &WallPos2, WIndex);
		}
		else if (!CutValid[0] && CutValid[1])
			UpdateWall (&Vector2dMake (WallCoord, WallCut[1]), &pWallDesc[WallIndex].Pos2, WallIndex);
		else if (CutValid[0] && !CutValid[1])
			UpdateWall (&pWallDesc[WallIndex].Pos1, &Vector2dMake (WallCoord, WallCut[0]), WallIndex);
		else
			DeleteWall (WallIndex);		
	}
	//Die Wnde optisch neu generieren (d.h. das Loch erstellen)
	pCP->GetWalls ()->Rebuild (this, pWallDesc[WallIndex].From, false);
}

float TronWallDesc::GetWP (D3DXVECTOR2 *pPos, D3DXVECTOR2 *pDir)
{
	int i;
	float ActDist, MinDist;
	D3DXVECTOR2 LPos;
	bool First = true;

	//Ermittelt die krzeste Entfernung von diesem Punkt in diese Richtung zu einem Hindernis auf der Landschaft
	LPos = *pPos;
	if (LPos.x < 0)
		LPos.x = 0;
	if (LPos.x > Options.LandSize)
		LPos.x = Options.LandSize;
	if (LPos.y < 0)
		LPos.y = 0;
	if (LPos.y > Options.LandSize)
		LPos.y = Options.LandSize;	
	//Wnde von den Mopeds berprfen
	for (i = 0; i < WallUsed; i++)
	{
		if (pWallDesc[i].Used)
		{
			if (pDir->y)
			{
				if ((pWallDesc[i].Pos1.y - LPos.y) * pDir->y > 0 && InAreaDelta (pWallDesc[i].Pos1.x, pWallDesc[i].Pos2.x, LPos.x, 2.5f - Options.OpponentSkill))
				{
					ActDist = (float) fabs (pWallDesc[i].Pos1.y - LPos.y);
					if (First || ActDist < MinDist)
					{
						First = false;
						MinDist = ActDist;
					}
				}
			}
			else if (pDir->x)
			{
				if ((pWallDesc[i].Pos1.x - LPos.x) * pDir->x > 0 && InAreaDelta (pWallDesc[i].Pos1.y, pWallDesc[i].Pos2.y, LPos.y, 2.5f - Options.OpponentSkill))
				{
					ActDist = (float) fabs (pWallDesc[i].Pos1.x - LPos.x);
					if (First || ActDist < MinDist)
					{
						First = false;
						MinDist = ActDist;
					}
				}
			}
		}
	}
	//Wnde von der Landschaft berprfen
	if (pDir->x == 1 && (First || MinDist > Options.LandSize - LPos.x))
	{
		MinDist = Options.LandSize - LPos.x;
		First = false;
	}
	if (pDir->x == -1 && (First || MinDist > LPos.x))
	{
		MinDist = LPos.x;
		First = false;
	}
	if (pDir->y == 1 && (First || MinDist > Options.LandSize - LPos.y))
	{
		MinDist = Options.LandSize - LPos.y;
		First = false;
	}
	if (pDir->y == -1 && (First || MinDist > LPos.y))
		MinDist = LPos.y;
	return (MinDist);
}

void TronWallDesc::WriteData (BYTE **pMem)
{
	int i, UseCnt = 0;

	for (i = 0; i < WallUsed; i++)
	{
		if (pWallDesc[i].Used)
			UseCnt++;
	}
	memcpy (*pMem, &UseCnt, sizeof (int));
	*pMem += sizeof (int);
	for (i = 0; i < WallUsed; i++)
	{
		if (pWallDesc[i].Used)
		{
			memcpy (*pMem, &pWallDesc[i], sizeof (TronWall));
			*pMem += sizeof (TronWall);
		}
	}
}

void TronWallDesc::ReadData (BYTE **pMem)
{
	int i;
	int RdInt;

	memcpy (&RdInt, *pMem, sizeof (int));
	*pMem += sizeof (int);
	if (RdInt > WallCnt)
	{
		//Speicher erweitern
		pWallDesc = (TronWall *) realloc (pWallDesc, RdInt * sizeof (TronWall));
		WallCnt = RdInt;
	}
	WallUsed = RdInt;
	for (i = 0; i < WallCnt; i++)
	{
		//Gesamten Speicher durchlaufen, aber nur die verwendeten Daten kopieren.
		if (i < RdInt)
		{
			memcpy (&pWallDesc[i], *pMem, sizeof (TronWall));
			*pMem += sizeof (TronWall);
		}
		else
			pWallDesc[i].Used = false;
	}
	//Die Wnde optisch neu generieren.
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (pPlayer[i].IsAlive ())
			pPlayer[i].GetWalls ()->Rebuild (this, i, true);
	}
}


//*******************TRON EXPLOSION*********************
void TronExplosion::Initialize (int NrStaticExplosion, Revo2dTexture *pTex)
{
	int i;

	ExplCnt = NrStaticExplosion;
	pExplDesc = (TronExplDesc *) malloc (ExplCnt * sizeof (TronExplDesc));
	for (i = 0; i < ExplCnt; i++)
		pExplDesc[i].Used = false;
	Explosion.Initialize (&Eng3d, true);
	Explosion.SetPolyFormat (TRIANGLELIST);
	Explosion.SetVisibility (BOTHSIDES);
	Explosion.SetTexture (pTex);
	Explosion.ChangeBufferSize (180, 0);
}

TronExplosion::TronExplosion ()
{
	pExplDesc = NULL;
}

TronExplosion::~TronExplosion ()
{
	if (pExplDesc)
		free (pExplDesc);
}

void TronExplosion::CreateExplosion (D3DXVECTOR2 *Pos, bool Small, bool MPServer)
{
	int i, j;
	int ActExpl;

	if (MPServer)
	{
		//Dem Client die Explosions-Information schicken
		SData[0] = CLIENT_EXPLOSION;
		SData[1] = Small;
		memcpy (&SData[2], Pos, sizeof (D3DXVECTOR2));
		GameServer.SendData (0, SData, 2 + sizeof (D3DXVECTOR2), 0);
	}
	for (i = 0; i < ExplCnt; i++)
	{
		if (!pExplDesc[i].Used)
			break;
	}
	if (i == ExplCnt)
	{
		//Speicher fr neue Explosionen erstellen.
		ExplCnt += 10;
		pExplDesc = (TronExplDesc *) realloc (pExplDesc, ExplCnt * sizeof (TronExplDesc));
		for (j = i; j < ExplCnt; j++)
			pExplDesc[j].Used = false;
	}
	ActExpl = i;
	pExplDesc[ActExpl].Used = true;
	pExplDesc[ActExpl].Trans = 0.2f;
	if (Small)
		pExplDesc[ActExpl].FlameCnt = 10;
	else
		pExplDesc[ActExpl].FlameCnt = 30;
	for (i = 0; i < pExplDesc[ActExpl].FlameCnt; i++)
	{
		//Flammen erstellen
		pExplDesc[ActExpl].Flames[i].Pos = Vector3dMake (Pos->x + GetRandomVal (-2, 2), 0, Pos->y + GetRandomVal (-2, 2));
		pExplDesc[ActExpl].Flames[i].Dir = Vector3dMake (GetRandomVal (-5, 5), GetRandomVal (5, 20), GetRandomVal (-5, 5));
	}
}

void TronExplosion::DoEvents (float RelTimeDiff)
{
	int i, j;
	TronFire *pActFlame;
	D3DXVECTOR3 AddVec;

	for (i = 0; i < ExplCnt; i++)
	{
		if (pExplDesc[i].Used)
		{
			pExplDesc[i].Trans += RelTimeDiff / 5.0f;
			if (pExplDesc[i].Trans >= 1)
			{
				pExplDesc[i].Used = false;
				continue;
			}
			for (j = 0; j < pExplDesc[i].FlameCnt; j++)
			{
				pActFlame = &pExplDesc[i].Flames[j];
				if (pActFlame->Pos.y >= 0)
				{
					pActFlame->Dir.y -= 10 * RelTimeDiff;
					D3DXVec3Scale (&AddVec, &pActFlame->Dir, RelTimeDiff);
					D3DXVec3Add (&pActFlame->Pos, &pActFlame->Pos, &AddVec);
					if (pActFlame->Pos.x < 0)
					{
						pActFlame->Pos.x = 0;
						pActFlame->Dir.x *= -1;
					}
					if (pActFlame->Pos.x > Options.LandSize)
					{
						pActFlame->Pos.x = Options.LandSize;
						pActFlame->Dir.x *= -1;
					}
					if (pActFlame->Pos.z < 0)
					{
						pActFlame->Pos.z = 0;
						pActFlame->Dir.z *= -1;
					}
					if (pActFlame->Pos.z > Options.LandSize)
					{
						pActFlame->Pos.z = Options.LandSize;
						pActFlame->Dir.z *= -1;
					}
				}
			}
		}
	}
}

void TronExplosion::Render (D3DXVECTOR3 *CamPos)
{
	int i, j, ActIndex;
	Vertex vData[180];
	D3DMATERIAL8 Mat;

	Mat = Eng3d.GetDefaultMaterial ();
	Eng3d.SetZState (true, false);
	for (i = 0; i < ExplCnt; i++)
	{
		if (pExplDesc[i].Used)
		{
			ActIndex = 0;
			for (j = 0; j < pExplDesc[i].FlameCnt; j++)
			{
				if (pExplDesc[i].Flames[j].Pos.y >= 0)
				{
					BillboardVertexMake (&pExplDesc[i].Flames[j].Pos, 3, 3, CamPos, true, &vData[ActIndex]);
					//Auf Format TRIANGLELIST "konvertieren".
					vData[ActIndex + 5] = vData[ActIndex + 3];
					vData[ActIndex + 4] = vData[ActIndex + 1];
					vData[ActIndex + 3] = vData[ActIndex + 2];
					ActIndex += 6;
				}
			}
			if (ActIndex != 0)
			{
				Explosion.SetVertexData (vData, 0, ActIndex);
				Mat.Diffuse.a = 1 - pExplDesc[i].Trans;
				Explosion.SetMaterial (&Mat);
				Eng3d.Render3dPolygons (&Explosion, 0, ActIndex / 3);
			}
		}
	}
	Eng3d.SetZState (true, true);
}



//*******************TRON SHOTS*********************
void TronShots::Initialize (int NrStaticShots, TronWallDesc *pWallDesc, RECT *pRect, TronExplosion *pExplosion, void *pPlayer)
{
	int x, y;
	float TransErg;
	Vertex VData[4];
	D3DMATERIAL8 UseMat;

	//Schuss-Objekt laden
	Shot.Initialize (&Eng3d);
	Shot.LoadFromFile ("objects\\shot.x", 1);
	//Textur laden
	LightMap.Initialize (&Eng3d);
	LightMap.LoadForEdit (64, 64);
	if (LightMap.EditStart ())
	{
		for (x = 0; x < 64; x++)
		{
			for (y = 0; y < 64; y++)
			{
				TransErg = 255 - 8 * (float) sqrt ((32 - x) * (32 - x) + (32 - y) * (32 - y));
				if (TransErg > 0)
					LightMap.Edit (&Vector2dMake ((float) x, (float) y), (int) TransErg, 255, 255, 255);
				else
					LightMap.Edit (&Vector2dMake ((float) x, (float) y), 0, 255, 255, 255);
			}
		}
		LightMap.EditEnd ();
	}
	//Lichtquelle erstellen
	ShotLight.Initialize (&Eng3d, false);
	VData[0] = VertexMake (-5, 0, 5, 0, 1, 0, 0, 0);
	VData[1] = VertexMake (5, 0, 5, 0, 1, 0, 1, 0);
	VData[2] = VertexMake (-5, 0, -5, 0, 1, 0, 0, 1);
	VData[3] = VertexMake (5, 0, -5, 0, 1, 0, 1, 1);
	ShotLight.SetVertexData (VData, 0, 4);
	ShotLight.SetTexture (&LightMap);
	UseMat = MaterialMake (255, 255, 255, 0, 0, 0, 0, 100);
	ShotLight.SetMaterial (&UseMat);
	ShotLight.SetPolyFormat (TRIANGLESTRIP);
	ShotLight.SetVisibility (BOTHSIDES);
	//Lokale Einstellungen (Feld anlegen und Variablen setzen) treffen
	ShotCnt = NrStaticShots;
	pShotDesc = (TronShotDesc *) malloc (ShotCnt * sizeof (TronShotDesc));
	Delete ();
	this->pWallDesc = pWallDesc;
	this->pExplosion = pExplosion;
}

void TronShots::Delete ()
{
	int i;

	for (i = 0; i < ShotCnt; i++)
		pShotDesc[i].Used = false;	
}

TronShots::TronShots ()
{
	pShotDesc = NULL;
}

TronShots::~TronShots ()
{
	if (pShotDesc)
		free (pShotDesc);
}

void TronShots::FireShot (D3DXVECTOR2 *Pos, D3DXVECTOR2 *Dir, int ShotType, void *pFrom, bool MPServer)
{
	int i, Index = -1;
	D3DXVECTOR2 SPos, SDir;

	for (i = 0; i < ShotCnt; i++)
	{
		if (!pShotDesc[i].Used)
		{
			Index = i;
			break;
		}
	}
	if (Index == -1)
	{
		//Speicher erweitern
		Index = ShotCnt;
		ShotCnt *= 2;
		pShotDesc = (TronShotDesc *) realloc (pShotDesc, ShotCnt * sizeof (TronShotDesc));
		for (i = Index + 1; i < ShotCnt; i++)
			pShotDesc[i].Used = false;
	}
	//Schuss ins lokale Feld bernehmen
	SPos.x = Pos->x + Dir->x * 2;
	SPos.y = Pos->y + Dir->y * 2;
	if (SPos.x < 0)
		SPos.x = 0;
	else if (SPos.x > Options.LandSize)
		SPos.x = Options.LandSize;
	if (SPos.y < 0)
		SPos.y = 0;
	else if (SPos.y > Options.LandSize)
		SPos.y = Options.LandSize;
	SDir.x = Dir->x + GetRandomVal (-0.1f, 0.1f);
	SDir.y = Dir->y + GetRandomVal (-0.1f, 0.1f);
	D3DXVec2Normalize (&SDir, &SDir);
	D3DXVec2Scale (&SDir, &SDir, Options.MopedSpeed * 2);
	pShotDesc[Index].Pos = SPos;
	pShotDesc[Index].Dir = SDir;
	pShotDesc[Index].NoDanger = 1;
	pShotDesc[Index].ShotType = ShotType;
	if (ShotType == 0)					//normaler Schuss
		pShotDesc[Index].TTL = 25;
	else if (ShotType == 1)				//zielsuchender Schuss
		pShotDesc[Index].TTL = 20;
	else								//Superschuss
		pShotDesc[Index].TTL = 40;
	pShotDesc[Index].pFrom = pFrom;
	pShotDesc[Index].Used = true;
	if (Options.UseSounds)
	{
		//Sound abspielen
		FireShotSound.Set3dPosition (&Vector3dMake (Pos->x, 0, Pos->y));
		FireShotSound.SetPlayPosition (0);
		EngSound.Update3dSounds ();
		EngSound.Play3dSound (&FireShotSound, false);
	}
	if (MPServer)
	{
		SData[0] = CLIENT_PLAYSOUND;
		SData[1] = 2;
		memcpy (&SData[2], Pos, sizeof (D3DXVECTOR2));
		GameServer.SendData (0, SData, 2 + sizeof (D3DXVECTOR2), 0);
	}
}

D3DXVECTOR2 TronShots::GetDest (int Index)
{
	int i, MinIndex = -1;
	float ActDist, MinDist;
	bool First = true;
	D3DXVECTOR2 ActVec;

	//Fr den zielsuchenden Schuss das Ziel suchen
	for (i = 0; i < Options.OpponentCnt + 1; i++)
	{
		if (pPlayer[i].IsAlive () && pShotDesc[Index].pFrom != &pPlayer[i])
		{
			//Dieses Moped ist ein mglicher Abschusskandidat
			D3DXVec2Subtract (&ActVec, &pPlayer[i].GetPos (), &pShotDesc[Index].Pos);
			ActDist = D3DXVec2Length (&ActVec);
			if (First || ActDist < MinDist)
			{
				First = false;
				MinDist = ActDist;
				MinIndex = i;
			}
		}
	}
	if (MinIndex != -1)
		ActVec = pPlayer[MinIndex].GetPos ();
	else
		//Es konnte kein Ziel ausgemacht werden (d.h. es sind bereits alle Gegner tot). Zuflliges Ziel verwenden.
		ActVec = Vector2dMake (GetRandomVal (0, Options.LandSize), GetRandomVal (0, Options.LandSize));
	return (ActVec);
}

void TronShots::DoEvents (float RelTimeDiff, bool MPServer)
{
	int i, j, WIndex;
	bool bx, by, xBounce, wBounce;
	D3DXVECTOR2 PrevPos, ActDir, PrevDir;

	for (i = 0; i < ShotCnt; i++)
	{
		if (pShotDesc[i].Used)
		{
			//Schuss muss beachtet werden
			if (pShotDesc[i].ShotType != 0)
			{
				//Bei den Zielsuchern: das Ziel ausmachen
				D3DXVec2Subtract (&ActDir, &GetDest (i), &pShotDesc[i].Pos);
				D3DXVec2Normalize (&ActDir, &ActDir);
				D3DXVec2Normalize (&PrevDir, &pShotDesc[i].Dir);
				if (pShotDesc[i].ShotType == 1)
				{
					D3DXVec2Lerp (&ActDir, &PrevDir, &ActDir, RelTimeDiff * 3);		//= Trgheit von 1/3 Sekunde
					D3DXVec2Scale (&pShotDesc[i].Dir, &ActDir, Options.MopedSpeed * 1.1f);
				}
				else
				{
					D3DXVec2Lerp (&ActDir, &PrevDir, &ActDir, RelTimeDiff * 5);		//= Trgheit von 1/5 Sekunde
					D3DXVec2Scale (&pShotDesc[i].Dir, &ActDir, Options.MopedSpeed * 2);
				}
			}
			PrevPos = pShotDesc[i].Pos;
			pShotDesc[i].Pos.x += pShotDesc[i].Dir.x * RelTimeDiff;
			pShotDesc[i].Pos.y += pShotDesc[i].Dir.y * RelTimeDiff;
			pShotDesc[i].NoDanger -= RelTimeDiff;
			if (pShotDesc[i].NoDanger < 0)
				pShotDesc[i].NoDanger = 0;
			bx = false;
			by = false;
			wBounce = false;

			//Auf Wnde von den Mopeds berprfen		
			if (!pWallDesc->CanPlace (&PrevPos, &pShotDesc[i].Pos, -1, &xBounce, &WIndex))
			{
				if (xBounce)
					bx = true;
				else
					by = true;
			}
			//Auf die Wnde der Landschaft berprfen
			else if (pShotDesc[i].Pos.x < 0 || pShotDesc[i].Pos.x > Options.LandSize)
			{
				by = true;
				wBounce = true;
			}
			else if (pShotDesc[i].Pos.y < 0 || pShotDesc[i].Pos.y > Options.LandSize)
			{
				bx = true;
				wBounce = true;
			}
			if (bx || by)
			{
				if (wBounce)
				{
					//An der Auenwand abspringen. Dies tun alle Schsse.
					if (by)
						pShotDesc[i].Dir.x *= -1;
					else
						pShotDesc[i].Dir.y *= -1;
					pShotDesc[i].Pos = PrevPos;
					pShotDesc[i].Pos.x += pShotDesc[i].Dir.x * RelTimeDiff;
					pShotDesc[i].Pos.y += pShotDesc[i].Dir.y * RelTimeDiff;
					//TTL abziehen, aber nicht an Auenwnden explodieren lassen.
					pShotDesc[i].TTL--;
					if (pShotDesc[i].TTL == 0)
						pShotDesc[i].TTL = 1;
					continue;
				}
				if (pShotDesc[i].ShotType != 0 || pShotDesc[i].TTL == 1)
				{
					//Zielsucher: Explodieren an einer Moped-Wand bzw. beim Ausleben eines normalen Schusses
					pWallDesc->DestroyWall (WIndex, bx, &pShotDesc[i].Pos);
					pExplosion->CreateExplosion (&pShotDesc[i].Pos, true, MPServer);
					if (Options.UseSounds)
					{
						ShotCrashSound.Set3dPosition (&Vector3dMake (pShotDesc[i].Pos.x, 0, pShotDesc[i].Pos.y));
						ShotCrashSound.SetPlayPosition (0);
						EngSound.Update3dSounds ();
						EngSound.Play3dSound (&ShotCrashSound, false);
					}
					if (MPServer)
					{
						SData[0] = CLIENT_PLAYSOUND;
						SData[1] = 1;
						memcpy (&SData[2], &pShotDesc[i].Pos, sizeof (D3DXVECTOR2));
						GameServer.SendData (0, SData, 2 + sizeof (D3DXVECTOR2), 0);
					}
				}
				else
				{
					//Normaler Schuss: Abprallen an einer Moped-Wand
					if (by)
						pShotDesc[i].Dir.x *= -1;
					else
						pShotDesc[i].Dir.y *= -1;
					pShotDesc[i].Pos = PrevPos;
					pShotDesc[i].Pos.x += pShotDesc[i].Dir.x * RelTimeDiff;
					pShotDesc[i].Pos.y += pShotDesc[i].Dir.y * RelTimeDiff;
				}
				pShotDesc[i].TTL--;
				if (pShotDesc[i].TTL == 0)
				{
					//Schuss lschen
					pShotDesc[i].Used = false;
				}
			}
			//Prfen, ob der Schuss ein Moped trifft.
			for (j = 0; j < Options.OpponentCnt + 1; j++)
			{
				if (pPlayer[j].IsAlive () && fabs (pPlayer[j].GetPos ().x - pShotDesc[i].Pos.x) < 2 && fabs (pPlayer[j].GetPos ().y - pShotDesc[i].Pos.y) < 2)
				{
					if (pShotDesc[i].NoDanger > 0 && pShotDesc[i].pFrom == &pPlayer[j])
						//Eigenes Moped innerhalb der ersten Zeit nicht zerstren!
						continue;
					if (pPlayer[j].GetShieldTime () > 0)
					{
						//Dieses Moped verwendet gerade ein Schild, davon abprallen.
						pShotDesc[i].Dir.x *= -1;
						pShotDesc[i].Dir.y *= -1;
						pShotDesc[i].Pos = PrevPos;
						pShotDesc[i].Pos.x += pShotDesc[i].Dir.x * RelTimeDiff;
						pShotDesc[i].Pos.y += pShotDesc[i].Dir.y * RelTimeDiff;
						//Kein TTL abziehen, da sonst mglicherweise der Schuss sofort weg ist.
					}
					else
						//Dieses Moped explodieren lassen.
						pPlayer[j].WillCrash = true;
				}
			}
		}
	}
}

void TronShots::Render (bool RenderLight)
{
	int i;

	if (RenderLight)
	{
		//Das Licht rendern.
		Eng3d.SetZState (false, false);
		for (i = 0; i < ShotCnt; i++)
		{
			if (pShotDesc[i].Used)
			{
				//Das Licht von diesem Schuss rendern.
				ShotLight.SetPosition (&Vector3dMake (pShotDesc[i].Pos.x, 0, pShotDesc[i].Pos.y));
				if (pShotDesc[i].ShotType == 0)
					Eng3d.SetGlobalLight (ColorMake (0, 255, 0));
				else if (pShotDesc[i].ShotType == 1)
					Eng3d.SetGlobalLight (ColorMake (255, 0, 0));
				else
					Eng3d.SetGlobalLight (ColorMake (0, 0, 0));
				Eng3d.Render3dPolygons (&ShotLight, 0, 2);
			}
		}
		Eng3d.SetZState (true, true);
	}
	for (i = 0; i < ShotCnt; i++)
	{
		if (pShotDesc[i].Used)
		{
			//Diesen Schuss rendern.
			Shot.SetPosition (&Vector3dMake (pShotDesc[i].Pos.x, 0, pShotDesc[i].Pos.y));
			if (pShotDesc[i].ShotType == 0)
				Eng3d.SetGlobalLight (ColorMake (0, 255, 0));
			else if (pShotDesc[i].ShotType == 1)
				Eng3d.SetGlobalLight (ColorMake (255, 0, 0));
			else
				Eng3d.SetGlobalLight (ColorMake (0, 0, 0));
			Eng3d.Render3dObject (&Shot);
		}
	}
}

void TronShots::WriteData (BYTE **pMem)
{
	int i, ShotsUsed = 0;

	for (i = 0; i < ShotCnt; i++)
	{
		if (pShotDesc[i].Used)
			ShotsUsed++;
	}
	memcpy (*pMem, &ShotsUsed, sizeof (int));
	*pMem += sizeof (int);
	for (i = 0; i < ShotCnt; i++)
	{
		if (pShotDesc[i].Used)
		{
			memcpy (*pMem, &pShotDesc[i], sizeof (TronShotDesc));
			*pMem += sizeof (TronShotDesc);
		}
	}
}

void TronShots::ReadData (BYTE **pMem)
{
	int i;
	int RdInt;
	
	memcpy (&RdInt, *pMem, sizeof (int));
	*pMem += sizeof (int);
	if (RdInt > ShotCnt)
	{
		//Speicher erweitern
		pShotDesc = (TronShotDesc *) realloc (pShotDesc, RdInt * sizeof (TronShotDesc));
		ShotCnt = RdInt;
	}
	for (i = 0; i < ShotCnt; i++)
	{
		//Gesamten Speicher durchlaufen, aber nur die verwendeten Daten kopieren.
		if (i < RdInt)
		{
			memcpy (&pShotDesc[i], *pMem, sizeof (TronShotDesc));
			*pMem += sizeof (TronShotDesc);
		}
		else
			pShotDesc[i].Used = false;
	}
}