//defines
#define TRON_MVERS 1				//Hauptversions-Nummer
#define TRON_SVERS 4				//Subversions-Nummer
#define TRON_MPVERS 1				//Multiplayer-Version. Muss bei einer nderung des Multiplayer-Protokolls erhht werden!

#define MAX_MENUITEM 16
#define MAX_MENUTEXT 64
#define MENUITEM_WIDTH 550
#define MENUITEM_HEIGHT 40

#define MENU_MAIN 0
#define MENU_MULTIPLAYER 1
#define MENU_MULTIPLAYER_SERVER 2
#define MENU_MULTIPLAYER_CLIENT 3
#define MENU_MULTIPLAYER_CLIENTWAIT 4
#define MENU_OPTIONS 5
#define MENU_OPTIONS_GRAFICS 6
#define MENU_OPTIONS_SOUNDS 7
#define MENU_OPTIONS_DETAILS 8
#define MENU_OPTIONS_GAMEPLAY 9

#define CONNECTION_OK 0				//Die Verbindung steht.
#define CONNECTION_LOST 255			//Die Verbindung wurde getrennt (nicht 0 verwenden!)
#define CLIENT_SETTINGS 1			//Erste Kontaktaufnahme zu jedem Client: Server-Einstellungen, alle Bike-Farbnummern, jeweilige Client-Bikenummer
#define SERVER_SETTINGSOK 2			//Client antwortet auf CLIENT_SETTINGS mit eigenem Client-Name (fr den Server)
#define CLIENT_LOAD 3				//Server sendet den Ladebefehl an jeden Client: bermittelt die Spieler-Namen
#define SERVER_LOADOK 4				//Client antwortet mit OK auf CLIENT_LOAD: Der Ladevorgang ist abgeschlossen.
#define CLIENT_START 5				//Nachdem SERVER_LOADOK von jedem Client eingetroffen ist, wird CLIENT_START vom Server geschickt.
#define SERVER_EXEC 6				//Der Server bekommt die Nachricht, dass ein Client eine Aktion durchfhren will. Beinhaltet die Client-Bikenummer und die Client-Aktionsnummer.
#define CLIENT_GAMESTATE 7			//Ein Client bekommt diese Nachricht am laufenden Band. Sie bermittelt dem Client die aktuellen Spieldaten.
#define CLIENT_PLAYSOUND 8			//Ein Client bekommt die Nachricht, dass er einen Sound abspielen soll. Beinhaltet die Sound-ID und die Position.
#define CLIENT_EXPLOSION 9			//Ein Client soll eine Explosion darstellen. Beinhaltet Gre und Position.



//**************MULTIPLAYER********************
typedef struct TronMPSettings
{
	int PlayerCnt;
	unsigned long *pIdents;
	int AIPlayerCnt;
} TronMPSettings;

typedef struct TronMPBikeIdent
{
	int BikeType;
	char Name[64];
} TronMPBikeIdent;




//Strukturen, Klassen
typedef struct TronSettings
{
	//Grafik
	Revo3dRes Res;
	int UseFSAA;
	bool UseReflection;
	int UseShadow;
	bool UseSpecular;
	bool UseAnisotropic;
	bool UseTransWalls;
	//Sound
	bool UseMusic;
	bool UseSounds;
	//Gameplay
	float LandSize;
	float MopedSpeed;
	int FragLimit;
	int OpponentCnt;
	int OpponentSkill;
	int Cam;
	bool EnableAction;
	char PlayerName[64];
} TronSettings;

typedef struct TronWall
{
	D3DXVECTOR2 Pos1, Pos2;
	int From;
	bool Used;
} TronWall;

typedef struct TronFire
{
	D3DXVECTOR3 Pos, Dir;
} TronFire;

typedef struct TronExplDesc
{
	TronFire Flames[30];
	int FlameCnt;
	float Trans;
	bool Used;
} TronExplDesc;

typedef struct TronShotDesc
{
	D3DXVECTOR2 Pos, Dir;
	float NoDanger;
	int TTL, ShotType;
	bool Used;
	void *pFrom;
} TronShotDesc;


class TronWalls
{
private:
	Revo3dPolygons Walls, WallsTop;
	int WallCnt, WallUsed;
	int r, g, b;
	D3DMATERIAL8 TransMat, SolidMat;

public:
	float Trans;

	void Initialize (int NrStaticWalls, int r, int g, int b, Revo2dTexture *pTex);
	void AddWall ();
	void UpdateWall (D3DXVECTOR2 *Pos1, D3DXVECTOR2 *Pos2);
	void Rebuild (void *pWDesc, int Player, bool Multiplayer);
	void Delete ();
	void TransIncr (float RelTimeDiff);
	void Render ();
	void RenderTrans ();
	D3DMATERIAL8 *GetTransMat () { return (&TransMat); }
	D3DMATERIAL8 *GetSolidMat () { return (&SolidMat); }
};

class TronWallDesc
{
private:
	int WallCnt, WallUsed;
	TronWall *pWallDesc;

public:
	TronWallDesc ();
	~TronWallDesc ();
	void Initialize (int NrStaticWalls);
	int AddWall (int From);
	void UpdateWall (D3DXVECTOR2 *Pos1, D3DXVECTOR2 *Pos2, int WallIndex);
	void DeleteWall (int WallIndex);
	void DestroyWall (int WallIndex, bool xWall, D3DXVECTOR2 *pPos);
	bool CanPlace (D3DXVECTOR2 *Pos1, D3DXVECTOR2 *Pos2, int OwnWallIndex, bool *xWall = NULL, int *pWIndex = NULL);
	void DeletePlayer (int Player);
	void Delete () { WallUsed = 0; }
	float GetWP (D3DXVECTOR2 *pPos, D3DXVECTOR2 *pDir);
	int GetWallUsed () { return (WallUsed); }
	TronWall *GetWallDesc () { return (pWallDesc); }
	void WriteData (BYTE **pMem);
	void ReadData (BYTE **pMem);
};

class TronExplosion
{
private:
	Revo3dPolygons Explosion;
	int ExplCnt;
	TronExplDesc *pExplDesc;

public:
	TronExplosion ();
	~TronExplosion ();
	void Initialize (int NrStaticExplosion, Revo2dTexture *pTex);
	void CreateExplosion (D3DXVECTOR2 *Pos, bool Small, bool MPServer);
	void DoEvents (float RelTimeDiff);
	void Render (D3DXVECTOR3 *CamPos);
	void WriteData (BYTE **pMem);
	void ReadData (BYTE **pMem);
};

class TronShots
{
private:
	Revo3dObject Shot;
	Revo3dPolygons ShotLight;
	Revo2dTexture LightMap;
	TronShotDesc *pShotDesc;
	TronWallDesc *pWallDesc;
	TronExplosion *pExplosion;
	int ShotCnt;

	D3DXVECTOR2 GetDest (int Index);

public:
	TronShots ();
	~TronShots ();
	void Initialize (int NrStaticShots, TronWallDesc *pWallDesc, RECT *pRect, TronExplosion *pExplosion, void *pPlayer);
	void Delete ();
	void FireShot (D3DXVECTOR2 *Pos, D3DXVECTOR2 *Dir, int ShotType, void *pFrom, bool MPServer);
	void DoEvents (float RelTimeDiff, bool MPServer);
	void Render (bool RenderLight);
	void WriteData (BYTE **pMem);
	void ReadData (BYTE **pMem);
};

class TronPlayer
{
private:
	Revo3dObject Moped, Wheel, SWheel;
	TronWalls Wall;
	TronWallDesc *pWallDesc;
	D3DXVECTOR2 Pos, PrevPos, LastWallPos, ActWallPos, LastOWPos;
	D3DXVECTOR2 Dir;
	D3DXVECTOR3 BackRPos, FrontRPos;
	float YAngle, RotationLeft, ItemTime, SpeedTime, ShieldTime;
	float AIReact, AILastCh, AILastHook;
	int WallDescIndex, Ident;
	bool Alive, MPSrv;

	bool CanMopedPlace (D3DXVECTOR2 *CrashPos, TronPlayer *pOpp);
	//TRON AI III
	float AIGetWayLen (D3DXVECTOR2 Pos, D3DXVECTOR2 LastDir, int d, int *pDir);
	bool AIChDir (TronPlayer **pField, int *pDir, float RelTimeDiff);

public:
	int Frags, ItemAviable, Rank;
	bool WillCrash;
	char Name[64];

	bool Initialize (int MopedNr, TronWallDesc *pWallDesc, int Ident, bool Multiplayer, bool IsServer, Revo2dTexture *pWallTex);
	void Reset ();
	void Crash ();
	void Delete ();
	void GeneratePos ();
	D3DXVECTOR2 GetChangeDir (int TurnDir);
	void ChangeDir (int TurnDir);
	void AI (float RelTimeDiff, TronShots *pShots);
	void DoEvents (float RelTimeDiff, float MopedSpeed);
	D3DXVECTOR2 GetPos () { return (Pos); }
	D3DXVECTOR2 GetPrevPos () { return (PrevPos); }
	D3DXVECTOR2 GetDir () { return (Dir); }
	TronWalls *GetWalls () { return (&Wall); }
	bool IsAlive () { return (Alive); }
	void SetAlive (bool Alive) { this->Alive = Alive; }
	void HandleShield ();
	float GetShieldTime () { return (ShieldTime); }
	bool CanPlace ();
	void FireItem (TronShots *pShots);
	void WallCrash ();
	void Render (bool Shadow, Revo3dObject *pShield);
	void RenderWalls (bool Mirrored);
	void WriteData (BYTE **pMem);
	void ReadData (BYTE **pMem);
};

class CameraHandler
{
private:
	D3DXVECTOR3 CamPos, BlendPos;
	D3DXVECTOR3 CamDest, BlendDest;
	float CamAng, BlendAng;
	float PosTime, DestTime, AngleTime;

public:
	CameraHandler ();
	void SetPos (D3DXVECTOR3 *Pos) { CamPos = *Pos; }
	void SetDest (D3DXVECTOR3 *Dest) { CamDest = *Dest; }
	float GetAngle () { return (CamAng); }
	D3DXVECTOR3 *GetPos () { return (&CamPos); }
	D3DXVECTOR3 *GetDest () { return (&CamDest); }
	bool IsAngleBlended () { return (AngleTime != 0); }
	void BlendPosition (D3DXVECTOR3 *NewPos, float Time);
	void BlendDestination (D3DXVECTOR3 *NewDest, float Time);
	void BlendAngle (float ActAngle, float NewAngle, float Time);
	void DoEvents (float RelTimeDiff);
	void SetCamera ();
};


//Networking-Klassen
class ActionBuffer
{
private:
	int ActionCnt;
	int Idents[100];
	BYTE Actions[100];

public:
	ActionBuffer ();
	void Reset ();
	int GetActionCnt () { return (ActionCnt); }
	void GetActions (int *DestIdents, BYTE *DestActions);
	void AddAction (int Ident, BYTE Action);
};

class RenderBuffer
{
private:
	BYTE pRenderData[131072];
	int DataSize;

public:
	bool NewData;

	RenderBuffer ();
	void Reset ();
	int GetDataSize () { return (DataSize); }
	void GetData (BYTE *Dest);
	void SetData (BYTE *Src, int Size);
};


//Funktionen
bool OneTimeDXInit (HINSTANCE hInst, int ResX, int ResY, D3DFORMAT Format, int AntiAlias);
void GeometryInit (float SizeX, float SizeZ, float WallHeight, Vertex *outV);

void SCallback (BYTE *pData, int DataSize);
DWORD WINAPI SenderThread (LPVOID pParam);
void CCallback (BYTE *pData, int DataSize);

void ShowMenu (int MenuID, D3DXVECTOR2 CursPos, Revo2dTexture *pCurs, Revo2dTexture *pBusy, Revo2dTexture *pButton, Revo2dText *pText, Revo2dTexture *pBackgr);
void ShowInfo (Revo2dTexture *pBackgr, bool Credits);
int CreateMenuText (int MenuID, char MenuStr[MAX_MENUITEM][MAX_MENUTEXT], Revo3dRes *pNR = NULL, int *pNF = NULL, bool EditName = false);
void MasterGameController (float RelTimeDiff, bool Multiplayer);
void RunGame (bool Multiplayer, bool IsServer);
void GetCamDesc (D3DXVECTOR2 ActPos, D3DXVECTOR2 CamDir, D3DXVECTOR3 *CamPos, D3DXVECTOR3 *CamDest);
int DoRanking ();
void InitRandom ();
float GetRandomVal (float Min, float Max);
bool InArea (float Min, float Max, float Val);
bool InAreaDelta (float Min, float Max, float Val, float Delta);
bool InLine (float v11, float v12, float v21, float v22);